/***********************************************************

             YUV image input/output functions

 ***********************************************************/
#pragma once

#include <stdio.h>

#include "GVDsetup.h"

#include <opencv2/opencv.hpp>

class CIYuv
{
public:
  CIYuv ();
  CIYuv (int h, int w, int chroma_format);
  virtual ~CIYuv ();

  bool Resize (int h, int w, int chroma_format);
  bool ReadOneFrame (FILE *fp, int frameno = -1);
  bool WriteOneFrame (FILE *fp);

  bool WriteOneFrame_inYUV(FILE *fp); // GIST test

  void setDataFromImgYUV( IplImage *imgYUV );
  void setDataFromDepth( IplImage *imgDepth );
  bool setData444_inIYUV( CIYuv *yuvSrc );
  bool setData444_inIYUV420_onlyLEFTSIDE(CIYuv *yuvSrc, bool half=false); // Oct25,2012

  bool depthEnhanced( CIYuv *yuvSrc, unsigned int th= 5 );

  void setTextureFromImgYUV444( IplImage *imgYUV ); // for 3 plains
  void setDepthFromImgYUV444( IplImage *imgYUV ); // for single plain

  bool setUpsampleFilter (unsigned int filter, unsigned int precision);
  bool upsampling (CIYuv *src, int padding_size=0);
  bool BQF_V (CIYuv *src);
  bool BQF_V3 (CIYuv *src, unsigned int side);
  bool BQF_D (CIYuv *src);
  bool BQF_D3 (CIYuv *src, unsigned int side);

  void BQF_V_Expand(CIYuv *src);
  void BQF_V_Single(CIYuv *src);
  void BQF_V_Split(CIYuv *src, unsigned int side);
  void BQF_H(CIYuv *src);
  void BQF_Single ( CIYuv *src );

  void BQF_StackSplit ( CIYuv *src, unsigned int side );

  //void looseEdgeBindResiduals( CIYuv *data, CIYuv *map, unsigned int m_mode );
  void looseEdgeBindResiduals( CIYuv *map, unsigned int m_mode );

  bool LPF_V (CIYuv *src, bool full= true);
  bool LPF_H (CIYuv *src, bool full= true);

  void LPF_V_DExpand(CIYuv *src, unsigned int th= 5, bool full= true);
  void LPF_V_DSingle(CIYuv *src, unsigned int th= 5, bool full= true);
  void LPF_V_DSplit(CIYuv *src, unsigned int side, unsigned int th= 5, bool full= true);
  void LPF_H_DSmooth(CIYuv *src, unsigned int th= 5, bool full= true);

  void LPF_V_RExpand(CIYuv *src, unsigned int th= 5);
  void LPF_V_RSingle(CIYuv *src, unsigned int th= 5);
  void LPF_V_RSplit(CIYuv *src, unsigned int side, unsigned int th= 5);
  void LPF_H_RSmooth(CIYuv *src, unsigned int th= 5);

  int getSampling () { return sampling; }
  int getHeight () { return height; }
  int getWidth () { return width; }
  int getHeightUV () { return heightUV; }
  int getWidthUV () { return widthUV; }
  BYTE*** getData () { return comp; }
  BYTE* getBuffer () { return pBuffer; }

  BYTE **Y;
  BYTE **U;
  BYTE **V;
  BYTE **Ytemp;

private:
  bool allocate_mem ();
  void release_mem ();

  int height;
  int width;
  int heightUV;
  int widthUV;
  int picsizeY;
  int picsizeUV;
  int size_in_byte;
  int sampling;

  void (*FuncFilter)(BYTE **in, BYTE **out, int width, int height, int padding_size);

  BYTE **comp[3];

  BYTE *pBuffer; //!> 1D array for the Y, U and V component.
};
