#pragma once
// View Synthesis Decoder
// Original code: View Synthesis Reference Software (VSRS) 3.5
// Last Update: Feb, 2014.
#include <utility>
#include <string>
#include "SubMixOptions.h"
#include "SubSep.h"
#include "ReturnCode.h"
#include "ParameterDec.h"
#include "ViewSynthesis.h"

namespace view_synthesis {
  namespace view_subsep_options {
    using namespace std;

    CParameterEncResiduals cParameter;

    unsigned int mode;
    unsigned int numView;
    typedef pair<unsigned int, string> modepair;
    modepair mode1(QUARTER, "QUARTER"); // VIEW_2
    modepair mode11(QUARTER_L, "QUARTER_L"); // for VIEW_2 with Full hight/width.
    // HALF_QUARTER, OUT                    VIEW_2
    modepair mode2(QUARTER_STACK, "QUARTER_STACK");
    modepair mode3(HALF_QUARTER_STACK, "HALF_QUARTER_STACK");
    modepair mode4(QUARTER_ODD, "QUARTER_ODD"); // Single global depth(quarter size) with Stacked Residuals(quarter size x 2)
    modepair mode5(QUARTER_ODD_L, "QUARTER_ODD_L"); // Single global depth(quarter size) with Stacked Residuals(quarter size x 2, padding to full size)
    modepair mode6(FULL_ODD_L, "FULL_ODD_L"); // Single global depth(full size) with Stacked Residuals(quarter size x 2, padding to full size)

    unsigned int checkOption(string option) {
      if ( 0 == option.compare(mode1.second) ) {mode= mode1.first; numView= VIEW_2; return mode;}
      if ( 0 == option.compare(mode11.second) ) {mode= mode11.first; numView= VIEW_2; return mode;}
      if ( 0 == option.compare(mode2.second) ) {mode= mode2.first; numView= VIEW_3; return mode;}
      if ( 0 == option.compare(mode3.second) ) {mode= mode3.first; numView= VIEW_3; return mode;}
      if ( 0 == option.compare(mode4.second) ) {mode= mode4.first; numView= VIEW_3; return mode;}
      if ( 0 == option.compare(mode5.second) ) {mode= mode5.first; numView= VIEW_3; return mode;}
      if ( 0 == option.compare(mode6.second) ) {mode= mode6.first; numView= VIEW_3; return mode;}
      cout << "GVD: ERR: Config parameter: Mode: ->" << option << "<- Unknown." << endl << endl;
      return UNKNOWN;
    }


    bool VariParam= false;
    unsigned int width, height, half_width, half_height;
    unsigned int start, total;
    double near1, far1, near2, far2, near3, far3;
    double nearV1, farV1, nearV2, farV2;
    double nearT1, farT1, nearT2, farT2;
    class NoEven {};
    class BadSeq {};
    class MixPosNeg {};

    int initParam() {
      try {
        width= cParameter.getSourceWidth();
        height= cParameter.getSourceHeight();
        half_width= width / 2;
        half_height= height / 2;
        if ( width % 2 != 0 || height % 2 != 0) throw NoEven();
      } catch (NoEven) {
        cout << "GVD: ERR: Config parameter: ""SourceWidth"" or ""SourceHeight"" Odd, No even." << endl << endl;
        return BAD_CONFIG_OPTIONS;
      } catch (...) {
        cout << "GVD: ERR: Config parameter: ""SourceWidth"" or ""SourceHeight""." << endl << endl;
        return BAD_CONFIG_OPTIONS;
      }

      try {
        start= cParameter.getStartFrame();
        total= cParameter.getNumberOfFrames();
        // if ( start >= total ) throw BadSeq();
      } catch (BadSeq) {
        cout << "GVD: ERR: Config parameter: ""StartFrame"", ""NumberOfFrames""  Bad order." << endl << endl;
        return BAD_CONFIG_OPTIONS;
      } catch (...) {
        cout << "GVD: ERR: Config parameter: ""StartFrame"" or ""NumberOfFrames""." << endl << endl;
        return BAD_CONFIG_OPTIONS;
      }

      if ( cParameter.getNearestDepthValue1() == 0 && cParameter.getFarthestDepthValue1() == 0 )
        VariParam= true;
      else
        try {
          near1= cParameter.getNearestDepthValue1();
          far1= cParameter.getFarthestDepthValue1();
          near2= cParameter.getNearestDepthValue2();
          far2= cParameter.getFarthestDepthValue2();
          if ( numView == VIEW_2 ) {
            near3= 0;
            far3= 0;
            if ( near1<0 && (far1>0 || near2>0 || far2>0) ) throw MixPosNeg();
            if ( near1>0 && (far1<0 || near2<0 || far2<0) ) throw MixPosNeg();
            if ( (near1==0 && far1==0) || (near2==0 && far2==0) ) throw MixPosNeg();
            if ( near1==far1 || near2==far2 ) throw MixPosNeg();
          } else {
            near3= cParameter.getNearestDepthValue3();
            far3= cParameter.getFarthestDepthValue3();
            if ( near1<0 && (far1>0 || near2>0 || far2>0 || near3>0 || far3>0) ) throw MixPosNeg();
            if ( near1>0 && (far1<0 || near2<0 || far2<0 || near3<0 || far3<0) ) throw MixPosNeg();
            if ( (near1==0 && far1==0) || (near2==0 && far2==0)  || (near3==0 && far3==0) ) throw MixPosNeg();
            if ( near1==far1 || near2==far2 || near3==far3 ) throw MixPosNeg();
          }
        } catch (MixPosNeg){
          cout << "GVD: ERR: Config parameter: ""NearestDepthValue"" and ""FarthestDepthValue""  Mixed positive and negative." << endl << endl;
          return BAD_CONFIG_OPTIONS;
        } catch (...) {
          cout << "GVD: ERR: Config parameter: ""NearestDepthValue"" or ""FarthestDepthValue""." << endl << endl;
          return BAD_CONFIG_OPTIONS;
        }

      return NORMAL_END;
    }


    FILE *f_V, *f_BindD, *f_BindR;
    FILE *f_T1, *f_T2, *f_T3, *f_T4, *f_T6, *f_T7, *f_T8, *f_T9;
    #ifdef WIN32
    errno_t err_T1, err_T2, err_T3, err_T4, err_T6, err_T7, err_T8, err_T9;
    #endif
    class NoInfile {};
    class NoOutfile {};
    // for VariParam
    FILE *f_NFTfile;

    int initFile() {
      try {
        #ifdef WIN32
        f_V= _fsopen(cParameter.get_VName().c_str(), "rb", _SH_DENYWR); if (f_V == NULL) throw NoInfile();
        f_BindD= _fsopen(cParameter.getBind_DName().c_str(), "rb", _SH_DENYWR); if (f_BindD == NULL) throw NoInfile();
        f_BindR= _fsopen(cParameter.getBind_RName().c_str(), "rb", _SH_DENYWR); if (f_BindR == NULL) throw NoInfile();
        err_T1= fopen_s(&f_T1, cParameter.get_T1Name().c_str(), "wb"); if (err_T1 != 0) throw NoOutfile();
        err_T2= fopen_s(&f_T2, cParameter.get_T2Name().c_str(), "wb"); if (err_T2 != 0) throw NoOutfile();
        err_T3= fopen_s(&f_T3, cParameter.get_T3Name().c_str(), "wb"); if (err_T3 != 0) throw NoOutfile();
        err_T4= fopen_s(&f_T4, cParameter.get_T4Name().c_str(), "wb"); if (err_T4 != 0) throw NoOutfile();

        if ( VariParam ) {
          f_NFTfile= _fsopen(cParameter.getNFTFileName().c_str(), "rb", _SH_DENYWR); if (f_NFTfile == NULL) throw NoInfile();
          if ( fseek(f_NFTfile, 0, SEEK_SET) )  throw NoInfile();
        }

        #else
        f_V= fopen(cParameter.get_VName().c_str(), "rb"); if (f_V == NULL) throw NoInfile();
        f_BindD= fopen(cParameter.getBind_DName().c_str(), "rb"); if (f_BindD == NULL) throw NoInfile();
        f_BindR= fopen(cParameter.getBind_RName().c_str(), "rb"); if (f_BindR == NULL) throw NoInfile();
        f_T1= fopen(cParameter.get_T1Name().c_str(), "wb"); if (f_T1 == NULL) throw NoOutfile();
        f_T2= fopen(cParameter.get_T2Name().c_str(), "wb"); if (f_T2 == NULL) throw NoOutfile();
        f_T3= fopen(cParameter.get_T3Name().c_str(), "wb"); if (f_T3 == NULL) throw NoOutfile();
        f_T4= fopen(cParameter.get_T4Name().c_str(), "wb"); if (f_T4 == NULL) throw NoOutfile();

        if ( VariParam ) {
          f_NFTfile= fopen(cParameter.getNFTFileName().c_str(), "rb"); if (f_NFTfile == NULL) throw NoInfile();
          if ( fseek(f_NFTfile, 0, SEEK_SET) )  throw NoInfile();
        }
        #endif

      } catch (NoInfile) {
        cout << "GVD: ERR:: Not specify files, or File path/IO error." << endl;
        if (f_V != NULL) fclose(f_V);
        if ( VariParam ) {
          if (f_NFTfile != NULL) fclose(f_NFTfile);
        }
        return NO_FILEIO;
      } catch (...) {
        cout << "GVD: ERR:: Not specify output file, or File path/IO error." << endl;
        return NO_FILEIO;
      }

      if ( mode != QUARTER ) {
        try {
          #ifdef WIN32
          err_T6= fopen_s(&f_T6, cParameter.get_T6Name().c_str(), "wb"); if (err_T6 != 0) throw NoOutfile();
          err_T7= fopen_s(&f_T7, cParameter.get_T7Name().c_str(), "wb"); if (err_T7 != 0) throw NoOutfile();
          err_T8= fopen_s(&f_T8, cParameter.get_T8Name().c_str(), "wb"); if (err_T8 != 0) throw NoOutfile();
          err_T9= fopen_s(&f_T9, cParameter.get_T9Name().c_str(), "wb"); if (err_T9 != 0) throw NoOutfile();

          #else
          f_T6= fopen(cParameter.get_T6Name().c_str(), "wb"); if (f_T6 == NULL) throw NoOutfile();
          f_T7= fopen(cParameter.get_T7Name().c_str(), "wb"); if (f_T7 == NULL) throw NoOutfile();
          f_T8= fopen(cParameter.get_T8Name().c_str(), "wb"); if (f_T8 == NULL) throw NoOutfile();
          f_T9= fopen(cParameter.get_T9Name().c_str(), "wb"); if (f_T9 == NULL) throw NoOutfile();

          #endif
        } catch (NoOutfile) {
          cout << "GVD: ERR: QUARTER_STACK or HALF_QUARTER_STACK: Not specify output files, or File path/IO error." << endl;
          return NO_FILEIO;
        } catch (...) {
          cout << "GVD: ERR: QUARTER_STACK or HALF_QUARTER_STACK: File error." << endl;
          return NO_FILEIO;
        }
      }

      return NORMAL_END;
    }



    void endFile() {
      fclose(f_V); fclose(f_BindD); fclose(f_BindR); fclose(f_T1); fclose(f_T2); fclose(f_T3); fclose(f_T4);
      if ( VariParam ) fclose(f_NFTfile);

      if ( numView == VIEW_3 ) {
        fclose(f_T6); fclose(f_T7); fclose(f_T8); fclose(f_T9);
      }
    }



    CSubSep cSubV;

    int initSub() {
      #ifdef PTRACE
        printf("TRACE: view_synthesis::view_subsep_options::initSub()\n");
      #endif

      if ( numView == VIEW_2 ) {
        if ( !cSubV.Init(
                  mode, numView, VariParam, width, height, half_width, half_height,
                  near1, far1, near2, far2, 0, 0,
                  cParameter.getCameraName1().c_str(),
                  cParameter.getCameraName2().c_str(),
                  "",
                  cParameter.getVertualCameraName1().c_str(),
                  "",
                  cParameter.getTargetCameraName1().c_str(),
                  cParameter.getTargetCameraName2().c_str(),
                  cParameter.getTargetCameraName3().c_str(),
                  cParameter.getTargetCameraName4().c_str(),
                  "", "", "", "",
                  cParameter.getMat_In_1(), cParameter.getMat_Ex_1(), cParameter.getMat_Trans_1(),
                  cParameter.getMat_In_2(), cParameter.getMat_Ex_2(), cParameter.getMat_Trans_2(),
                  NULL, NULL, NULL,
                  cParameter.getMat_In_T3(), cParameter.getMat_Ex_T3(), cParameter.getMat_Trans_T3(),
                  NULL, NULL, NULL,
                  cParameter.getMat_In_T1(), cParameter.getMat_Ex_T1(), cParameter.getMat_Trans_T1(),
                  cParameter.getMat_In_T2(), cParameter.getMat_Ex_T2(), cParameter.getMat_Trans_T2(),
                  cParameter.getMat_In_T3(), cParameter.getMat_Ex_T3(), cParameter.getMat_Trans_T3(),
                  cParameter.getMat_In_T4(), cParameter.getMat_Ex_T4(), cParameter.getMat_Trans_T4(),
                  NULL, NULL, NULL,
                  NULL, NULL, NULL,
                  NULL, NULL, NULL,
                  NULL, NULL, NULL
        ) ) { noInitSUBMIX; return NO_INIT_SUBMIX_D; }
      }

      if ( mode == QUARTER_STACK || mode == HALF_QUARTER_STACK ) {
        if ( !cSubV.Init(
                  mode, numView, VariParam, width, height, half_width, height,
                  near1, far1, near2, far2, near3, far3,
                  cParameter.getCameraName1().c_str(),
                  cParameter.getCameraName2().c_str(),
                  cParameter.getCameraName3().c_str(),
                  cParameter.getVertualCameraName1().c_str(),
                  cParameter.getVertualCameraName2().c_str(),
                  cParameter.getTargetCameraName1().c_str(),
                  cParameter.getTargetCameraName2().c_str(),
                  cParameter.getTargetCameraName3().c_str(),
                  cParameter.getTargetCameraName4().c_str(),
                  cParameter.getTargetCameraName6().c_str(),
                  cParameter.getTargetCameraName7().c_str(),
                  cParameter.getTargetCameraName8().c_str(),
                  cParameter.getTargetCameraName9().c_str(),
                  cParameter.getMat_In_1(), cParameter.getMat_Ex_1(), cParameter.getMat_Trans_1(),
                  cParameter.getMat_In_2(), cParameter.getMat_Ex_2(), cParameter.getMat_Trans_2(),
                  cParameter.getMat_In_3(), cParameter.getMat_Ex_3(), cParameter.getMat_Trans_3(),
                  cParameter.getMat_In_V1(), cParameter.getMat_Ex_V1(), cParameter.getMat_Trans_V1(),
                  cParameter.getMat_In_V2(), cParameter.getMat_Ex_V2(), cParameter.getMat_Trans_V2(),
                  cParameter.getMat_In_T1(), cParameter.getMat_Ex_T1(), cParameter.getMat_Trans_T1(),
                  cParameter.getMat_In_T2(), cParameter.getMat_Ex_T2(), cParameter.getMat_Trans_T2(),
                  cParameter.getMat_In_T3(), cParameter.getMat_Ex_T3(), cParameter.getMat_Trans_T3(),
                  cParameter.getMat_In_T4(), cParameter.getMat_Ex_T4(), cParameter.getMat_Trans_T4(),
                  cParameter.getMat_In_T6(), cParameter.getMat_Ex_T6(), cParameter.getMat_Trans_T6(),
                  cParameter.getMat_In_T7(), cParameter.getMat_Ex_T7(), cParameter.getMat_Trans_T7(),
                  cParameter.getMat_In_T8(), cParameter.getMat_Ex_T8(), cParameter.getMat_Trans_T8(),
                  cParameter.getMat_In_T9(), cParameter.getMat_Ex_T9(), cParameter.getMat_Trans_T9()
        ) ) { noInitSUBMIX; return NO_INIT_SUBMIX_D; }
      }

      if ( mode == QUARTER_ODD || mode == QUARTER_ODD_L || mode == FULL_ODD_L ) {
        if ( !cSubV.Init(
                  mode, numView, VariParam, width, height, half_width, height,
                  near1, far1, near2, far2, near3, far3,
                  cParameter.getCameraName1().c_str(),
                  cParameter.getCameraName2().c_str(),
                  cParameter.getCameraName3().c_str(),
                  cParameter.getCameraName2().c_str(),
                  cParameter.getCameraName2().c_str(),
                  cParameter.getTargetCameraName1().c_str(),
                  cParameter.getTargetCameraName2().c_str(),
                  cParameter.getTargetCameraName3().c_str(),
                  cParameter.getTargetCameraName4().c_str(),
                  cParameter.getTargetCameraName6().c_str(),
                  cParameter.getTargetCameraName7().c_str(),
                  cParameter.getTargetCameraName8().c_str(),
                  cParameter.getTargetCameraName9().c_str(),
                  cParameter.getMat_In_1(), cParameter.getMat_Ex_1(), cParameter.getMat_Trans_1(),
                  cParameter.getMat_In_2(), cParameter.getMat_Ex_2(), cParameter.getMat_Trans_2(),
                  cParameter.getMat_In_3(), cParameter.getMat_Ex_3(), cParameter.getMat_Trans_3(),
                  cParameter.getMat_In_2(), cParameter.getMat_Ex_2(), cParameter.getMat_Trans_2(),
                  cParameter.getMat_In_2(), cParameter.getMat_Ex_2(), cParameter.getMat_Trans_2(),
                  cParameter.getMat_In_T1(), cParameter.getMat_Ex_T1(), cParameter.getMat_Trans_T1(),
                  cParameter.getMat_In_T2(), cParameter.getMat_Ex_T2(), cParameter.getMat_Trans_T2(),
                  cParameter.getMat_In_T3(), cParameter.getMat_Ex_T3(), cParameter.getMat_Trans_T3(),
                  cParameter.getMat_In_T4(), cParameter.getMat_Ex_T4(), cParameter.getMat_Trans_T4(),
                  cParameter.getMat_In_T6(), cParameter.getMat_Ex_T6(), cParameter.getMat_Trans_T6(),
                  cParameter.getMat_In_T7(), cParameter.getMat_Ex_T7(), cParameter.getMat_Trans_T7(),
                  cParameter.getMat_In_T8(), cParameter.getMat_Ex_T8(), cParameter.getMat_Trans_T8(),
                  cParameter.getMat_In_T9(), cParameter.getMat_Ex_T9(), cParameter.getMat_Trans_T9()
        ) ) { noInitSUBMIX; return NO_INIT_SUBMIX_D; }
      }

      return NORMAL_END;
    }



    int convOneFrames(unsigned int fnum) {
      #ifdef PTRACE
        printf("TRACE: view_synthesis::view_subsep_options::convOneFrames()\n");
      #endif
      //printf("[%d]", fnum);

      if ( numView == VIEW_2 ) {
        if (!cSubV.ReadOneFrame( f_V, f_BindR, f_BindD, fnum, f_NFTfile )) return DATA_SET_ERR;
        cout << ".";
        if (!cSubV.NoMix(f_T1, f_T2, f_T3, f_T4, NULL, NULL, NULL, NULL)) return DATA_OUT_ERR; // DEBUG OUT
        cout << ".";

      } else {
        if (!cSubV.ReadOneFrame( f_V, f_BindR, f_BindD, fnum, f_NFTfile )) return DATA_SET_ERR;
        cout << ".";
        if (!cSubV.NoMix(f_T1, f_T2, f_T3, f_T4, f_T6, f_T7, f_T8, f_T9)) return DATA_OUT_ERR; // DEBUG OUT
        cout << ".";

      }

      return NORMAL_END;
    }

  }

}
