// View Synthesis Decoder
// Dev Enviroment:
//    Windows 8.1 64bit with Visual Studio 2013(VS12) C++ x64
//    OpenCV 2.4.9 64bit
// Original code: View Synthesis Reference Software (VSRS) 3.5
// Last Update: June, 2014.
#pragma once
#ifdef WIN32
#pragma warning(disable:4996)
#pragma warning(disable:4244)
#pragma warning(disable:4819)
#endif

#include <iostream>
#include <fstream>
#include "yuv.h"
#include <string>

#define LEFT 1
#define RIGHT 2
#define CENTER 3
#define NOP 100

#define INIT 1
#define RESET 2

#define NARROW 0
#define WIDE 1

#define ONESET 24

#define C1 0
#define C2 1
#define C3 2
#define C4 3
#define C6 4
#define C7 5
#define C8 6
#define C9 7

class CViewInterpolation
{
public:
  CViewInterpolation ();
  virtual ~CViewInterpolation ();

  bool InitOneSet (
    unsigned int width, unsigned int height, bool variparam,
    double near1, double far1, double near2, double far2,
    const char *RefCamID1, const char *RefCamID2, const char *VirCamID, const char *TarCamID,
    const unsigned int Side,
    double Mat_In_1[9], double Mat_Ex_1[9], double Mat_Trans_1[3],
    double Mat_In_2[9], double Mat_Ex_2[9], double Mat_Trans_2[3],
    double Mat_In_V[9], double Mat_Ex_V[9], double Mat_Trans_V[3],
    double Mat_In_T[9], double Mat_Ex_T[9], double Mat_Trans_T[3]
  );

  bool InitOneCh (
    unsigned int width, unsigned int height, bool variparam,
    // double disp255,
    double near, double far,
    const char *RefCamID, const char *VirCamID,
    const unsigned int Side,
    double Mat_In_Ref[9], double Mat_Ex_Ref[9], double Mat_Trans_Ref[3],
    double Mat_In_Vir[9], double Mat_Ex_Vir[9], double Mat_Trans_Vir[3]
  );

  void xReleaseMemory ();

  double getBaselineDistance () { return m_dBaselineDistance; }
  double getLeftBaselineDistance () { return LeftBaselineDistance; }
  double getRightBaselineDistance () { return RightBaselineDistance; }
  IplImage* getHolePixels () { return m_imgHoles; }
  IplImage* getSynthesizedPixels () { return m_imgSuccessSynthesis; }
  IplImage* getVirtualImage () { return m_imgVirtualImage; }
  IplImage* getVirtualDepthMap () { return m_imgVirtualDepth; }
  IplImage* getWorkDepthMap () { return m_imgWorkDepth; }
  BYTE** getVirtualDepth () { return m_pVirtualDepth; }
  BYTE** getWorkDepth () { return m_pWorkDepth; }
  BYTE** getNonHoles () { return m_pSuccessSynthesis; }

  void SetWidth (int sWidth ) { Width  = sWidth;  }
  void SetHeight(int sHeight) { Height = sHeight; }
  void SetDepthType(int sDepthType) {DepthType = sDepthType; }

  void depthInpaint( IplImage* map, IplImage* hole );

  unsigned char weighted2( unsigned char src1, unsigned char src2, unsigned int steps, unsigned int count );
  unsigned char weighted3( unsigned char src1, unsigned char src2, unsigned char src3, unsigned int steps, unsigned int count );
  unsigned int simple2( unsigned char src1, unsigned char src2 );
  unsigned int simple3( unsigned char src1, unsigned char src2, unsigned char src3 );
  void commonholeInpaint( IplImage* pView, IplImage* pHole );

  int  DoOneFrame(
          CIYuv* pSynYuvBuffer,
          double near_T1, double far_T1, double trans_T1
  );

  std::string getRefCameraName () { return RefCameraName; };
  std::string getVirtualCameraName () { return VirtualCameraName; };

  BYTE*** ViewM= NULL;
  BYTE*** ViewR= NULL;
  BYTE** Depth= NULL;
  // CIYuv* pSynYuvBuffer= NULL;
  double near_C1= 0.0;
  double far_C1= 0.0;
  double trans_C1= 0.0;
  double near_C2= 0.0;
  double far_C2= 0.0;
  double trans_C2= 0.0;
  double near_V1= 0.0;
  double far_V1= 0.0;
  double trans_V1= 0.0;
  // double near_T1= 0.0;
  // double far_T1= 0.0;
  // double trans_T1= 0.0;
  unsigned int fnum= 0;
  unsigned int from= 0;
  bool debug=false;

protected:

private:
  void convertCameraParam (CvMat *exMat_dst, CvMat *exMat_src);

  CViewInterpolation*  m_pD;
  CViewInterpolation*  m_pS;
  CViewInterpolation*  m_pV;
  CViewInterpolation*  m_pR;


  bool init_camera_param (
            double Mat_In_Ref[9], double Mat_Ex_Ref[9], double Mat_Trans_Ref[3],
            double Mat_In_Vir[9], double Mat_Ex_Vir[9], double Mat_Trans_Vir[3],
            CvMat *mat_in[2], CvMat *mat_ex_c2w[2], CvMat *mat_proj[2]);

  bool init_3Dwarp( double Z_near, double Z_far );
  bool reset_3Dwarp (
            double near_C, double far_C, double trans_C,
            double near_V, double far_V, double trans_V );

  void image2world_with_z (
            CvMat *mat_Rc2w_invIN_from, CvMat *matEX_c2w_from,
            CvMat *image, CvMat *world);

  void makeHomography (
            CvMat *matH_F2T[256],
            CvMat *matH_T2F[256],
            double adTable[256],
            CvMat *matIN_from, CvMat *matEX_c2w_from, CvMat *matProj_to,
            unsigned int work= RESET );

  bool setDepth(BYTE **pDepthMap);
  bool setView(BYTE ***src);
  bool depthsynthesis_reverse_3Dwarp (BYTE **pDepthMap);
  bool depthShift( BYTE **pDepthMap, unsigned int from );

  bool viewsynthesis_reverse_3Dwarp(BYTE ***src, unsigned int from, IplImage* pHole= NULL );

  void looseHoledge( IplImage* pView, IplImage* pHole );
  void drawHoledge( IplImage* pView, IplImage* pHole, char Y, char U, char V );

  double m_dWeightLeft;
  double m_dWeightRight;
  double WeightLeft;
  double WeightRight;

  unsigned int m_uiWidth;
  unsigned int m_uiHeight;
  unsigned int m_uiPicsize;
  unsigned int m_uiPrecision;
  const static unsigned int m_uiDepthType= 0;

  unsigned int m_uiResidualsSide;
  unsigned int m_uiResidualsWidth;
  double m_dMaxDisp;
  double m_dispA;
  double m_dispB;
  double m_disp255;
  double m_weightVR;

  std::string CameraParameterFile;
  std::string RefCameraName;
  std::string VirtualCameraName;

  unsigned int Width;
  unsigned int Height;
  unsigned int Picsize;
  unsigned int Precision;
  unsigned int DepthType;
  unsigned int Side;
  bool VariParam;

  // for VeriParam
  double Mat_Ex_C[9];
  double Mat_Ex_V[9];
  double Mat_In_C[9];
  double Mat_In_V[9];
  double Mat_Trans_C[3];
  double Mat_Trans_V[3];

  double m_dBaselineDistance;
  double LeftBaselineDistance;
  double RightBaselineDistance;

  unsigned int m_ResidualsSide;

  BYTE** m_pVirtualDepth;
  BYTE** m_pWorkDepth;
  BYTE** m_pSuccessSynthesis;

  IplImage* m_imgVirtualImage;
  IplImage* m_imgVirtualDepth;
  IplImage* m_imgWorkDepth;
  IplImage* m_imgSuccessSynthesis;
  IplImage* m_imgHoles;

  static const unsigned int m_MaskTh= 1;

  IplImage* m_depthInterpolatedView;
  void ToFilterDepth(IplImage* pDepthSRC, IplImage* pDepthDST, IplImage* pDepthNEGSRC, IplImage* pDepthNEGDST, unsigned int fsize);


  // for VeriParam
  CvMat* m_matH_R2V[256];
  CvMat* m_matH_V2R[256];
  CvMat* m_Mat_in[2]; //intrinsic parameter of left camera 3x3 matrix
  CvMat* m_Mat_ex_c2w[2]; //extrinsic parameter of left camera 3x4 matrix
  CvMat* m_Mat_proj_w2i[2]; // projection matrix from world to image
};
