// NICT GVD: View Synthesis Decoder
// Dev Enviroment:
//    Windows 8.1 64bit with Visual Studio 2013(VS12) C++ x64
//    OpenCV 2.4.9 64bit
// Original code: View Synthesis Reference Software (VSRS) 3.5
// Designer: Senoh, NiCT, Tokyo. Japan.
// Last Update: June, 2014.
#define GVD_MAJOR_VERSION 4
#define GVD_MINOR_VERSION 4
#define GVD_VERSION CVAUX_STR(GVD_MAJOR_VERSION) "." CVAUX_STR(GVD_MINOR_VERSION)


#ifdef WIN32
#define _CRTDBG_MAP_ALLOC
#include <crtdbg.h>
static void MemReport_Init() {
  _CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF ); //| _CRTDBG_DELAY_FREE_MEM_DF );
  _CrtSetReportMode(_CRT_ERROR, _CRTDBG_MODE_FILE); // _FILE _DEBUG
  _CrtSetReportFile(_CRT_ERROR, _CRTDBG_FILE_STDOUT);
}
  //_CRT_WARN _CRT_ERROR
/*
#ifdef _DEBUG
   #ifndef DBG_NEW
      #define DBG_NEW new ( _NORMAL_BLOCK , __FILE__ , __LINE__ )
      #define new DBG_NEW
   #endif
#endif
*/
#endif

#include <opencv2/opencv.hpp>
#include <stdlib.h>
#include <time.h>

#include "yuv.h"

#include "ParameterDec.h"
#include "SubSep.h"
#include "ReturnCode.h"
#include "SubSepOptions.hpp"

namespace option= view_synthesis::view_subsep_options;

using namespace std;

int main(int argc , char *argv[])
{
  // MemReport_Init();
  cv::setUseOptimized(true);

  int rc;

  clock_t start, finish, first;
  first = start = clock();

  cout << "GVD Decoder, Version " << GVD_VERSION << " - MPEG 3DV, June 2014 NiCT." << endl << endl;

  if ( NORMAL_END != option::cParameter.Init( argc, argv ) )
    return BAD_COMAMND_LINE;
  if ( UNKNOWN == option::checkOption(option::cParameter.getMode()) )
    return BAD_CONFIG_OPTIONS;

  rc= option::initParam(); if ( rc != NORMAL_END ) return rc;
  rc= option::initFile(); if ( rc != NORMAL_END ) return rc;
  rc= option::initSub(); if ( rc != NORMAL_END ) { option::endFile(); return rc; }

  finish= clock();
  printf( "\nInitialization: %.4f sec\n", (double)(finish - start) / CLOCKS_PER_SEC);
  start= finish;

  #ifdef PTRACE
    printf("TRACE: main() LOOP start.\n");
  #endif
  for(unsigned int fnum= option::start; fnum < option::start + option::total; fnum++) {
    cout << "frame number = " << fnum;
    if (option::convOneFrames(fnum) != NORMAL_END) {genABEND; break;}
    finish= clock();
    printf("->End (%.4f sec)\n", (double)(finish - start) / CLOCKS_PER_SEC);
    start= finish;

    //_CrtDumpMemoryLeaks();
  }
  #ifdef PTRACE
    printf("TRACE: main() LOOP end.\n");
  #endif

  finish = clock();
  printf("Total: %.4f sec\n", ((double)(finish-first))/((double)CLOCKS_PER_SEC));

  option::endFile();

  return 0;
}


