// View Synthesis Encoder for Residuals
// Dev Enviroment:
//    Windows 8.1 64bit with Visual Studio 2013(VS12) C++ x64
//    OpenCV 2.4.9 64bit
// Original code: View Synthesis Reference Software (VSRS) 3.5
// Designer: Senoh, NiCT, Tokyo. Japan.
// Last Update: June, 2014.

#ifdef WIN32
#pragma warning(disable : 4996)
#endif

#include "yuv.h"
#include "SubMixOptions.h"
#include "SubSep.h"
#include "ViewSynthesis.h"
#include <omp.h>

CSubSep::CSubSep() {
  #ifdef PTRACE
    printf("TRACE: CSubSep::CSubSep()\n");
  #endif

  pYuvRBind= NULL;
  pYuvR1= NULL;
  pYuvR2= NULL;
  pYuvR1_temp= NULL;
  pYuvR2_temp= NULL;

  pYuvDBind= NULL;
  pYuvDBindPRE= NULL;
  pYuvD1= NULL;
  pYuvD2= NULL;
  pYuvD1_temp= NULL;
  pYuvD2_temp= NULL;

  pYuvR1_half= NULL;
  pYuvR2_half= NULL;
  pYuvD1_half= NULL;
  pYuvD2_half= NULL;

  pYuvV= NULL;
  pYuvT[C1]= NULL;
  pYuvT[C2]= NULL;
  pYuvT[C3]= NULL;
  pYuvT[C4]= NULL;
  pYuvT[C6]= NULL;
  pYuvT[C7]= NULL;
  pYuvT[C8]= NULL;
  pYuvT[C9]= NULL;

  pYuvV_work= NULL;
  pYuvV_temp= NULL;

  pYuvRBind_work= NULL;
  pYuvDBind_work= NULL;

  m_img = NULL;
}



CSubSep::~CSubSep() {
  #ifdef PTRACE
    printf("TRACE: CSubSep::~CSubSep()\n");
  #endif

  if (pYuvRBind != NULL) { delete pYuvRBind; pYuvRBind= NULL; }
  if (pYuvRBind_work != NULL) { delete pYuvRBind_work; pYuvRBind_work= NULL; }
  if (pYuvR1 != NULL) { delete pYuvR1; pYuvR1= NULL; }
  if (pYuvR2 != NULL) { delete pYuvR2; pYuvR2= NULL; }
  if (pYuvR1_temp != NULL) { delete pYuvR1_temp; pYuvR1_temp= NULL; }
  if (pYuvR2_temp != NULL) { delete pYuvR2_temp; pYuvR2_temp= NULL; }

  if (pYuvR1_half != NULL) { delete pYuvR1_half; pYuvR1_half= NULL; }
  if (pYuvR2_half != NULL) { delete pYuvR2_half; pYuvR2_half= NULL; }
  if (pYuvD1_half != NULL) { delete pYuvD1_half; pYuvD1_half= NULL; }
  if (pYuvD2_half != NULL) { delete pYuvD2_half; pYuvD2_half= NULL; }

  if (pYuvV_work != NULL) { delete pYuvV_work; pYuvV_work= NULL; }
  if (pYuvV_temp != NULL) { delete pYuvV_temp; pYuvV_temp= NULL; }

  if (pYuvDBind != NULL) { delete pYuvDBind; pYuvDBind= NULL; }
  if (pYuvDBindPRE != NULL) { delete pYuvDBindPRE; pYuvDBindPRE= NULL; }
  if (pYuvDBind_work != NULL) { delete pYuvDBind_work; pYuvDBind_work= NULL; }
  if (pYuvD1 != NULL) { delete pYuvD1; pYuvD1= NULL; }
  if (pYuvD2 != NULL) { delete pYuvD2; pYuvD2= NULL; }
  if (pYuvD1_temp != NULL) { delete pYuvD1_temp; pYuvD1_temp= NULL; }
  if (pYuvD2_temp != NULL) { delete pYuvD2_temp; pYuvD2_temp= NULL; }

  if (pYuvV != NULL) { delete pYuvV; pYuvV= NULL; }
  if (pYuvT[C1] != NULL) { delete pYuvT[C1]; pYuvT[C1]= NULL; }
  if (pYuvT[C2] != NULL) { delete pYuvT[C2]; pYuvT[C2]= NULL; }
  if (pYuvT[C3] != NULL) { delete pYuvT[C3]; pYuvT[C3]= NULL; }
  if (pYuvT[C4] != NULL) { delete pYuvT[C4]; pYuvT[C4]= NULL; }
  if (pYuvT[C6] != NULL) { delete pYuvT[C6]; pYuvT[C6]= NULL; }
  if (pYuvT[C7] != NULL) { delete pYuvT[C7]; pYuvT[C7]= NULL; }
  if (pYuvT[C8] != NULL) { delete pYuvT[C8]; pYuvT[C8]= NULL; }
  if (pYuvT[C9] != NULL) { delete pYuvT[C9]; pYuvT[C9]= NULL; }

  if (m_img != NULL) { cvReleaseImage(&m_img); m_img= NULL; }
}



bool CSubSep::Init(
      unsigned int mode, unsigned int numView, bool variparam,
      unsigned int width, unsigned int height, unsigned int width_bind, unsigned int height_bind,
      double near1, double far1, double near2, double far2, double near3, double far3,
      const char *RefCamID1, const char *RefCamID2, const char *RefCamID3,
      const char *VirCamID1, const char *VirCamID2,
      const char *TarCamID1, const char *TarCamID2, const char *TarCamID3, const char *TarCamID4,
      const char *TarCamID6, const char *TarCamID7, const char *TarCamID8, const char *TarCamID9,
      double Mat_In_1[9], double Mat_Ex_1[9], double Mat_Trans_1[3],
      double Mat_In_2[9], double Mat_Ex_2[9], double Mat_Trans_2[3],
      double Mat_In_3[9], double Mat_Ex_3[9], double Mat_Trans_3[3],
      double Mat_In_V1[9], double Mat_Ex_V1[9], double Mat_Trans_V1[3],
      double Mat_In_V2[9], double Mat_Ex_V2[9], double Mat_Trans_V2[3],
      double Mat_In_T1[9], double Mat_Ex_T1[9], double Mat_Trans_T1[3],
      double Mat_In_T2[9], double Mat_Ex_T2[9], double Mat_Trans_T2[3],
      double Mat_In_T3[9], double Mat_Ex_T3[9], double Mat_Trans_T3[3],
      double Mat_In_T4[9], double Mat_Ex_T4[9], double Mat_Trans_T4[3],
      double Mat_In_T6[9], double Mat_Ex_T6[9], double Mat_Trans_T6[3],
      double Mat_In_T7[9], double Mat_Ex_T7[9], double Mat_Trans_T7[3],
      double Mat_In_T8[9], double Mat_Ex_T8[9], double Mat_Trans_T8[3],
      double Mat_In_T9[9], double Mat_Ex_T9[9], double Mat_Trans_T9[3]
) {
  #ifdef PTRACE
    printf("TRACE: CSubSep::Init()\n");
  #endif

  m_mode= mode;
  m_width= width;
  m_height= height;
  m_width_bind= width_bind;
  m_height_bind= height_bind;

  m_numView= numView;
  if ( mode == QUARTER_STACK || mode == HALF_QUARTER_STACK ) m_numDepth= DEPTH_2; else m_numDepth= DEPTH_1;
  m_globalDepth= ( mode == QUARTER_ODD || mode == QUARTER_ODD_L || mode == FULL_ODD_L );
  m_fullSize= ( mode == QUARTER_ODD || mode == QUARTER_ODD_L || mode == FULL_ODD_L );

  pYuvV= new CIYuv(height, width, 420);

  pYuvR1= new CIYuv(height, width, 444);
  pYuvD1= new CIYuv(height, width, 444);
  pYuvT[C1]= new CIYuv(height, width, 420);
  pYuvT[C2]= new CIYuv(height, width, 420);
  pYuvT[C3]= new CIYuv(height, width, 420);
  pYuvT[C4]= new CIYuv(height, width, 420);
  if ( m_numView == VIEW_3 ) {
    pYuvR2= new CIYuv(height, width, 444);
    if ( m_numDepth == DEPTH_2  )
      pYuvD2= new CIYuv(height, width, 444);
    pYuvT[C6]= new CIYuv(height, width, 420);
    pYuvT[C7]= new CIYuv(height, width, 420);
    pYuvT[C8]= new CIYuv(height, width, 420);
    pYuvT[C9]= new CIYuv(height, width, 420);
  }

  if ( m_mode == QUARTER ) {
    pYuvRBind= new CIYuv(height_bind, width_bind, 420);
    pYuvDBind= new CIYuv(height_bind, width_bind, 420);
    pYuvDBindPRE= new CIYuv(height_bind, width_bind, 420);
    pYuvRBind_work= new CIYuv(height_bind, width_bind, 444);
    pYuvDBind_work= new CIYuv(height_bind, width_bind, 444);
    pYuvR1_temp= new CIYuv(height, width, 444);
    pYuvD1_temp= new CIYuv(height, width, 444);
  } else if ( m_mode == QUARTER_L ) {
    pYuvRBind= new CIYuv(height, width, 420);
    pYuvDBind= new CIYuv(height_bind, width_bind, 420);
    pYuvDBindPRE= new CIYuv(height_bind, width_bind, 420);
    pYuvRBind_work= new CIYuv(height_bind, width_bind, 444);
    pYuvDBind_work= new CIYuv(height_bind, width_bind, 444);
    pYuvR1_temp= new CIYuv(height, width, 444);
    pYuvD1_temp= new CIYuv(height, width, 444);
  } else { // QUARTER_STACK
    if ( m_mode == QUARTER_STACK ) {
      pYuvRBind= new CIYuv(height_bind, width_bind, 420);
      pYuvDBind= new CIYuv(height_bind, width_bind, 420);
      pYuvDBindPRE= new CIYuv(height_bind, width_bind, 420);
      pYuvRBind_work= new CIYuv(height_bind, width_bind, 444);
      pYuvDBind_work= new CIYuv(height_bind, width_bind, 444);
    }
    if ( m_mode == HALF_QUARTER_STACK ) {
      pYuvRBind= new CIYuv(height_bind / 2, width_bind, 420);
      pYuvDBind= new CIYuv(height_bind / 2, width_bind, 420);
      pYuvRBind_work= new CIYuv(height_bind / 2, width_bind, 444);
      pYuvDBind_work= new CIYuv(height_bind / 2, width_bind, 444);
    }
    if ( m_mode == QUARTER_ODD ) {
      pYuvRBind= new CIYuv(height_bind, width_bind, 420);
      pYuvDBind= new CIYuv(height_bind / 2, width_bind, 420);
      pYuvDBindPRE= new CIYuv(height_bind / 2, width_bind, 420);
      pYuvRBind_work= new CIYuv(height_bind, width_bind, 444);
      pYuvDBind_work= new CIYuv(height_bind / 2, width_bind, 444);
    }
    if ( m_mode == QUARTER_ODD_L ) {
      pYuvRBind= new CIYuv(height_bind, width, 420); // Full width
      pYuvDBind= new CIYuv(height_bind / 2, width_bind, 420);
      pYuvDBindPRE= new CIYuv(height_bind / 2, width_bind, 420);
      pYuvRBind_work= new CIYuv(height_bind, width_bind, 444);
      pYuvDBind_work= new CIYuv(height_bind / 2, width_bind, 444);
    }
    if ( m_mode == FULL_ODD_L ) {
      pYuvRBind= new CIYuv(height_bind, width, 420); // Full width
      pYuvDBind= new CIYuv(height_bind, width, 420); // DUMMY
      pYuvDBindPRE= new CIYuv(height_bind, width, 420);
      pYuvRBind_work= new CIYuv(height_bind, width_bind, 444);
      pYuvDBind_work= new CIYuv(height_bind, width, 444); // DUMMY
    }
    pYuvR1_temp= new CIYuv(height_bind, width, 444);
    pYuvR2_temp= new CIYuv(height_bind, width, 444);
    pYuvD1_temp= new CIYuv(height_bind, width, 444);
    if ( m_mode != QUARTER_ODD && m_mode != QUARTER_ODD_L&& m_mode != FULL_ODD_L  )
      pYuvD2_temp= new CIYuv(height_bind, width, 444);
  }

  if (
    pYuvR1 == NULL || pYuvRBind == NULL || pYuvT[C1] == NULL || pYuvV == NULL || pYuvD1 == NULL || pYuvDBind == NULL || pYuvDBindPRE == NULL
    || ( m_numDepth == DEPTH_2 && pYuvD2==NULL )
    || ( m_numView == VIEW_3 && ( pYuvR2==NULL || pYuvRBind_work == NULL || pYuvDBind_work == NULL
                                || pYuvT[C2] == NULL || pYuvT[C3] == NULL || pYuvT[C4] == NULL
                                || pYuvT[C6] == NULL || pYuvT[C7] == NULL || pYuvT[C8] == NULL || pYuvT[C9] == NULL ) )
  ) {
    if (pYuvR1 != NULL) delete pYuvR1;
    if (pYuvR2 != NULL) delete pYuvR2;
    if (pYuvRBind != NULL) delete pYuvRBind;
    if (pYuvRBind_work != NULL) delete pYuvRBind_work;
    if (pYuvD1 != NULL) delete pYuvD1;
    if (pYuvD2 != NULL) delete pYuvD2;
    if (pYuvDBind != NULL) delete pYuvDBind;
    if (pYuvDBindPRE != NULL) delete pYuvDBindPRE;
    if (pYuvDBind_work != NULL) delete pYuvDBind_work;
    if (pYuvT[C1] != NULL) delete pYuvT[C1];
    if (pYuvT[C2] != NULL) delete pYuvT[C2];
    if (pYuvT[C3] != NULL) delete pYuvT[C3];
    if (pYuvT[C4] != NULL) delete pYuvT[C4];
    if (pYuvT[C6] != NULL) delete pYuvT[C6];
    if (pYuvT[C7] != NULL) delete pYuvT[C7];
    if (pYuvT[C8] != NULL) delete pYuvT[C8];
    if (pYuvT[C9] != NULL) delete pYuvT[C9];
    return false;
  }

  pYuvV_work= new CIYuv(height, width, 444);
  pYuvV_temp= new CIYuv(height, width, 444);
  pYuvV_temp->setUpsampleFilter(1, 1); // General, QUARTER PEL Dec30,2011 // 1 Full-pel

  if ( m_numView == VIEW_3  ) {
    if ( m_mode == QUARTER_STACK ) {
      pYuvD1_half= new CIYuv(m_height_bind, width, 444);
      pYuvD2_half= new CIYuv(m_height_bind, width, 444);
      pYuvR1_half= new CIYuv(m_height_bind, width, 444);
      pYuvR2_half= new CIYuv(m_height_bind, width, 444);
    }
    if ( m_mode == HALF_QUARTER_STACK ) {
      pYuvD1_half= new CIYuv(m_height_bind / 2, width, 444);
      pYuvD2_half= new CIYuv(m_height_bind / 2, width, 444);
      pYuvR1_half= new CIYuv(m_height_bind / 2, width, 444);
      pYuvR2_half= new CIYuv(m_height_bind / 2, width, 444);
    }
    if ( m_numDepth == DEPTH_1 ) {
      pYuvD1_half= new CIYuv(m_height_bind, width, 444);
      pYuvD2_half= NULL;
      pYuvR1_half= new CIYuv(m_height_bind, width, 444);
      pYuvR2_half= new CIYuv(m_height_bind, width, 444);
    }
  }

  if ( pYuvV_work == NULL || pYuvV_temp == NULL
      || ( ( m_mode == QUARTER_STACK || m_mode == HALF_QUARTER_STACK )
          && ( pYuvD1_half == NULL || pYuvD2_half == NULL || pYuvR1_half == NULL || pYuvR2_half == NULL) )
      || ( ( m_mode == QUARTER_ODD || m_mode == QUARTER_ODD_L || m_mode == FULL_ODD_L )
          && ( pYuvD1_half == NULL || pYuvR1_half == NULL || pYuvR2_half == NULL) )
  ) {
    if (pYuvV_work != NULL) delete pYuvV_work;
    if (pYuvV_temp != NULL) delete pYuvV_temp;
    if (pYuvDBind_work != NULL) delete pYuvDBind_work;

    if (pYuvR1_half != NULL) delete pYuvR1_half;
    if (pYuvR2_half != NULL) delete pYuvR2_half;
    if (pYuvD1_half != NULL) delete pYuvD1_half;
    if (pYuvD2_half != NULL) delete pYuvD2_half;

    return false;
  }

  m_img= cvCreateImage(cvSize(width_bind, height_bind), 8, 3);

  if ( m_mode == QUARTER_ODD || m_mode == QUARTER_ODD_L || m_mode == FULL_ODD_L ) {

    // m_pViewSyn1
    m_pViewSyn[C1]= new CViewInterpolation();
    m_pViewSyn[C1]->SetWidth( width );
    m_pViewSyn[C1]->SetHeight( height );
    if (
      !m_pViewSyn[C1]->InitOneSet(
                        width, height, variparam,
                        near1, far1, near2, far2,
                        RefCamID1, RefCamID2, RefCamID2, TarCamID1,
                        LEFT,
                        Mat_In_1, Mat_Ex_1, Mat_Trans_1,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_T1, Mat_Ex_T1, Mat_Trans_T1
                        )
    ) return false;

    // m_pViewSyn2
    m_pViewSyn[C2]= new CViewInterpolation();
    m_pViewSyn[C2]->SetWidth( width );
    m_pViewSyn[C2]->SetHeight( height );
    if (
      !m_pViewSyn[C2]->InitOneSet(
                        width, height, variparam,
                        near1, far1, near2, far2,
                        RefCamID1, RefCamID2, RefCamID2, TarCamID2,
                        LEFT,
                        Mat_In_1, Mat_Ex_1, Mat_Trans_1,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_T2, Mat_Ex_T2, Mat_Trans_T2
                        )
    ) return false;

    // m_pViewSyn3
    m_pViewSyn[C3]= new CViewInterpolation();
    m_pViewSyn[C3]->SetWidth( width );
    m_pViewSyn[C3]->SetHeight( height );
    if (
      !m_pViewSyn[C3]->InitOneSet(
                        width, height, variparam,
                        near1, far1, near2, far2,
                        RefCamID1, RefCamID2, RefCamID2, TarCamID3,
                        LEFT,
                        Mat_In_1, Mat_Ex_1, Mat_Trans_1,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_T3, Mat_Ex_T3, Mat_Trans_T3
                        )
    ) return false;

    // m_pViewSyn4
    m_pViewSyn[C4]= new CViewInterpolation();
    m_pViewSyn[C4]->SetWidth( width );
    m_pViewSyn[C4]->SetHeight( height );
    if (
      !m_pViewSyn[C4]->InitOneSet(
                        width, height, variparam,
                        near1, far1, near2, far2,
                        RefCamID1, RefCamID2, RefCamID2, TarCamID4,
                        LEFT,
                        Mat_In_1, Mat_Ex_1, Mat_Trans_1,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_T4, Mat_Ex_T4, Mat_Trans_T4
                        )
    ) return false;

    // m_pViewSyn6
    m_pViewSyn[C6]= new CViewInterpolation();
    m_pViewSyn[C6]->SetWidth( width );
    m_pViewSyn[C6]->SetHeight( height );
    if (
      !m_pViewSyn[C6]->InitOneSet( // CameraOrder: Outer(Right) to Center !!!!!!
                      width, height, variparam,
                      near3, far3, near2, far2,
                      RefCamID3, RefCamID2, RefCamID2, TarCamID6,
                      RIGHT,
                      Mat_In_3, Mat_Ex_3, Mat_Trans_3,
                      Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                      Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                      Mat_In_T6, Mat_Ex_T6, Mat_Trans_T6
                      )
    ) return false;

    // m_pViewSyn7
    m_pViewSyn[C7]= new CViewInterpolation();
    m_pViewSyn[C7]->SetWidth( width );
    m_pViewSyn[C7]->SetHeight( height );
    if (
      !m_pViewSyn[C7]->InitOneSet( // CameraOrder: Outer(Right) to Center !!!!!!
                      width, height, variparam,
                      near3, far3, near2, far2,
                      RefCamID3, RefCamID2, RefCamID2, TarCamID7,
                      RIGHT,
                      Mat_In_3, Mat_Ex_3, Mat_Trans_3,
                      Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                      Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                      Mat_In_T7, Mat_Ex_T7, Mat_Trans_T7
                      )
    ) return false;

    // m_pViewSyn8
    m_pViewSyn[C8]= new CViewInterpolation();
    m_pViewSyn[C8]->SetWidth( width );
    m_pViewSyn[C8]->SetHeight( height );
    if (
      !m_pViewSyn[C8]->InitOneSet( // CameraOrder: Outer(Right) to Center !!!!!!
                      width, height, variparam,
                      near3, far3, near2, far2,
                      RefCamID3, RefCamID2, RefCamID2, TarCamID8,
                      RIGHT,
                      Mat_In_3, Mat_Ex_3, Mat_Trans_3,
                      Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                      Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                      Mat_In_T8, Mat_Ex_T8, Mat_Trans_T8
                      )
    ) return false;

    // m_pViewSyn9
    m_pViewSyn[C9]= new CViewInterpolation();
    m_pViewSyn[C9]->SetWidth( width );
    m_pViewSyn[C9]->SetHeight( height );
    if (
      !m_pViewSyn[C9]->InitOneSet( // CameraOrder: Outer(Right) to Center !!!!!!
                      width, height, variparam,
                      near3, far3, near2, far2,
                      RefCamID3, RefCamID2, RefCamID2, TarCamID9,
                      RIGHT,
                      Mat_In_3, Mat_Ex_3, Mat_Trans_3,
                      Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                      Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                      Mat_In_T9, Mat_Ex_T9, Mat_Trans_T9
                      )
    ) return false;

  } else {

    m_pViewSyn[C1]= new CViewInterpolation();
    m_pViewSyn[C1]->SetWidth( width );
    m_pViewSyn[C1]->SetHeight( height );
    if (
      !m_pViewSyn[C1]->InitOneSet(
                        width, height, variparam,
                        near1, far1, near2, far2,
                        RefCamID1, RefCamID2, TarCamID3, TarCamID1,
                        LEFT,
                        Mat_In_1, Mat_Ex_1, Mat_Trans_1,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_T3, Mat_Ex_T3, Mat_Trans_T3,
                        Mat_In_T1, Mat_Ex_T1, Mat_Trans_T1
                        )
    ) return false;

    m_pViewSyn[C2]= new CViewInterpolation();
    m_pViewSyn[C2]->SetWidth( width );
    m_pViewSyn[C2]->SetHeight( height );
    if (
      !m_pViewSyn[C2]->InitOneSet(
                        width, height, variparam,
                        near1, far1, near2, far2,
                        RefCamID1, RefCamID2, TarCamID3, TarCamID2,
                        LEFT,
                        Mat_In_1, Mat_Ex_1, Mat_Trans_1,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_T3, Mat_Ex_T3, Mat_Trans_T3,
                        Mat_In_T2, Mat_Ex_T2, Mat_Trans_T2
                        )
    ) return false;

    m_pViewSyn[C3]= new CViewInterpolation();
    m_pViewSyn[C3]->SetWidth( width );
    m_pViewSyn[C3]->SetHeight( height );
    if (
      !m_pViewSyn[C3]->InitOneSet(
                        width, height, variparam,
                        near1, far1, near2, far2,
                        RefCamID1, RefCamID2, TarCamID3, TarCamID3,
                        LEFT,
                        Mat_In_1, Mat_Ex_1, Mat_Trans_1,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_T3, Mat_Ex_T3, Mat_Trans_T3,
                        Mat_In_T3, Mat_Ex_T3, Mat_Trans_T3
                        )
    ) return false;

    m_pViewSyn[C4]= new CViewInterpolation();
    m_pViewSyn[C4]->SetWidth( width );
    m_pViewSyn[C4]->SetHeight( height );
    if (
      !m_pViewSyn[C4]->InitOneSet(
                        width, height, variparam,
                        near1, far1, near2, far2,
                        RefCamID1, RefCamID2, TarCamID3, TarCamID4,
                        LEFT,
                        Mat_In_1, Mat_Ex_1, Mat_Trans_1,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_T3, Mat_Ex_T3, Mat_Trans_T3,
                        Mat_In_T4, Mat_Ex_T4, Mat_Trans_T4
                        )
    ) return false;

    if ( m_numView == VIEW_3 ) {
      m_pViewSyn[C6]= new CViewInterpolation();
      m_pViewSyn[C6]->SetWidth( width );
      m_pViewSyn[C6]->SetHeight( height );
      if (
        !m_pViewSyn[C6]->InitOneSet( // CameraOrder: Outer(Right) to Center !!!!!!
                        width, height, variparam,
                        near3, far3, near2, far2,
                        RefCamID3, RefCamID2, TarCamID7, TarCamID6,
                        RIGHT,
                        Mat_In_3, Mat_Ex_3, Mat_Trans_3,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_T7, Mat_Ex_T7, Mat_Trans_T7,
                        Mat_In_T6, Mat_Ex_T6, Mat_Trans_T6
                        )
      ) return false;

      m_pViewSyn[C7]= new CViewInterpolation();
      m_pViewSyn[C7]->SetWidth( width );
      m_pViewSyn[C7]->SetHeight( height );
      if (
        !m_pViewSyn[C7]->InitOneSet( // CameraOrder: Outer(Right) to Center !!!!!!
                        width, height, variparam,
                        near3, far3, near2, far2,
                        RefCamID3, RefCamID2, TarCamID7, TarCamID7,
                        RIGHT,
                        Mat_In_3, Mat_Ex_3, Mat_Trans_3,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_T7, Mat_Ex_T7, Mat_Trans_T7,
                        Mat_In_T7, Mat_Ex_T7, Mat_Trans_T7
                        )
      ) return false;

      m_pViewSyn[C8]= new CViewInterpolation();
      m_pViewSyn[C8]->SetWidth( width );
      m_pViewSyn[C8]->SetHeight( height );
      if (
        !m_pViewSyn[C8]->InitOneSet( // CameraOrder: Outer(Right) to Center !!!!!!
                        width, height, variparam,
                        near3, far3, near2, far2,
                        RefCamID3, RefCamID2, TarCamID7, TarCamID8,
                        RIGHT,
                        Mat_In_3, Mat_Ex_3, Mat_Trans_3,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_T7, Mat_Ex_T7, Mat_Trans_T7,
                        Mat_In_T8, Mat_Ex_T8, Mat_Trans_T8
                        )
      ) return false;

      m_pViewSyn[C9]= new CViewInterpolation();
      m_pViewSyn[C9]->SetWidth( width );
      m_pViewSyn[C9]->SetHeight( height );
      if (
        !m_pViewSyn[C9]->InitOneSet( // CameraOrder: Outer(Right) to Center !!!!!!
                        width, height, variparam,
                        near3, far3, near2, far2,
                        RefCamID3, RefCamID2, TarCamID7, TarCamID9,
                        RIGHT,
                        Mat_In_3, Mat_Ex_3, Mat_Trans_3,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_T7, Mat_Ex_T7, Mat_Trans_T7,
                        Mat_In_T9, Mat_Ex_T9, Mat_Trans_T9
                        )
      ) return false;
    }

  }


  return true;
}



bool CSubSep::ReadOneFrame( FILE *fpV, FILE *fpR, FILE *fpD, unsigned int fnum, FILE *f_NFTfile ) {
  BYTE*** RefV= pYuvV_temp->getData();
  BYTE*** RefR1= pYuvR1->getData();
  BYTE*** RefR2;  if ( m_numView == VIEW_3 ) RefR2= pYuvR2->getData();
  BYTE** RefD1= pYuvD1->Y;
  BYTE** RefD2;  if (  m_numDepth == DEPTH_2 ) RefD2= pYuvD2->Y;
  double D1near, D1far, D1trans, D2near, D2far, D2trans, D3near, D3far, D3trans;
  double V1near, V1far, V1trans, V2near, V2far, V2trans;
  double Tnear[8], Tfar[8], Ttrans[8];

  #ifdef PTRACE
    printf("TRACE: CSubSep::ReadOneFrame()\n");
  #endif

  try {
    pYuvV->ReadOneFrame( fpV, fnum );
    pYuvV_work->setData444_inIYUV(pYuvV);
    pYuvV_temp->upsampling(pYuvV_work);

    // Residuals Split/Resize
    pYuvRBind->ReadOneFrame( fpR, fnum );
    if ( m_numView == VIEW_2 ) {
      if ( m_mode == QUARTER_L )
        pYuvRBind_work->setData444_inIYUV420_onlyLEFTSIDE( pYuvRBind, true );
      else
        pYuvRBind_work->setData444_inIYUV( pYuvRBind );

      pYuvR1_temp->LPF_V_RSingle( pYuvRBind_work ); // BQF_V_Single // BQF_Single
      pYuvR1->LPF_H_RSmooth( pYuvR1_temp ); // BQF_H

    } else { // 3view
      if ( m_mode == QUARTER_ODD_L || m_mode == FULL_ODD_L )
        pYuvRBind_work->setData444_inIYUV420_onlyLEFTSIDE( pYuvRBind );
      else
        pYuvRBind_work->setData444_inIYUV( pYuvRBind );

      pYuvR1_half->LPF_V_RSplit( pYuvRBind_work, LEFT ); // BQF_V_Split
      pYuvR2_half->LPF_V_RSplit( pYuvRBind_work, RIGHT ); // BQF_V_Split

      if ( m_mode != HALF_QUARTER_STACK ) {
        pYuvR1->LPF_H_RSmooth( pYuvR1_half ); // BQF_H _temp
        pYuvR2->LPF_H_RSmooth( pYuvR2_half ); // BQF_H
      } else { // HALF_QUARTER_STACK
        pYuvR1_temp->LPF_V_RExpand( pYuvR1_half ); // BQF_V_Expand
        pYuvR2_temp->LPF_V_RExpand( pYuvR2_half ); // BQF_V_Expand
        pYuvR1->LPF_H_RSmooth( pYuvR1_temp ); // BQF_H _temp
        pYuvR2->LPF_H_RSmooth( pYuvR2_temp ); // BQF_H
      }

    }

    // DEBUG  Residuals OUT /////////////////////////////////
    // return true;

    // Depth Split/Resize
    pYuvDBindPRE->ReadOneFrame( fpD, fnum );
    if ( m_mode == FULL_ODD_L ) {
      pYuvD1->setData444_inIYUV( pYuvDBindPRE );
    } else {
      pYuvDBind->depthEnhanced( pYuvDBindPRE );
      if ( m_numView == VIEW_2 ) { // for 2view ////////////////////////////////////////////
        pYuvDBind_work->setData444_inIYUV( pYuvDBind );
        pYuvD1_temp->LPF_V_DSingle( pYuvDBind_work, 5, false );
        pYuvD1->LPF_H_DSmooth( pYuvD1_temp, 5, false );

      } else { // for 3view /////////////////////////////////////////////
        pYuvDBind_work->setData444_inIYUV( pYuvDBind );

        if ( m_mode == QUARTER_ODD || m_mode == QUARTER_ODD_L ) {
          pYuvD1_half->LPF_V_DSplit( pYuvDBind_work, CENTER, 5, false );
        } else {
          pYuvD1_half->LPF_V_DSplit( pYuvDBind_work, LEFT, 5, false );
          pYuvD2_half->LPF_V_DSplit( pYuvDBind_work, RIGHT, 5, false );
        }

        if ( m_mode == QUARTER_STACK ) {
          pYuvD1->LPF_H_DSmooth( pYuvD1_half, 5, false );
          pYuvD2->LPF_H_DSmooth( pYuvD2_half, 5, false );
        }
        if ( m_mode == HALF_QUARTER_STACK ) {
          pYuvD1_temp->LPF_V_DExpand( pYuvD1_half, 5, false );
          pYuvD2_temp->LPF_V_DExpand( pYuvD2_half, 5, false );
          pYuvD1->LPF_H_DSmooth( pYuvD1_temp, 5, false );
          pYuvD2->LPF_H_DSmooth( pYuvD2_temp, 5, false );
        }
        if ( m_mode == QUARTER_ODD || m_mode == QUARTER_ODD_L ) {
          pYuvD1->LPF_H_DSmooth( pYuvD1_half, 5, false );
        }
      }
    }

    unsigned int from1, from2;
    if ( m_numDepth == DEPTH_1 ) {
      from1= RIGHT; from2= LEFT;
    } else {
      from1= LEFT; from2= RIGHT;
    }
    int cend;

    // DEBUG  Depth OUT ///////////////////////////////////
    // return true;

    if ( f_NFTfile != NULL ) {
      #ifdef WIN32
      if ( fscanf_s( f_NFTfile, "%lf,", &Tnear[C1] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Tfar[C1] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Ttrans[C1] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Tnear[C2] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Tfar[C2] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Ttrans[C2] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Tnear[C3] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Tfar[C3] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Ttrans[C3] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Tnear[C4] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Tfar[C4] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Ttrans[C4] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &D2near ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &D2far ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &D2trans ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Tnear[C6] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Tfar[C6] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Ttrans[C6] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Tnear[C7] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Tfar[C7] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Ttrans[C7] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Tnear[C8] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Tfar[C8] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Ttrans[C8] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Tnear[C9] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf,", &Tfar[C9] ) != 1 ) return false;
      if ( fscanf_s( f_NFTfile, "%lf", &Ttrans[C9] ) != 1 ) return false;

      #else
      if ( fscanf( f_NFTfile, "%lf,", &Tnear[C1] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Tfar[C1] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Ttrans[C1] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Tnear[C2] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Tfar[C2] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Ttrans[C2] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Tnear[C3] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Tfar[C3] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Ttrans[C3] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Tnear[C4] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Tfar[C4] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Ttrans[C4] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &D2near ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &D2far ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &D2trans ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Tnear[C6] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Tfar[C6] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Ttrans[C6] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Tnear[C7] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Tfar[C7] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Ttrans[C7] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Tnear[C8] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Tfar[C8] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Ttrans[C8] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Tnear[C9] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf,", &Tfar[C9] ) != 1 ) return false;
      if ( fscanf( f_NFTfile, "%lf", &Ttrans[C9] ) != 1 ) return false;

      #endif

      D1near= Tnear[C1]; D1far= Tfar[C1]; D1trans= Ttrans[C1];
      D3near= Tnear[C9]; D3far= Tfar[C9]; D3trans= Ttrans[C9];
      if ( m_mode == QUARTER_ODD || m_mode == QUARTER_ODD_L || m_mode == FULL_ODD_L ) {
        V1near= D2near; V1far= D2far; V1trans= D2trans;
      } else {
        V1near= Tnear[C3]; V1far= Tfar[C3]; V1trans= Ttrans[C3];
        V2near= Tnear[C7]; V2far= Tfar[C7]; V2trans= Ttrans[C7];
      }

      m_pViewSyn[C1]->ViewM= m_pViewSyn[C2]->ViewM= m_pViewSyn[C3]->ViewM= m_pViewSyn[C4]->ViewM= RefV;
      m_pViewSyn[C6]->ViewM= m_pViewSyn[C7]->ViewM= m_pViewSyn[C8]->ViewM= m_pViewSyn[C9]->ViewM= RefV;
      m_pViewSyn[C1]->fnum= m_pViewSyn[C2]->fnum= m_pViewSyn[C3]->fnum= m_pViewSyn[C4]->fnum= fnum;
      m_pViewSyn[C6]->fnum= m_pViewSyn[C7]->fnum= m_pViewSyn[C8]->fnum= m_pViewSyn[C9]->fnum= fnum;

      if ( m_mode == QUARTER_ODD || m_mode == QUARTER_ODD_L || m_mode == FULL_ODD_L ) {

        m_pViewSyn[C1]->ViewR= m_pViewSyn[C2]->ViewR= m_pViewSyn[C3]->ViewR= m_pViewSyn[C4]->ViewR= RefR1;
        m_pViewSyn[C6]->ViewR= m_pViewSyn[C7]->ViewR= m_pViewSyn[C8]->ViewR= m_pViewSyn[C9]->ViewR= RefR2;
        m_pViewSyn[C1]->Depth= m_pViewSyn[C2]->Depth= m_pViewSyn[C3]->Depth= m_pViewSyn[C4]->Depth= RefD1;
        m_pViewSyn[C6]->Depth= m_pViewSyn[C7]->Depth= m_pViewSyn[C8]->Depth= m_pViewSyn[C9]->Depth= RefD1;

        m_pViewSyn[C1]->near_C1= m_pViewSyn[C2]->near_C1= m_pViewSyn[C3]->near_C1= m_pViewSyn[C4]->near_C1= D1near;
        m_pViewSyn[C6]->near_C1= m_pViewSyn[C7]->near_C1= m_pViewSyn[C8]->near_C1= m_pViewSyn[C9]->near_C1= D3near;
        m_pViewSyn[C1]->far_C1= m_pViewSyn[C2]->far_C1= m_pViewSyn[C3]->far_C1= m_pViewSyn[C4]->far_C1= D1far;
        m_pViewSyn[C6]->far_C1= m_pViewSyn[C7]->far_C1= m_pViewSyn[C8]->far_C1= m_pViewSyn[C9]->far_C1= D3far;
        m_pViewSyn[C1]->trans_C1= m_pViewSyn[C2]->trans_C1= m_pViewSyn[C3]->trans_C1= m_pViewSyn[C4]->trans_C1= D1trans;
        m_pViewSyn[C6]->trans_C1= m_pViewSyn[C7]->trans_C1= m_pViewSyn[C8]->trans_C1= m_pViewSyn[C9]->trans_C1= D3trans;

        m_pViewSyn[C1]->near_C2= m_pViewSyn[C2]->near_C2= m_pViewSyn[C3]->near_C2= m_pViewSyn[C4]->near_C2= D2near;
        m_pViewSyn[C6]->near_C2= m_pViewSyn[C7]->near_C2= m_pViewSyn[C8]->near_C2= m_pViewSyn[C9]->near_C2= D2near;
        m_pViewSyn[C1]->far_C2= m_pViewSyn[C2]->far_C2= m_pViewSyn[C3]->far_C2= m_pViewSyn[C4]->far_C2= D2far;
        m_pViewSyn[C6]->far_C2= m_pViewSyn[C7]->far_C2= m_pViewSyn[C8]->far_C2= m_pViewSyn[C9]->far_C2= D2far;
        m_pViewSyn[C1]->trans_C2= m_pViewSyn[C2]->trans_C2= m_pViewSyn[C3]->trans_C2= m_pViewSyn[C4]->trans_C2= D2trans;
        m_pViewSyn[C6]->trans_C2= m_pViewSyn[C7]->trans_C2= m_pViewSyn[C8]->trans_C2= m_pViewSyn[C9]->trans_C2= D2trans;

        m_pViewSyn[C1]->near_V1= m_pViewSyn[C2]->near_V1= m_pViewSyn[C3]->near_V1= m_pViewSyn[C4]->near_V1= D2near;
        m_pViewSyn[C6]->near_V1= m_pViewSyn[C7]->near_V1= m_pViewSyn[C8]->near_V1= m_pViewSyn[C9]->near_V1= D2near;
        m_pViewSyn[C1]->far_V1= m_pViewSyn[C2]->far_V1= m_pViewSyn[C3]->far_V1= m_pViewSyn[C4]->far_V1= D2far;
        m_pViewSyn[C6]->far_V1= m_pViewSyn[C7]->far_V1= m_pViewSyn[C8]->far_V1= m_pViewSyn[C9]->far_V1= D2far;
        m_pViewSyn[C1]->trans_V1= m_pViewSyn[C2]->trans_V1= m_pViewSyn[C3]->trans_V1= m_pViewSyn[C4]->trans_V1= D2trans;
        m_pViewSyn[C6]->trans_V1= m_pViewSyn[C7]->trans_V1= m_pViewSyn[C8]->trans_V1= m_pViewSyn[C9]->trans_V1= D2trans;

        m_pViewSyn[C1]->from= m_pViewSyn[C2]->from= m_pViewSyn[C3]->from= m_pViewSyn[C4]->from= RIGHT;
        m_pViewSyn[C6]->from= m_pViewSyn[C7]->from= m_pViewSyn[C8]->from= m_pViewSyn[C9]->from= LEFT;

#ifdef DEBUGOUT
        m_pViewSyn[C1]->debug= m_pViewSyn[C9]->debug= true;
#endif
        cend= 8;

      } else {

        m_pViewSyn[C1]->ViewR= m_pViewSyn[C2]->ViewR= m_pViewSyn[C3]->ViewR= m_pViewSyn[C4]->ViewR= RefR1;
        m_pViewSyn[C1]->Depth= m_pViewSyn[C2]->Depth= m_pViewSyn[C3]->Depth= m_pViewSyn[C4]->Depth= RefD1;

        m_pViewSyn[C1]->near_C1= m_pViewSyn[C2]->near_C1= m_pViewSyn[C3]->near_C1= m_pViewSyn[C4]->near_C1= D1near;
        m_pViewSyn[C1]->far_C1= m_pViewSyn[C2]->far_C1= m_pViewSyn[C3]->far_C1= m_pViewSyn[C4]->far_C1= D1far;
        m_pViewSyn[C1]->trans_C1= m_pViewSyn[C2]->trans_C1= m_pViewSyn[C3]->trans_C1= m_pViewSyn[C4]->trans_C1= D1trans;

        m_pViewSyn[C1]->near_C2= m_pViewSyn[C2]->near_C2= m_pViewSyn[C3]->near_C2= m_pViewSyn[C4]->near_C2= D2near;
        m_pViewSyn[C1]->far_C2= m_pViewSyn[C2]->far_C2= m_pViewSyn[C3]->far_C2= m_pViewSyn[C4]->far_C2= D2far;
        m_pViewSyn[C1]->trans_C2= m_pViewSyn[C2]->trans_C2= m_pViewSyn[C3]->trans_C2= m_pViewSyn[C4]->trans_C2= D2trans;

        m_pViewSyn[C1]->near_V1= m_pViewSyn[C2]->near_V1= m_pViewSyn[C3]->near_V1= m_pViewSyn[C4]->near_V1= V1near;
        m_pViewSyn[C1]->far_V1= m_pViewSyn[C2]->far_V1= m_pViewSyn[C3]->far_V1= m_pViewSyn[C4]->far_V1= V1far;
        m_pViewSyn[C1]->trans_V1= m_pViewSyn[C2]->trans_V1= m_pViewSyn[C3]->trans_V1= m_pViewSyn[C4]->trans_V1= V1trans;

        m_pViewSyn[C1]->from= m_pViewSyn[C2]->from= RIGHT;
        m_pViewSyn[C3]->from= m_pViewSyn[C4]->from= from1;

#ifdef DEBUGOUT
        m_pViewSyn[C1]->debug= true;
#endif
        cend= 4;

        if (  m_numView == VIEW_3 ) {
          m_pViewSyn[C6]->ViewR= m_pViewSyn[C7]->ViewR= m_pViewSyn[C8]->ViewR= m_pViewSyn[C9]->ViewR= RefR2;
          m_pViewSyn[C6]->Depth= m_pViewSyn[C7]->Depth= m_pViewSyn[C8]->Depth= m_pViewSyn[C9]->Depth= RefD2;

          m_pViewSyn[C6]->near_C1= m_pViewSyn[C7]->near_C1= m_pViewSyn[C8]->near_C1= m_pViewSyn[C9]->near_C1= D3near;
          m_pViewSyn[C6]->far_C1= m_pViewSyn[C7]->far_C1= m_pViewSyn[C8]->far_C1= m_pViewSyn[C9]->far_C1= D3far;
          m_pViewSyn[C6]->trans_C1= m_pViewSyn[C7]->trans_C1= m_pViewSyn[C8]->trans_C1= m_pViewSyn[C9]->trans_C1= D3trans;

          m_pViewSyn[C6]->near_C2= m_pViewSyn[C7]->near_C2= m_pViewSyn[C8]->near_C2= m_pViewSyn[C9]->near_C2= D2near;
          m_pViewSyn[C6]->far_C2= m_pViewSyn[C7]->far_C2= m_pViewSyn[C8]->far_C2= m_pViewSyn[C9]->far_C2= D2far;
          m_pViewSyn[C6]->trans_C2= m_pViewSyn[C7]->trans_C2= m_pViewSyn[C8]->trans_C2= m_pViewSyn[C9]->trans_C2= D2trans;

          m_pViewSyn[C6]->near_V1= m_pViewSyn[C7]->near_V1= m_pViewSyn[C8]->near_V1= m_pViewSyn[C9]->near_V1= V2near;
          m_pViewSyn[C6]->far_V1= m_pViewSyn[C7]->far_V1= m_pViewSyn[C8]->far_V1= m_pViewSyn[C9]->far_V1= V2far;
          m_pViewSyn[C6]->trans_V1= m_pViewSyn[C7]->trans_V1= m_pViewSyn[C8]->trans_V1= m_pViewSyn[C9]->trans_V1= V2trans;

          m_pViewSyn[C6]->from= m_pViewSyn[C7]->from= RIGHT; // LEFT ? June 09
          m_pViewSyn[C8]->from= m_pViewSyn[C9]->from= from2;

#ifdef DEBUGOUT
          m_pViewSyn[C9]->debug= true;
#endif
          cend= 8;
        }
      }

    } else {

      if ( m_mode == QUARTER_ODD || m_mode == QUARTER_ODD_L || m_mode == FULL_ODD_L ) {

        m_pViewSyn[C1]->ViewR= m_pViewSyn[C2]->ViewR= m_pViewSyn[C3]->ViewR= m_pViewSyn[C4]->ViewR= RefR1;
        m_pViewSyn[C6]->ViewR= m_pViewSyn[C7]->ViewR= m_pViewSyn[C8]->ViewR= m_pViewSyn[C9]->ViewR= RefR2;
        m_pViewSyn[C1]->Depth= m_pViewSyn[C2]->Depth= m_pViewSyn[C3]->Depth= m_pViewSyn[C4]->Depth= RefD1;
        m_pViewSyn[C6]->Depth= m_pViewSyn[C7]->Depth= m_pViewSyn[C8]->Depth= m_pViewSyn[C9]->Depth= RefD1;

        m_pViewSyn[C1]->near_C1= m_pViewSyn[C2]->near_C1= m_pViewSyn[C3]->near_C1= m_pViewSyn[C4]->near_C1= 0.0;
        m_pViewSyn[C6]->near_C1= m_pViewSyn[C7]->near_C1= m_pViewSyn[C8]->near_C1= m_pViewSyn[C9]->near_C1= 0.0;
        m_pViewSyn[C1]->far_C1= m_pViewSyn[C2]->far_C1= m_pViewSyn[C3]->far_C1= m_pViewSyn[C4]->far_C1= 0.0;
        m_pViewSyn[C6]->far_C1= m_pViewSyn[C7]->far_C1= m_pViewSyn[C8]->far_C1= m_pViewSyn[C9]->far_C1= 0.0;
        m_pViewSyn[C1]->trans_C1= m_pViewSyn[C2]->trans_C1= m_pViewSyn[C3]->trans_C1= m_pViewSyn[C4]->trans_C1= 0.0;
        m_pViewSyn[C6]->trans_C1= m_pViewSyn[C7]->trans_C1= m_pViewSyn[C8]->trans_C1= m_pViewSyn[C9]->trans_C1= 0.0;

        m_pViewSyn[C1]->near_C2= m_pViewSyn[C2]->near_C2= m_pViewSyn[C3]->near_C2= m_pViewSyn[C4]->near_C2= 0.0;
        m_pViewSyn[C6]->near_C2= m_pViewSyn[C7]->near_C2= m_pViewSyn[C8]->near_C2= m_pViewSyn[C9]->near_C2= 0.0;
        m_pViewSyn[C1]->far_C2= m_pViewSyn[C2]->far_C2= m_pViewSyn[C3]->far_C2= m_pViewSyn[C4]->far_C2= 0.0;
        m_pViewSyn[C6]->far_C2= m_pViewSyn[C7]->far_C2= m_pViewSyn[C8]->far_C2= m_pViewSyn[C9]->far_C2= 0.0;
        m_pViewSyn[C1]->trans_C2= m_pViewSyn[C2]->trans_C2= m_pViewSyn[C3]->trans_C2= m_pViewSyn[C4]->trans_C2= 0.0;
        m_pViewSyn[C6]->trans_C2= m_pViewSyn[C7]->trans_C2= m_pViewSyn[C8]->trans_C2= m_pViewSyn[C9]->trans_C2= 0.0;

        m_pViewSyn[C1]->near_V1= m_pViewSyn[C2]->near_V1= m_pViewSyn[C3]->near_V1= m_pViewSyn[C4]->near_V1= 0.0;
        m_pViewSyn[C6]->near_V1= m_pViewSyn[C7]->near_V1= m_pViewSyn[C8]->near_V1= m_pViewSyn[C9]->near_V1= 0.0;
        m_pViewSyn[C1]->far_V1= m_pViewSyn[C2]->far_V1= m_pViewSyn[C3]->far_V1= m_pViewSyn[C4]->far_V1= 0.0;
        m_pViewSyn[C6]->far_V1= m_pViewSyn[C7]->far_V1= m_pViewSyn[C8]->far_V1= m_pViewSyn[C9]->far_V1= 0.0;
        m_pViewSyn[C1]->trans_V1= m_pViewSyn[C2]->trans_V1= m_pViewSyn[C3]->trans_V1= m_pViewSyn[C4]->trans_V1= 0.0;
        m_pViewSyn[C6]->trans_V1= m_pViewSyn[C7]->trans_V1= m_pViewSyn[C8]->trans_V1= m_pViewSyn[C9]->trans_V1= 0.0;

        Tnear[C1]= Tnear[C2]= Tnear[C3]= Tnear[C4]= 0.0;
        Tnear[C6]= Tnear[C7]= Tnear[C8]= Tnear[C9]= 0.0;

        m_pViewSyn[C1]->from= m_pViewSyn[C2]->from= m_pViewSyn[C3]->from= m_pViewSyn[C4]->from= RIGHT;
        m_pViewSyn[C6]->from= m_pViewSyn[C7]->from= m_pViewSyn[C8]->from= m_pViewSyn[C9]->from= LEFT;

#ifdef DEBUGOUT
        m_pViewSyn[C1]->debug= m_pViewSyn[C9]->debug= true;
#endif
        cend= 8;

      } else {

        m_pViewSyn[C1]->ViewR= m_pViewSyn[C2]->ViewR= m_pViewSyn[C3]->ViewR= m_pViewSyn[C4]->ViewR= RefR1;
        m_pViewSyn[C1]->Depth= m_pViewSyn[C2]->Depth= m_pViewSyn[C3]->Depth= m_pViewSyn[C4]->Depth= RefD1;

        m_pViewSyn[C1]->near_C1= m_pViewSyn[C2]->near_C1= m_pViewSyn[C3]->near_C1= m_pViewSyn[C4]->near_C1= 0.0;
        m_pViewSyn[C1]->far_C1= m_pViewSyn[C2]->far_C1= m_pViewSyn[C3]->far_C1= m_pViewSyn[C4]->far_C1= 0.0;
        m_pViewSyn[C1]->trans_C1= m_pViewSyn[C2]->trans_C1= m_pViewSyn[C3]->trans_C1= m_pViewSyn[C4]->trans_C1= 0.0;

        m_pViewSyn[C1]->near_C2= m_pViewSyn[C2]->near_C2= m_pViewSyn[C3]->near_C2= m_pViewSyn[C4]->near_C2= 0.0;
        m_pViewSyn[C1]->far_C2= m_pViewSyn[C2]->far_C2= m_pViewSyn[C3]->far_C2= m_pViewSyn[C4]->far_C2= 0.0;
        m_pViewSyn[C1]->trans_C2= m_pViewSyn[C2]->trans_C2= m_pViewSyn[C3]->trans_C2= m_pViewSyn[C4]->trans_C2= 0.0;

        m_pViewSyn[C1]->near_V1= m_pViewSyn[C2]->near_V1= m_pViewSyn[C3]->near_V1= m_pViewSyn[C4]->near_V1= 0.0;
        m_pViewSyn[C1]->far_V1= m_pViewSyn[C2]->far_V1= m_pViewSyn[C3]->far_V1= m_pViewSyn[C4]->far_V1= 0.0;
        m_pViewSyn[C1]->trans_V1= m_pViewSyn[C2]->trans_V1= m_pViewSyn[C3]->trans_V1= m_pViewSyn[C4]->trans_V1= 0.0;

        Tnear[C1]= Tnear[C2]= Tnear[C3]= Tnear[C4]= 0.0;

        m_pViewSyn[C1]->from= m_pViewSyn[C2]->from= RIGHT;
        m_pViewSyn[C3]->from= m_pViewSyn[C4]->from= from1;

#ifdef DEBUGOUT
        m_pViewSyn[C1]->debug= true;
#endif
        cend= 4;

        if (  m_numView == VIEW_3 ) {
          m_pViewSyn[C6]->ViewR= m_pViewSyn[C7]->ViewR= m_pViewSyn[C8]->ViewR= m_pViewSyn[C9]->ViewR= RefR2;
          m_pViewSyn[C6]->Depth= m_pViewSyn[C7]->Depth= m_pViewSyn[C8]->Depth= m_pViewSyn[C9]->Depth= RefD2;

          m_pViewSyn[C6]->near_C1= m_pViewSyn[C7]->near_C1= m_pViewSyn[C8]->near_C1= m_pViewSyn[C9]->near_C1= 0.0;
          m_pViewSyn[C6]->far_C1= m_pViewSyn[C7]->far_C1= m_pViewSyn[C8]->far_C1= m_pViewSyn[C9]->far_C1= 0.0;
          m_pViewSyn[C6]->trans_C1= m_pViewSyn[C7]->trans_C1= m_pViewSyn[C8]->trans_C1= m_pViewSyn[C9]->trans_C1= 0.0;

          m_pViewSyn[C6]->near_C2= m_pViewSyn[C7]->near_C2= m_pViewSyn[C8]->near_C2= m_pViewSyn[C9]->near_C2= 0.0;
          m_pViewSyn[C6]->far_C2= m_pViewSyn[C7]->far_C2= m_pViewSyn[C8]->far_C2= m_pViewSyn[C9]->far_C2= 0.0;
          m_pViewSyn[C6]->trans_C2= m_pViewSyn[C7]->trans_C2= m_pViewSyn[C8]->trans_C2= m_pViewSyn[C9]->trans_C2= 0.0;

          m_pViewSyn[C6]->near_V1= m_pViewSyn[C7]->near_V1= m_pViewSyn[C8]->near_V1= m_pViewSyn[C9]->near_V1= 0.0;
          m_pViewSyn[C6]->far_V1= m_pViewSyn[C7]->far_V1= m_pViewSyn[C8]->far_V1= m_pViewSyn[C9]->far_V1= 0.0;
          m_pViewSyn[C6]->trans_V1= m_pViewSyn[C7]->trans_V1= m_pViewSyn[C8]->trans_V1= m_pViewSyn[C9]->trans_V1= 0.0;

          Tnear[C6]= Tnear[C7]= Tnear[C8]= Tnear[C9]= 0.0;

          m_pViewSyn[C6]->from= m_pViewSyn[C7]->from= LEFT;
          m_pViewSyn[C8]->from= m_pViewSyn[C9]->from= from2;

#ifdef DEBUGOUT
          m_pViewSyn[C9]->debug= true;
#endif
          cend= 8;
        }
      }

    }

#pragma omp parallel for
    for (int cpos= 0; cpos < 8; cpos++)
      m_pViewSyn[cpos]->DoOneFrame( pYuvT[cpos], Tnear[cpos], Tfar[cpos], Ttrans[cpos] );

  } catch (...) {
    return false;
  }
  return true;
}



bool CSubSep::NoMix( FILE *fpT1, FILE *fpT2, FILE *fpT3, FILE *fpT4, FILE *fpT6, FILE *fpT7, FILE *fpT8, FILE *fpT9 ) {
  // DEBUG  Residuals OUT
  //  if (!pYuvR1->WriteOneFrame(fpT1)) return false; //_inYUV
  //  if (m_mode != QUARTER)
  //      if (!pYuvR2->WriteOneFrame(fpT2)) return false;
  //
  //  return true;

  // DEBUG  Depth OUT
  //  if (!pYuvD1->WriteOneFrame(fpT1)) return false; //_inYUV
  //  if (m_mode != QUARTER)
  //    if (!pYuvD2->WriteOneFrame(fpT2)) return false;
  //
  //  return true;

  #ifdef PTRACE
    printf("TRACE: CSubSep::NoMix()\n");
  #endif

  if (!pYuvT[C1]->WriteOneFrame(fpT1)) return false;
  if (!pYuvT[C2]->WriteOneFrame(fpT2)) return false;
  if (!pYuvT[C3]->WriteOneFrame(fpT3)) return false;
  if (!pYuvT[C4]->WriteOneFrame(fpT4)) return false;

  if ( m_numView == VIEW_3 ) {
    if (!pYuvT[C6]->WriteOneFrame(fpT6)) return false;
    if (!pYuvT[C7]->WriteOneFrame(fpT7)) return false;
    if (!pYuvT[C8]->WriteOneFrame(fpT8)) return false;
    if (!pYuvT[C9]->WriteOneFrame(fpT9)) return false;
  }

  return true;
}
