#pragma once
// View Synthesis Encoder for Depth
// Dev Enviroment:
//    Windows 7 64bit with Visual Studio 2010 C++ x64
//    OpenCV 2.4.2 64bit
// Original code: View Synthesis Reference Software (VSRS) 3.5
// Designer: Senoh, NiCT, Tokyo. Japan.
// Last Update: Apr, 2013.
#ifdef WIN32
#pragma warning(disable:4996)
#pragma warning(disable:4244)
#pragma warning(disable:4819)
#endif

#ifndef WIN32
#define BYTE unsigned char
#endif

#include <iostream>
#include <fstream>

#include "yuv.h" 

typedef unsigned char uchar;
//using namespace std;

#ifndef UInt
#define UInt    unsigned int
#endif

#define NONDIR 0
#define LEFT 1
#define RIGHT 2
#define CENTER 3

namespace view_synthesis {
  namespace fnum_tri {
    struct tri {
      unsigned int current;
      unsigned int start;
      unsigned int total;
    };
  }
}

namespace fn= view_synthesis::fnum_tri;

class CDepthInterpolationGeneral 
{
public:
  CDepthInterpolationGeneral (); 
  virtual ~CDepthInterpolationGeneral (); 

  // Init both(all) depth channel.
  bool InitLR (  
        UInt uiWidth, UInt uiHeight, bool variparam, unsigned int mode,
        double dZnearL, double dZfarL, double dZnearR, double dZfarR, double dZnearO, double dZfarO, 
        const char *strRefLCamID, const char *strRefRCamID, const char *strRefOCamID, const char *strVirCamID, 
        double Mat_In_Left[9], double Mat_Ex_Left[9], double Mat_Trans_Left[3], 
        double Mat_In_Right[9], double Mat_Ex_Right[9], double Mat_Trans_Right[3], 
        double Mat_In_Outside[9], double Mat_Ex_Outside[9], double Mat_Trans_Outside[3], 
        double Mat_In_Virtual[9], double Mat_Ex_Virtual[9], double Mat_Trans_Virtual[3]
  );

  // Init each(single) depth channel. Launch from InitLR.
  bool Init (
        UInt uiWidth, UInt uiHeight, bool variparam,
        double dZnear, double dZfar, 
        const char *strRefCamID, 
        const char *strVirCamID,
        double Mat_In_Ref[9], double Mat_Ex_Ref[9], double Mat_Trans_Ref[3], 
        double Mat_In_Vir[9], double Mat_Ex_Vir[9], double Mat_Trans_Vir[3]
  ); 
  
  void xReleaseMemory ();

  bool xSynthesizeView (BYTE ***src, BYTE **pDepthMap, int th_same_depth=5);
  bool xSynthesizeView_reverse (BYTE ***src, BYTE **pDepthMap, int th_same_depth=5);

  double getBaselineDistance () { return m_dBaselineDistance; }
  double getLeftBaselineDistance () { return LeftBaselineDistance; }
  double getRightBaselineDistance () { return RightBaselineDistance; }

  IplImage* getHolePixels () { return m_imgHoles; }
  IplImage* getSynthesizedPixels () { return m_imgSuccessSynthesis; }
  IplImage* getJumpPixels () { return m_imgJump; }
  IplImage* getUnstablePixels () { return m_imgMask[0]; }
  IplImage* getVirtualDepthMap () { return m_imgVirtualDepth; }
  BYTE** getVirtualDepth () { return m_pVirtualDepth; }
  BYTE** getNonHoles () { return m_pSuccessSynthesis; }
  
  double getWeightLeft() { return m_dWeightLeft; }
  double getWeightRight() { return m_dWeightRight; }
  void setWeightLeft(double WeightLeft) { m_dWeightLeft= WeightLeft; }
  void setWeightRight(double WeightRight) { m_dWeightRight= WeightRight; }
  
  void SetWidth (int sWidth ) { Width  = sWidth;  }
  void SetHeight(int sHeight) { Height = sHeight; }
  void SetDepthType(int sDepthType) {DepthType = sDepthType; }
  void SetSubPelOption(int sPrecision) { Precision = sPrecision; }
  void SetViewBlending(int sViewBlending) { ViewBlending = sViewBlending; }

  void setDepthMap(IplImage* img) { cvCopy(img, m_imgVirtualDepth); }

  void depthInpaint( IplImage* map, IplImage* hole );
  void depthEnhanced( IplImage* map, IplImage* out, unsigned int th=5 );

  bool DoOneFrameDepth(
          BYTE** RefDepthLeft, BYTE** RefDepthRight, BYTE** RefDepthOutside, CIYuv* pSynYuvBuffer,
          double near_C1, double far_C1, double trans_C1, 
          double near_C2, double far_C2, double trans_C2, 
          double near_C3, double far_C3, double trans_C3, 
          double near_V, double far_V, double trans_V,
          fn::tri fnum,
          unsigned int side, bool mode2view );
  
  IplImage*  getImgSynthesizedViewLeft  ();
  IplImage*  getImgSynthesizedViewRight  ();
  IplImage*  getImgSynthesizedViewOutside  ();

protected:

private:
  CDepthInterpolationGeneral*  m_pcDepthSynthesisLeft; 
  CDepthInterpolationGeneral*  m_pcDepthSynthesisRight; 
  CDepthInterpolationGeneral*  m_pcDepthSynthesisOutside; 

  void convertCameraParam ( CvMat *exMat_dst, CvMat *exMat_src );

  bool init_camera_param (
          double Mat_In_Ref[9], double Mat_Ex_Ref[9], double Mat_Trans_Ref[3], 
          double Mat_In_Vir[9], double Mat_Ex_Vir[9], double Mat_Trans_Vir[3], 
          CvMat *mat_in[2], CvMat *mat_ex_c2w[2], CvMat *mat_proj[2]  ); 

  bool init_3Dwarp( double Z_near, double Z_far );

  bool reset_3Dwarp ( 
          double near_C, double far_C, double trans_C, 
          double near_V, double far_V, double trans_V );

  void image2world_with_z ( 
          CvMat *mat_Rc2w_invIN_from, CvMat *matEX_c2w_from, CvMat *image, CvMat *world );
  
  void makeHomography ( 
          CvMat *matH_F2T[256], CvMat *matH_T2F[256], double adTable[256], 
          CvMat *matIN_from, CvMat *matEX_c2w_from, CvMat *matProj_to );

  bool depthsynthesis_3Dwarp ( BYTE **pDepthMap, int side );

  double        m_dWeightLeft;
  double        m_dWeightRight;

  unsigned int  m_uiViewBlending; 
  unsigned int  ViewBlending; 

  unsigned int  m_uiWidth;
  unsigned int  m_uiHeight;
  unsigned int  m_uiPicsize;
  unsigned int  m_uiPrecision;
  unsigned int  m_uiDepthType;

  std::string   CameraParameterFile;
  std::string   LeftCameraName;
  std::string   RightCameraName;
  std::string   OutsideCameraName;
  std::string   VirtualCameraName;

  unsigned int  Width;
  unsigned int  Height;
  unsigned int  Picsize;
  unsigned int  Precision;
  unsigned int  DepthType;
  bool VariParam;

  // for VeriParam
  double  Mat_Ex_C[9];
  double  Mat_Ex_V[9];
  double  Mat_In_C[9];
  double  Mat_In_V[9];
  double  Mat_Trans_C[3];
  double  Mat_Trans_V[3];

  IplImage*  m_imgInterpolatedView; //!> The final image buffer to be output
  IplImage*  getImgInterpolatedView    ()    { return m_imgInterpolatedView; }

  double  m_dBaselineDistance;
  double  LeftBaselineDistance;
  double  RightBaselineDistance;

  BYTE**  m_pVirtualDepth;
  BYTE**  m_pSuccessSynthesis;
  BYTE**  m_pJump;

  IplImage*      m_imgVirtualDepth;
  IplImage*      m_imgSuccessSynthesis;
  IplImage*      m_imgJump;
  IplImage*      m_imgHoles;
  IplImage*      m_imgBound;
  IplImage*      m_imgMask[5];
  IplImage*      m_imgTemp[5];

  IplImage*      m_imgLHole;
  IplImage*      m_imgLSuccess;
  IplImage*      m_imgRHole;
  IplImage*      m_imgRSuccess;
  IplImage*      m_imgOHole;
  IplImage*      m_imgOSuccess;
  IplImage*      m_imgComHole;

  IplImage* m_depthPre2;
  IplImage* m_depthPre1;
  IplImage* m_depthBlended0;
  IplImage* m_depthBlended1;
  IplImage* m_depthBlended2;
  IplImage* m_depthBlendedMix;
  IplImage* m_depthBlendedMix2;
  IplImage* m_depthSuccess;
  void ToFilterDepth(IplImage* pDepthSRC, IplImage* pDepthDST, IplImage* pDepthNEGSRC, IplImage* pDepthNEGDST, unsigned int fsize);

  void mf3Depth( const IplImage* src1, const IplImage* src2, const IplImage* src3, IplImage* dst );
  void setDepth( BYTE **pDepthMap, IplImage* dst );

  CvMat*  m_matH_R2V[256];
  CvMat*  m_matH_V2R[256];
  
  // for VeriParam
  CvMat* m_Mat_in[2]; //intrinsic parameter of left camera 3x3 matrix
  CvMat* m_Mat_ex_c2w[2]; //extrinsic parameter of left camera 3x4 matrix
  CvMat* m_Mat_proj_w2i[2]; // projection matrix from world to image

};
