#pragma once
// View Synthesis Encoder for Depth
// Dev Enviroment:
//    Windows 7 64bit with Visual Studio 2010 C++ x64
//    OpenCV 2.4.2 64bit
// Original code: View Synthesis Reference Software (VSRS) 3.5
// Last Update: Feb, 2014.
#include <utility>
#include <string>
#include "SubMixOptions.h"
#include "SubMix.h"
#include "ReturnCode.h" 
#include "ParameterEncDepth.h"
#include "DepthSynthesis.h"

namespace view_synthesis {
  namespace depth_submix_options {
    using namespace std;
    namespace fn= view_synthesis::fnum_tri;

    CParameterEncDepth cParameter;

    unsigned int mode;
    typedef pair<unsigned int, string> modepair;
    modepair mode1(QUARTER, "QUARTER"); 
    modepair mode2(QUARTER_STACK, "QUARTER_STACK");
    modepair mode3(HALF_QUARTER_STACK, "HALF_QUARTER_STACK");
    modepair mode4(QUARTER_ODD, "QUARTER_ODD"); // Single global depth(quarter size)
    modepair mode6(FULL_ODD, "FULL_ODD"); // Single global depth(full size)

    unsigned int checkOption(string option) {
      if ( 0 == option.compare(mode1.second) ) {mode= mode1.first; return mode;}
      if ( 0 == option.compare(mode2.second) ) {mode= mode2.first; return mode;}
      if ( 0 == option.compare(mode3.second) ) {mode= mode3.first; return mode;}
      if ( 0 == option.compare(mode4.second) ) {mode= mode4.first; return mode;}
      if ( 0 == option.compare(mode6.second) ) {mode= mode6.first; return mode;}
      cout << "VSencD: ERR: Config parameter: Mode: ->" << option << "<- Unknown." << endl << endl;
      return UNKNOWN;
    }


    bool VariParam= false;
    unsigned int width, height, half_width, half_height;
    unsigned int start, total;
    double near1, far1, near2, far2, near3, far3;
    class NoEven {};
    class BadSeq {};
    class MixPosNeg {};

    int initParam() {
      try {
        width= cParameter.getSourceWidth();
        height= cParameter.getSourceHeight();
        half_width= width / 2;
        half_height= height / 2;
        if ( width % 2 != 0 || height % 2 != 0) throw NoEven();
      } catch (NoEven) {
        cout << "VSencD: ERR: Config parameter: ""SourceWidth"" or ""SourceHeight"" Odd, No even." << endl << endl;
        return BAD_CONFIG_OPTIONS;
      } catch (...) {
        cout << "VSencD: ERR: Config parameter: ""SourceWidth"" or ""SourceHeight""." << endl << endl;
        return BAD_CONFIG_OPTIONS;
      }

      try {
        start= cParameter.getStartFrame();
        total= cParameter.getNumberOfFrames();
        // if ( start >= total ) throw BadSeq();
      } catch (BadSeq) {
        cout << "VSencD: ERR: Config parameter: ""StartFrame"", ""NumberOfFrames""  Bad order." << endl << endl;
        return BAD_CONFIG_OPTIONS;
      } catch (...) {
        cout << "VSencD: ERR: Config parameter: ""StartFrame"" or ""NumberOfFrames""." << endl << endl;
        return BAD_CONFIG_OPTIONS;
      }
      
      if ( cParameter.getNearestDepthValue1() == 0 && cParameter.getFarthestDepthValue1() == 0 ) 
        VariParam= true;
      else
        try {
          near1= cParameter.getNearestDepthValue1();
          far1= cParameter.getFarthestDepthValue1();
          near2= cParameter.getNearestDepthValue2();
          far2= cParameter.getFarthestDepthValue2();
          if ( mode == QUARTER || mode == HALF_QUARTER ) { // 2 View ////////////////////////////////////////////////////////////////////
            near3= 0;
            far3= 0;
            if ( near1<0 && (far1>0 || near2>0 || far2>0) ) throw MixPosNeg();
            if ( near1>0 && (far1<0 || near2<0 || far2<0) ) throw MixPosNeg();
            if ( (near1==0 && far1==0) || (near2==0 && far2==0) ) throw MixPosNeg();
            if ( near1==far1 || near2==far2 ) throw MixPosNeg();
          } else { // FULL_ODD, QUARTER_ODD, QUARTER_STACK, HALF_QUARTER_STACK, for 3view //////////////////////////////////////////////////////////////////
            near3= cParameter.getNearestDepthValue3();
            far3= cParameter.getFarthestDepthValue3();
            if ( near1<0 && (far1>0 || near2>0 || far2>0 || near3>0 || far3>0) ) throw MixPosNeg();
            if ( near1>0 && (far1<0 || near2<0 || far2<0 || near3<0 || far3<0) ) throw MixPosNeg();
            if ( (near1==0 && far1==0) || (near2==0 && far2==0)  || (near3==0 && far3==0) ) throw MixPosNeg();
            if ( near1==far1 || near2==far2 || near3==far3 ) throw MixPosNeg();
          }
        } catch (MixPosNeg){
          cout << "VSencD: ERR: Config parameter: ""NearestDepthValue"" and ""FarthestDepthValue""  Mixed positive and negative." << endl << endl;
          return BAD_CONFIG_OPTIONS;
        } catch (...) {
          cout << "VSencD: ERR: Config parameter: ""NearestDepthValue"" or ""FarthestDepthValue""." << endl << endl;
          return BAD_CONFIG_OPTIONS;
        }
      
      return NORMAL_END;
    }

    // QUARTER_STACK
    //   Depth: [LEFT]=D1, [CENTER]=D2, [RIGHT]=D3  -> Stack D1.5(upper) with D2.5(bottom)
    // QUARTER
    //   Depth: [LEFT]=D1, [CENTER]=D2  -> D1.5 (Alone, not binded!)
    FILE *f_D1, *f_D2, *f_D3, *f_BindD, *f_BindDD;
    #ifdef WIN32
    errno_t err_BindD, err_BindDD;
    #endif
    bool workD= true;
    class NoInfile {};
    class NoOutfile {};
    // for VariParam
    FILE *f_D1near, *f_D1far, *f_D1trans;
    FILE *f_D2near, *f_D2far, *f_D2trans;
    FILE *f_D3near, *f_D3far, *f_D3trans;
    FILE *f_V1near, *f_V1far, *f_V1trans;
    FILE *f_V2near, *f_V2far, *f_V2trans;

    int initFile() {
      try {
        #ifdef WIN32
        f_D1= _fsopen(cParameter.get_D1Name().c_str(), "rb", _SH_DENYWR); if (f_D1 == NULL) throw NoInfile();
        f_D2= _fsopen(cParameter.get_D2Name().c_str(), "rb", _SH_DENYWR); if (f_D2 == NULL) throw NoInfile();
        err_BindD= fopen_s(&f_BindD, cParameter.getBind_DName().c_str(), "wb"); if (err_BindD != 0) throw NoOutfile();
        cParameter.setBind_DDName( "Dummy" + cParameter.getBind_DName() );  // Apr16 For Dummy Depth
        err_BindD= fopen_s(&f_BindDD, cParameter.getBind_DDName().c_str(), "wb"); if (err_BindDD != 0) throw NoOutfile();
        if ( VariParam ) {
          f_D1near= _fsopen(cParameter.getNearFN1().c_str(), "rb", _SH_DENYWR); if (f_D1near == NULL) throw NoInfile();
          f_D1far= _fsopen(cParameter.getFarFN1().c_str(), "rb", _SH_DENYWR); if (f_D1far == NULL) throw NoInfile();
          f_D1trans= _fsopen(cParameter.getTransFN1().c_str(), "rb", _SH_DENYWR); if (f_D1trans == NULL) throw NoInfile();
          f_D2near= _fsopen(cParameter.getNearFN2().c_str(), "rb", _SH_DENYWR); if (f_D2near == NULL) throw NoInfile();
          f_D2far= _fsopen(cParameter.getFarFN2().c_str(), "rb", _SH_DENYWR); if (f_D2far == NULL) throw NoInfile();
          f_D2trans= _fsopen(cParameter.getTransFN2().c_str(), "rb", _SH_DENYWR); if (f_D2trans == NULL) throw NoInfile();
          f_V1near= _fsopen(cParameter.getNearFNV1().c_str(), "rb", _SH_DENYWR); if (f_V1near == NULL) throw NoInfile();
          f_V1far= _fsopen(cParameter.getFarFNV1().c_str(), "rb", _SH_DENYWR); if (f_V1far == NULL) throw NoInfile();
          f_V1trans= _fsopen(cParameter.getTransFNV1().c_str(), "rb", _SH_DENYWR); if (f_V1trans == NULL) throw NoInfile();
          if ( fseek(f_D1near, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_D1far, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_D1trans, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_D2near, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_D2far, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_D2trans, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_V1near, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_V1far, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_V1trans, 0, SEEK_SET) )  throw NoInfile();
        }

        if ( mode != QUARTER && mode != HALF_QUARTER ) {
          f_D3= _fsopen(cParameter.get_D3Name().c_str(), "rb", _SH_DENYWR); if (f_D3 == NULL) throw NoInfile();
          if ( VariParam ) {
            f_D3near= _fsopen(cParameter.getNearFN3().c_str(), "rb", _SH_DENYWR); if (f_D3near == NULL) throw NoInfile();
            f_D3far= _fsopen(cParameter.getFarFN3().c_str(), "rb", _SH_DENYWR); if (f_D3far == NULL) throw NoInfile();
            f_D3trans= _fsopen(cParameter.getTransFN3().c_str(), "rb", _SH_DENYWR); if (f_D3trans == NULL) throw NoInfile();
            if ( fseek(f_D3near, 0, SEEK_SET) )  throw NoInfile();
            if ( fseek(f_D3far, 0, SEEK_SET) )  throw NoInfile();
            if ( fseek(f_D3trans, 0, SEEK_SET) )  throw NoInfile();
            if ( mode != QUARTER_ODD && mode != QUARTER_ODD_L && mode != FULL_ODD ) {
              f_V2near= _fsopen(cParameter.getNearFNV2().c_str(), "rb", _SH_DENYWR); if (f_V2near == NULL) throw NoInfile();
              f_V2far= _fsopen(cParameter.getFarFNV2().c_str(), "rb", _SH_DENYWR); if (f_V2far == NULL) throw NoInfile();
              f_V2trans= _fsopen(cParameter.getTransFNV2().c_str(), "rb", _SH_DENYWR); if (f_V2trans == NULL) throw NoInfile();
              if ( fseek(f_V2near, 0, SEEK_SET) )  throw NoInfile();
              if ( fseek(f_V2far, 0, SEEK_SET) )  throw NoInfile();
              if ( fseek(f_V2trans, 0, SEEK_SET) )  throw NoInfile();
            }
          }
        }
        
        #else
        f_D1= fopen(cParameter.get_D1Name().c_str(), "rb"); if (f_D1 == NULL) throw NoInfile();
        f_D2= fopen(cParameter.get_D2Name().c_str(), "rb"); if (f_D2 == NULL) throw NoInfile();
        f_BindD= fopen(cParameter.getBind_DName().c_str(), "wb"); if (f_BindD == NULL) throw NoOutfile();
        cParameter.setBind_DDName( "Dummy" + cParameter.getBind_DName() );  // Apr16 For Dummy Depth
        f_BindDD= fopen(cParameter.getBind_DDName().c_str(), "wb"); if (f_BindDD == NULL) throw NoOutfile();
        if ( VariParam ) {
          f_D1near= fopen(cParameter.getNearFN1().c_str(), "rb"); if (f_D1near == NULL) throw NoInfile();
          f_D1far= fopen(cParameter.getFarFN1().c_str(), "rb"); if (f_D1far == NULL) throw NoInfile();
          f_D1trans= fopen(cParameter.getTransFN1().c_str(), "rb"); if (f_D1trans == NULL) throw NoInfile();
          f_D2near= fopen(cParameter.getNearFN2().c_str(), "rb"); if (f_D2near == NULL) throw NoInfile();
          f_D2far= fopen(cParameter.getFarFN2().c_str(), "rb"); if (f_D2far == NULL) throw NoInfile();
          f_D2trans= fopen(cParameter.getTransFN2().c_str(), "rb"); if (f_D2trans == NULL) throw NoInfile();
          f_V1near= fopen(cParameter.getNearFNV1().c_str(), "rb"); if (f_V1near == NULL) throw NoInfile();
          f_V1far= fopen(cParameter.getFarFNV1().c_str(), "rb"); if (f_V1far == NULL) throw NoInfile();
          f_V1trans= fopen(cParameter.getTransFNV1().c_str(), "rb"); if (f_V1trans == NULL) throw NoInfile();
          if ( fseek(f_D1near, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_D1far, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_D1trans, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_D2near, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_D2far, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_D2trans, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_V1near, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_V1far, 0, SEEK_SET) )  throw NoInfile();
          if ( fseek(f_V1trans, 0, SEEK_SET) )  throw NoInfile();
        }

        if ( mode != QUARTER && mode != HALF_QUARTER ) {
          f_D3= fopen(cParameter.get_D3Name().c_str(), "rb"); if (f_D3 == NULL) throw NoInfile();
          if ( VariParam ) {
            f_D3near= fopen(cParameter.getNearFN3().c_str(), "rb"); if (f_D3near == NULL) throw NoInfile();
            f_D3far= fopen(cParameter.getFarFN3().c_str(), "rb"); if (f_D3far == NULL) throw NoInfile();
            f_D3trans= fopen(cParameter.getTransFN3().c_str(), "rb"); if (f_D3trans == NULL) throw NoInfile();
            if ( fseek(f_D3near, 0, SEEK_SET) )  throw NoInfile();
            if ( fseek(f_D3far, 0, SEEK_SET) )  throw NoInfile();
            if ( fseek(f_D3trans, 0, SEEK_SET) )  throw NoInfile();
            if ( mode != QUARTER_ODD && mode != QUARTER_ODD_L && mode != FULL_ODD ) {
              f_V2near= fopen(cParameter.getNearFNV2().c_str(), "rb"); if (f_V2near == NULL) throw NoInfile();
              f_V2far= fopen(cParameter.getFarFNV2().c_str(), "rb"); if (f_V2far == NULL) throw NoInfile();
              f_V2trans= fopen(cParameter.getTransFNV2().c_str(), "rb"); if (f_V2trans == NULL) throw NoInfile();
              if ( fseek(f_V2near, 0, SEEK_SET) )  throw NoInfile();
              if ( fseek(f_V2far, 0, SEEK_SET) )  throw NoInfile();
              if ( fseek(f_V2trans, 0, SEEK_SET) )  throw NoInfile();
            }
          }
        }
        #endif
      } catch (NoInfile) { 
        cout << "VSencD: ERR: Not specify files, or File path/IO error." << endl;
        if (f_D1 != NULL) fclose(f_D1); 
        if (f_D2 != NULL) fclose(f_D2); 
        if ( VariParam ) {
          if (f_D1near != NULL) fclose(f_D1near); 
          if (f_D1far != NULL) fclose(f_D1far); 
          if (f_D1trans != NULL) fclose(f_D1trans); 
          if (f_D2near != NULL) fclose(f_D2near); 
          if (f_D2far != NULL) fclose(f_D2far); 
          if (f_D2trans != NULL) fclose(f_D2trans); 
          if (f_V1near != NULL) fclose(f_V1near); 
          if (f_V1far != NULL) fclose(f_V1far); 
          if (f_V1trans != NULL) fclose(f_V1trans); 
        }

        if ( mode != QUARTER && mode != HALF_QUARTER ) {
          if (f_D3 != NULL) fclose(f_D3); 
          if ( VariParam ) {
            if (f_D3near != NULL) fclose(f_D3near); 
            if (f_D3far != NULL) fclose(f_D3far); 
            if (f_D3trans != NULL) fclose(f_D3trans); 
            if ( mode != QUARTER_ODD && mode != QUARTER_ODD_L && mode != FULL_ODD ) {
              if (f_V2near != NULL) fclose(f_V2near); 
              if (f_V2far != NULL) fclose(f_V2far); 
              if (f_V2trans != NULL) fclose(f_V2trans); 
            }
          }
       }
        workD= false;
      } catch (...) {
        cout << "VSencD: ERR: Not specify output file, or File path/IO error." << endl;
        workD= false;
      }
      if ( !workD ) return NO_FILEIO;


      if ( VariParam ) {
        #ifdef WIN32
        if ( fscanf_s(f_D1near, "%lf", &near1) != 1 ) return false;
        if ( fscanf_s(f_D1far, "%lf", &far1) != 1 ) return false;
        if ( fscanf_s(f_D2near, "%lf", &near2) != 1 ) return false;
        if ( fscanf_s(f_D2far, "%lf", &far2) != 1 ) return false;
        if ( fseek(f_D1near, 0, SEEK_SET) )  return false;
        if ( fseek(f_D1far, 0, SEEK_SET) )  return false;
        if ( fseek(f_D2near, 0, SEEK_SET) )  return false;
        if ( fseek(f_D2far, 0, SEEK_SET) )  return false;
        if ( mode != QUARTER && mode != HALF_QUARTER ) {
          if ( fscanf_s(f_D3near, "%lf", &near3) != 1 ) return false;
          if ( fscanf_s(f_D3far, "%lf", &far3) != 1 ) return false;
          if ( fseek(f_D3near, 0, SEEK_SET) )  return false;
          if ( fseek(f_D3far, 0, SEEK_SET) )  return false;
        }
        
        #else
        if ( fscanf(f_D1near, "%lf", &near1) != 1 ) return false;
        if ( fscanf(f_D1far, "%lf", &far1) != 1 ) return false;
        if ( fscanf(f_D2near, "%lf", &near2) != 1 ) return false;
        if ( fscanf(f_D2far, "%lf", &far2) != 1 ) return false;
        if ( fseek(f_D1near, 0, SEEK_SET) )  return false;
        if ( fseek(f_D1far, 0, SEEK_SET) )  return false;
        if ( fseek(f_D2near, 0, SEEK_SET) )  return false;
        if ( fseek(f_D2far, 0, SEEK_SET) )  return false;
        if ( mode != QUARTER && mode != HALF_QUARTER ) {
          if ( fscanf(f_D3near, "%lf", &near3) != 1 ) return false;
          if ( fscanf(f_D3far, "%lf", &far3) != 1 ) return false;
          if ( fseek(f_D3near, 0, SEEK_SET) )  return false;
          if ( fseek(f_D3far, 0, SEEK_SET) )  return false;
        }
        #endif
      }


      return NORMAL_END;
    }


    void endFile() {
      if (workD) { 
        fclose(f_D1); fclose(f_D2); fclose(f_BindD); fclose(f_BindDD); 
        if ( VariParam ) {
          fclose(f_D1near); fclose(f_D1far); fclose(f_D1trans);
          fclose(f_D2near); fclose(f_D2far); fclose(f_D2trans);
          fclose(f_V1near); fclose(f_V1far); fclose(f_V1trans);
        }

        if ( mode != QUARTER && mode != HALF_QUARTER ) {
          fclose(f_D3);
          if ( VariParam ) {
            fclose(f_D3near); fclose(f_D3far); fclose(f_D3trans);
            if ( mode != QUARTER_ODD && mode != QUARTER_ODD_L && mode != FULL_ODD ) {
              fclose(f_V2near); fclose(f_V2far); fclose(f_V2trans);
            }
          }
        }
      }
    }



    CSubMix cSubD;

    int initSubMix() {
      // printf( "DEBUG: initSubMix() start.\n" );
      if ( mode == QUARTER || mode == HALF_QUARTER ) { // for 2view ///////////////////////////////////////////////////////////////////
        if ( !cSubD.Init(
                  mode, VariParam, width, height, half_width, half_height, 
                  near1, far1, near2, far2, 0, 0,
                  cParameter.getCameraName1().c_str(), 
                  cParameter.getCameraName2().c_str(), 
                  "", 
                  cParameter.getVertualCameraName1().c_str(),
                  "",
                  cParameter.getMat_In_1(), cParameter.getMat_Ex_1(), cParameter.getMat_Trans_1(),
                  cParameter.getMat_In_2(), cParameter.getMat_Ex_2(), cParameter.getMat_Trans_2(),
                  NULL, NULL, NULL,
                  cParameter.getMat_In_V1(), cParameter.getMat_Ex_V1(), cParameter.getMat_Trans_V1(),
                  NULL, NULL, NULL
        ) ) { noInitSUBMIX; return NO_INIT_SUBMIX_D; }
      } else if ( mode == QUARTER_ODD || mode == QUARTER_ODD_L ) {
        if ( !cSubD.Init(
                  mode, VariParam, width, height, half_width, height,
                  near1, far1, near2, far2, near3, far3,
                  cParameter.getCameraName1().c_str(), 
                  cParameter.getCameraName2().c_str(), 
                  cParameter.getCameraName3().c_str(), 
                  cParameter.getVertualCameraName1().c_str(),
                  cParameter.getVertualCameraName2().c_str(),
                  cParameter.getMat_In_1(), cParameter.getMat_Ex_1(), cParameter.getMat_Trans_1(),
                  cParameter.getMat_In_2(), cParameter.getMat_Ex_2(), cParameter.getMat_Trans_2(),
                  cParameter.getMat_In_3(), cParameter.getMat_Ex_3(), cParameter.getMat_Trans_3(),
                  cParameter.getMat_In_V1(), cParameter.getMat_Ex_V1(), cParameter.getMat_Trans_V1(),
                  NULL, NULL, NULL
        ) ) { noInitSUBMIX; return NO_INIT_SUBMIX_D; }
      } else if ( mode == FULL_ODD ) {
        if ( !cSubD.Init(
                  mode, VariParam, width, height, width, height,
                  near1, far1, near2, far2, near3, far3,
                  cParameter.getCameraName1().c_str(), 
                  cParameter.getCameraName2().c_str(), 
                  cParameter.getCameraName3().c_str(), 
                  cParameter.getVertualCameraName1().c_str(),
                  cParameter.getVertualCameraName2().c_str(),
                  cParameter.getMat_In_1(), cParameter.getMat_Ex_1(), cParameter.getMat_Trans_1(),
                  cParameter.getMat_In_2(), cParameter.getMat_Ex_2(), cParameter.getMat_Trans_2(),
                  cParameter.getMat_In_3(), cParameter.getMat_Ex_3(), cParameter.getMat_Trans_3(),
                  cParameter.getMat_In_V1(), cParameter.getMat_Ex_V1(), cParameter.getMat_Trans_V1(),
                  NULL, NULL, NULL
        ) ) { noInitSUBMIX; return NO_INIT_SUBMIX_D; }
      } else { // QUARTER_STACK, HALF_QUARTER_STACK for 3view //////////////////////////////////////////////////////////////////
        if ( !cSubD.Init(
                  mode, VariParam, width, height, half_width, height,
                  near1, far1, near2, far2, near3, far3,
                  cParameter.getCameraName1().c_str(), 
                  cParameter.getCameraName2().c_str(), 
                  cParameter.getCameraName3().c_str(), 
                  cParameter.getVertualCameraName1().c_str(),
                  cParameter.getVertualCameraName2().c_str(),
                  cParameter.getMat_In_1(), cParameter.getMat_Ex_1(), cParameter.getMat_Trans_1(),
                  cParameter.getMat_In_2(), cParameter.getMat_Ex_2(), cParameter.getMat_Trans_2(),
                  cParameter.getMat_In_3(), cParameter.getMat_Ex_3(), cParameter.getMat_Trans_3(),
                  cParameter.getMat_In_V1(), cParameter.getMat_Ex_V1(), cParameter.getMat_Trans_V1(),
                  cParameter.getMat_In_V2(), cParameter.getMat_Ex_V2(), cParameter.getMat_Trans_V2()
        ) ) { noInitSUBMIX; return NO_INIT_SUBMIX_D; }
      }
      // printf( "DEBUG: initSubMix() done.\n" );
      return NORMAL_END;
    }



    int convOneFrames(fn::tri fnum) {
      if ( mode == QUARTER || mode == HALF_QUARTER ) { // for 2view ///////////////////////////////////////////////////////////////////
        if (workD) {
          if (!cSubD.ReadOneFrame(
                  f_D1, f_D2, NULL, fnum,
                  // for VariParam
                  f_D1near, f_D1far, f_D1trans,
                  f_D2near, f_D2far, f_D2trans,
                  NULL, NULL, NULL,
                  f_V1near, f_V1far, f_V1trans,
                  NULL, NULL, NULL
           )) return DATA_SET_ERR;
          cout << ".";
          if (!cSubD.DoSubMix(f_BindD, f_BindDD)) return DATA_OUT_ERR;
          cout << ".";
        }
      } else { // QUARTER_STACK, HALF_QUARTER_STACK for 3view //////////////////////////////////////////////////////////////////
        if (workD) {
          if (!cSubD.ReadOneFrame(
                  f_D1, f_D2, f_D3, fnum,
                  // for VariParam
                  f_D1near, f_D1far, f_D1trans,
                  f_D2near, f_D2far, f_D2trans,
                  f_D3near, f_D3far, f_D3trans,
                  f_V1near, f_V1far, f_V1trans,
                  f_V2near, f_V2far, f_V2trans
           )) return DATA_SET_ERR;
          cout << ".";
          if (!cSubD.DoSubMix(f_BindD, f_BindDD)) return DATA_OUT_ERR;
          cout << ".";
        }
      }
      return NORMAL_END;
    }

  }

}
