// NICT GVD: View Synthesis Encoder for Depth
// Dev Enviroment:
//    Windows 7 64bit with Visual Studio 2010 C++ x64
//    OpenCV 2.4.5 64bit
// Other Test Enviroment:
//    Ubuntu 12.04 LTS 32bit
//    OpenCV 2.4.2 32bit
//    (VMware Player 5.0.0 on Windows 7 64bit)
// Original code: View Synthesis Reference Software (VSRS) 3.5
// Designer: Senoh, NiCT, Tokyo. Japan.
// Last Update: Feb, 2014.
#define GVDENC_MAJOR_VERSION 4
#define GVDENC_MINOR_VERSION 0
#define GVDENC_VERSION CVAUX_STR(GVDENC_MAJOR_VERSION) "." CVAUX_STR(GVDENC_MINOR_VERSION)

#define _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#include <cstdlib>

#ifdef WIN32
#include <crtdbg.h> 
static void MemReport_Init() {
  _CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF ); //| _CRTDBG_DELAY_FREE_MEM_DF );
  _CrtSetReportMode(_CRT_ERROR, _CRTDBG_MODE_FILE); //
  _CrtSetReportFile(_CRT_ERROR, _CRTDBG_FILE_STDOUT);
}
// _CRT_WARN _CRT_ERROR _CRTDBG_MODE_FILE _CRTDBG_MODE_DEBUG
//#ifdef _DEBUG
//   #ifndef DBG_NEW
//      #define DBG_NEW new ( _NORMAL_BLOCK , __FILE__ , __LINE__ )
//      #define new DBG_NEW
//   #endif
//#endif  // _DEBUG
#endif

#include <opencv2/opencv.hpp>
#include <time.h>

#include "yuv.h"

#include "ParameterEncDepth.h"
#include "SubMix.h"
#include "ReturnCode.h" 
#include "SubMixOptions.hpp"

namespace op= view_synthesis::depth_submix_options;
namespace fn= view_synthesis::fnum_tri;

using namespace std;

int main(int argc , char *argv[]) 
{
  
  // MemReport_Init();
  cv::setUseOptimized(true);

  int rc;
  fn::tri fnum;
  clock_t start, finish, first;
  first = start = clock();

  cout << "GVD Encoder, Version " << GVDENC_VERSION << " for Depth - MPEG 3DV, Feb 2014 NiCT." << endl << endl;
  
  if ( !op::cParameter.Init( argc, argv ) )
    return BAD_COMAMND_LINE;
  if ( UNKNOWN == op::checkOption(op::cParameter.getMode()) ) 
    return BAD_CONFIG_OPTIONS;

  rc= op::initParam(); if ( rc != NORMAL_END ) return rc; 
  rc= op::initFile(); if ( rc != NORMAL_END ) return rc; 
  rc= op::initSubMix(); if ( rc != NORMAL_END ) { op::endFile(); return rc; }

  finish= clock();
  printf( "\nInitialization: %.4f sec\n", (double)(finish - start) / CLOCKS_PER_SEC);
  start= finish;

  fnum.current= fnum.start= op::start;
  fnum.total= op::total;
  for(; fnum.current < fnum.start + fnum.total; fnum.current++) {
    cout << "frame number = " << fnum.current;
    if (op::convOneFrames(fnum) != NORMAL_END) {genABEND; break;}
    finish= clock();
    printf("->End (%.4f sec)\n", (double)(finish - start) / CLOCKS_PER_SEC);
    start= finish;
  }

  finish = clock();
  printf("Total: %.4f sec\n", ((double)(finish-first))/((double)CLOCKS_PER_SEC));

  op::endFile();
  
  // _CrtDumpMemoryLeaks();

  return 0;
}


