// View Synthesis Encoder for Depth
// Dev Enviroment:
//    Windows 7 64bit with Visual Studio 2010 C++ x64
//    OpenCV 2.4.2 64bit
// Original code: View Synthesis Reference Software (VSRS) 3.5
// Last Update: Feb, 2014.

#ifdef WIN32
#pragma warning(disable : 4996)
#endif

#include <stdio.h>
#include "yuv.h"
#include "SubMixOptions.h"
#include "SubMix.h"
#include "DepthSynthesis.h"

CSubMix::CSubMix() {
  pYuv1= NULL;
  pYuv2= NULL;
  pYuv3= NULL;
  pYuvBind= NULL;
  pYuvBindPre0= pYuvBindPre1= pYuvBindPre2= pYuvBindOut= NULL;
  pYuvOut= NULL;
  pYuvOutD= NULL;
  pYuvUnBind1= NULL;
  pYuvUnBind2= NULL;
//  if ( m_mode == HALF_QUARTER_STACK ) {
//    pYuvUnBind1a= NULL;
//    pYuvUnBind2a= NULL;
//  }
  pYuv1_work= NULL;
  pYuv2_work= NULL;
  pYuv1_temp= NULL;
  pYuv2_temp= NULL;

  m_img = NULL;
  m_imgD = NULL;
}



CSubMix::~CSubMix() {
  if (pYuv1 != NULL) { delete pYuv1; pYuv1= NULL; }
  if (pYuv2 != NULL) { delete pYuv2; pYuv2= NULL; }
  if (pYuv3 != NULL) { delete pYuv3; pYuv3= NULL; }

  if (pYuv1_work != NULL) { delete pYuv1_work; pYuv1_work= NULL; }
  if (pYuv1_temp != NULL) { delete pYuv1_temp; pYuv1_temp= NULL; }
  if (pYuv2_work != NULL) { delete pYuv2_work; pYuv2_work= NULL; }
  if (pYuv2_temp != NULL) { delete pYuv2_temp; pYuv2_temp= NULL; }

  if (pYuvBind != NULL) { delete pYuvBind; pYuvBind= NULL; }
  if (pYuvBindPre0 != NULL) { delete pYuvBindPre0; pYuvBindPre0= NULL; }
  if (pYuvBindPre1 != NULL) { delete pYuvBindPre1; pYuvBindPre1= NULL; }
  if (pYuvBindPre2 != NULL) { delete pYuvBindPre2; pYuvBindPre2= NULL; }
  if (pYuvBindOut != NULL) { delete pYuvBindOut; pYuvBindOut= NULL; }
  if (pYuvOut != NULL) { delete pYuvOut; pYuvOut= NULL; }
  if (pYuvOutD != NULL) { delete pYuvOutD; pYuvOutD= NULL; }
  if (pYuvUnBind1 != NULL) { delete pYuvUnBind1; pYuvUnBind1= NULL; }
  if (pYuvUnBind2 != NULL) { delete pYuvUnBind2; pYuvUnBind2= NULL; }
//  if ( m_mode == HALF_QUARTER_STACK ) {
//    if (pYuvUnBind1a != NULL) delete pYuvUnBind1a;
//    if (pYuvUnBind2a != NULL) delete pYuvUnBind2a;
//  }

  if (m_img != NULL) { cvReleaseImage(&m_img); m_img= NULL; }
  if (m_imgD != NULL) { cvReleaseImage(&m_imgD); m_imgD= NULL; }
}



bool CSubMix::Init(
      unsigned int mode, bool variparam,
      unsigned int width, unsigned int height, unsigned int width_bind, unsigned int height_bind,
      double near1, double far1, double near2, double far2, double near3, double far3,
      const char *strRefCamID1, const char *strRefCamID2, const char *strRefCamID3,
      const char *strVirCamID1, const char *strVirCamID2,
      double Mat_In_1[9], double Mat_Ex_1[9], double Mat_Trans_1[3],
      double Mat_In_2[9], double Mat_Ex_2[9], double Mat_Trans_2[3],
      double Mat_In_3[9], double Mat_Ex_3[9], double Mat_Trans_3[3],
      double Mat_In_V1[9], double Mat_Ex_V1[9], double Mat_Trans_V1[3],
      double Mat_In_V2[9], double Mat_Ex_V2[9], double Mat_Trans_V2[3]
) {
  m_mode= mode;
  m_width= width;
  m_height= height;
  m_width_bind= width_bind;
  m_height_bind= height_bind;

  pYuv1= new CIYuv(height, width, 420); // 400
  pYuv2= new CIYuv(height, width, 420); // 400
  if ( m_mode != QUARTER ) pYuv3= new CIYuv(height, width, 420); // 400
  pYuvBind= new CIYuv(height_bind, width_bind, 444);
  pYuvBindPre0= new CIYuv(height_bind, width_bind, 444);
  pYuvBindPre1= new CIYuv(height_bind, width_bind, 444);
  pYuvBindPre2= new CIYuv(height_bind, width_bind, 444);
  pYuvBindOut= new CIYuv(height_bind, width_bind, 444);
  if ( m_mode == QUARTER_STACK || m_mode == HALF_QUARTER_STACK ) {
    pYuvUnBind1= new CIYuv(height_bind / 2, width_bind, 444);
    pYuvUnBind2= new CIYuv(height_bind / 2, width_bind, 444);
    //if ( mode == HALF_QUARTER_STACK ) {
    //  pYuvUnBind1a= new CIYuv(height_bind / 2, width_bind, 444);
    //  pYuvUnBind2a= new CIYuv(height_bind / 2, width_bind, 444);
    //}
  }
  if ( m_mode == QUARTER ) { pYuvOut= new CIYuv(height_bind, width_bind, 420); pYuvOutD= new CIYuv(height_bind, width_bind, 420); }
  if ( m_mode == HALF_QUARTER_STACK ) { pYuvOut= new CIYuv(height_bind / 2, width_bind, 420); pYuvOutD= new CIYuv(height_bind / 2, width_bind, 420); }
  if ( m_mode == QUARTER_STACK ) { pYuvOut= new CIYuv(height_bind, width_bind, 420); pYuvOutD= new CIYuv(height_bind, width_bind, 420); }
// ATM13.1 out if ( m_mode == QUARTER_ODD ) { pYuvOut= new CIYuv(height_bind / 2, width_bind, 420); pYuvOutD= new CIYuv(height_bind / 2, width_bind, 420); }
  if ( m_mode == QUARTER_ODD ) { pYuvOut= new CIYuv(height_bind / 2, width_bind, 420); pYuvOutD= new CIYuv(height_bind, width_bind * 2, 420); }
  if ( m_mode == FULL_ODD ) { pYuvOut= new CIYuv(height_bind, width_bind, 420); pYuvOutD= new CIYuv(height_bind, width_bind, 420); }

  if ( pYuv1 == NULL || pYuv2 == NULL || pYuvBind == NULL || pYuvOut == NULL || pYuvOutD == NULL
  || (m_mode!=QUARTER && pYuv3==NULL) ) {
    if (pYuv1 != NULL) delete pYuv1;
    if (pYuv2 != NULL) delete pYuv2;
    if (pYuv3 != NULL) delete pYuv3;
    if (pYuvBind != NULL) delete pYuvBind;
    if (pYuvBindPre0 != NULL) delete pYuvBindPre0;
    if (pYuvBindPre1 != NULL) delete pYuvBindPre1;
    if (pYuvBindPre2 != NULL) delete pYuvBindPre2;
    if (pYuvBindOut != NULL) delete pYuvBindOut;
    if (pYuvOut != NULL) delete pYuvOut;
    if (pYuvOutD != NULL) delete pYuvOutD;
    return false;
  }

  pYuv1_work= new CIYuv(height, width, 444);
  if ( m_mode == QUARTER_STACK || m_mode == HALF_QUARTER_STACK ) pYuv2_work= new CIYuv(height, width, 444);
  pYuv1_temp= new CIYuv(height, width, 444);
  if ( m_mode == QUARTER_STACK || m_mode == HALF_QUARTER_STACK ) pYuv2_temp= new CIYuv(height, width, 444);

  if (pYuv1_work == NULL || pYuv1_temp == NULL
  || ((m_mode == QUARTER_STACK || m_mode == HALF_QUARTER_STACK) && (pYuv2_work == NULL || pYuv2_temp == NULL)) ) {
    if (pYuv1_work != NULL) delete pYuv1_work;
    if (pYuv2_work != NULL) delete pYuv2_work;
    if (pYuv1_temp != NULL) delete pYuv1_temp;
    if (pYuv2_temp != NULL) delete pYuv2_temp;
    return false;
  }

  m_img= cvCreateImage(cvSize(width_bind, height_bind), 8, 3);
  // for ATM11 m_imgD= cvCreateImage(cvSize(width_bind, height_bind), 8, 3);
  // for ATM13.1 Sep,2014
  m_imgD= cvCreateImage(cvSize(width_bind * 2, height_bind * 2), 8, 3);

  m_pDepthSyn1= new CDepthInterpolationGeneral();
  m_pDepthSyn1->SetWidth( width );
  m_pDepthSyn1->SetHeight( height );

  if ( mode == QUARTER || mode == HALF_QUARTER ) {
    if (
      !m_pDepthSyn1->InitLR(
                        width, height, variparam, m_mode,
                        near1, far1, near2, far2, near3, far3,
                        strRefCamID1, strRefCamID2, "", strVirCamID1,
                        Mat_In_1, Mat_Ex_1, Mat_Trans_1,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        NULL, NULL, NULL,
                        Mat_In_V1, Mat_Ex_V1, Mat_Trans_V1
                        )
    ) return false;
  }
  if ( m_mode == QUARTER_STACK || m_mode == HALF_QUARTER_STACK ) {
    if (
      !m_pDepthSyn1->InitLR(
                        width, height, variparam, m_mode,
                        near1, far1, near2, far2, near3, far3,
                        strRefCamID1, strRefCamID2, strRefCamID3, strVirCamID1,
                        Mat_In_1, Mat_Ex_1, Mat_Trans_1,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_3, Mat_Ex_3, Mat_Trans_3,
                        Mat_In_V1, Mat_Ex_V1, Mat_Trans_V1
                        )
    ) return false;
  }
  if ( m_mode == QUARTER_ODD || m_mode == QUARTER_ODD_L || m_mode == FULL_ODD ) {
    if (
      // LEFT, RIGHT, CENTER
      !m_pDepthSyn1->InitLR(
                        width, height, variparam, m_mode,
                        near1, far1, near3, far3, near2, far2,
                        strRefCamID1, strRefCamID3, strRefCamID2, strVirCamID1,
                        Mat_In_1, Mat_Ex_1, Mat_Trans_1,
                        Mat_In_3, Mat_Ex_3, Mat_Trans_3,
                        Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                        Mat_In_V1, Mat_Ex_V1, Mat_Trans_V1
                        )
    ) return false;
  }

  // 2nd phase for STACK type.
  if ( m_mode == QUARTER_STACK || m_mode == HALF_QUARTER_STACK ) {
    m_pDepthSyn2= new CDepthInterpolationGeneral();
    m_pDepthSyn2->SetWidth( width );
    m_pDepthSyn2->SetHeight( height );
    if (
      !m_pDepthSyn2->InitLR(
                      width, height, variparam, m_mode,
                      near2, far2, near3, far3, near1, far1,
                      strRefCamID2, strRefCamID3, strRefCamID1, strVirCamID2,
                      Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                      Mat_In_3, Mat_Ex_3, Mat_Trans_3,
                      Mat_In_1, Mat_Ex_1, Mat_Trans_1,
                      Mat_In_V2, Mat_Ex_V2, Mat_Trans_V2
                      )
    ) return false;
  }

  return true;
}



bool CSubMix::ReadOneFrame(
                FILE *fp1, FILE *fp2, FILE *fp3, fn::tri fnum,
                // for VariParam
                FILE *f_D1near, FILE *f_D1far, FILE *f_D1trans,
                FILE *f_D2near, FILE *f_D2far, FILE *f_D2trans,
                FILE *f_D3near, FILE *f_D3far, FILE *f_D3trans,
                FILE *f_V1near, FILE *f_V1far, FILE *f_V1trans,
                FILE *f_V2near, FILE *f_V2far, FILE *f_V2trans
) {
  BYTE** RefD1= pYuv1->Y;
  BYTE** RefD2= pYuv2->Y;
  BYTE** RefD3;
  if ( m_mode != QUARTER && m_mode != HALF_QUARTER ) RefD3= pYuv3->Y;
  double D1near, D1far, D1trans, D2near, D2far, D2trans, D3near, D3far, D3trans;
  double V1near, V1far, V1trans, V2near, V2far, V2trans;
  bool mode2view= false;
  mode2view= ( m_mode == QUARTER || m_mode == HALF_QUARTER );

  m_fnum= fnum;
  m_vparam= ( f_D1near != NULL );

  try {
    pYuv1->ReadOneFrame(fp1, fnum.current);
    pYuv2->ReadOneFrame(fp2, fnum.current);
    if ( m_mode != QUARTER && m_mode != HALF_QUARTER )
      pYuv3->ReadOneFrame(fp3, fnum.current);

    if ( m_vparam ) {
      #ifdef WIN32
      if ( fscanf_s(f_D1near, "%lf", &D1near) != 1 ) return false;
      if ( fscanf_s(f_D1far, "%lf", &D1far) != 1 ) return false;
      if ( fscanf_s(f_D1trans, "%lf", &D1trans) != 1 ) return false;
      if ( fscanf_s(f_D2near, "%lf", &D2near) != 1 ) return false;
      if ( fscanf_s(f_D2far, "%lf", &D2far) != 1 ) return false;
      if ( fscanf_s(f_D2trans, "%lf", &D2trans) != 1 ) return false;
      if ( fscanf_s(f_V1near, "%lf", &V1near) != 1 ) return false;
      if ( fscanf_s(f_V1far, "%lf", &V1far) != 1 ) return false;
      if ( fscanf_s(f_V1trans, "%lf", &V1trans) != 1 ) return false;
      if ( m_mode != QUARTER && m_mode != HALF_QUARTER ) {
        if ( fscanf_s(f_D3near, "%lf", &D3near) != 1 ) return false;
        if ( fscanf_s(f_D3far, "%lf", &D3far) != 1 ) return false;
        if ( fscanf_s(f_D3trans, "%lf", &D3trans) != 1 ) return false;
        if ( m_mode != QUARTER_ODD && m_mode != QUARTER_ODD_L && m_mode != FULL_ODD ) {
          if ( fscanf_s(f_V2near, "%lf", &V2near) != 1 ) return false;
          if ( fscanf_s(f_V2far, "%lf", &V2far) != 1 ) return false;
          if ( fscanf_s(f_V2trans, "%lf", &V2trans) != 1 ) return false;
        }
      } else {
        D3near= D3far= D3trans= 0;
      }
      #else
      if ( fscanf(f_D1near, "%lf", &D1near) != 1 ) return false;
      if ( fscanf(f_D1far, "%lf", &D1far) != 1 ) return false;
      if ( fscanf(f_D1trans, "%lf", &D1trans) != 1 ) return false;
      if ( fscanf(f_D2near, "%lf", &D2near) != 1 ) return false;
      if ( fscanf(f_D2far, "%lf", &D2far) != 1 ) return false;
      if ( fscanf(f_D2trans, "%lf", &D2trans) != 1 ) return false;
      if ( fscanf(f_V1near, "%lf", &V1near) != 1 ) return false;
      if ( fscanf(f_V1far, "%lf", &V1far) != 1 ) return false;
      if ( fscanf(f_V1trans, "%lf", &V1trans) != 1 ) return false;
      if ( m_mode != QUARTER && m_mode != HALF_QUARTER ) {
        if ( fscanf(f_D3near, "%lf", &D3near) != 1 ) return false;
        if ( fscanf(f_D3far, "%lf", &D3far) != 1 ) return false;
        if ( fscanf(f_D3trans, "%lf", &D3trans) != 1 ) return false;
        if ( m_mode != QUARTER_ODD && m_mode != QUARTER_ODD_L && m_mode != FULL_ODD ) {
          if ( fscanf(f_V2near, "%lf", &V2near) != 1 ) return false;
          if ( fscanf(f_V2far, "%lf", &V2far) != 1 ) return false;
          if ( fscanf(f_V2trans, "%lf", &V2trans) != 1 ) return false;
        }
      } else {
        D3near= D3far= D3trans= 0;
      }
      #endif

      if ( m_mode != QUARTER_ODD && m_mode != QUARTER_ODD_L && m_mode != FULL_ODD ) {
        if (!m_pDepthSyn1->DoOneFrameDepth(
                            RefD1, RefD2, RefD3, pYuv1_work,
                            D1near, D1far, D1trans, D2near, D2far, D2trans, D3near, D3far, D3trans,
                            V1near, V1far, V1trans, fnum, LEFT, mode2view
        ) ) return false;
      } else {
        if (!m_pDepthSyn1->DoOneFrameDepth(
                            RefD1, RefD3, RefD2, pYuv1_work,
                            D1near, D1far, D1trans, D3near, D3far, D3trans, D2near, D2far, D2trans,
                            V1near, V1far, V1trans, fnum, CENTER, false
        ) ) return false;
      }

      if ( m_mode == QUARTER_STACK || m_mode == HALF_QUARTER_STACK ) {
        if (!m_pDepthSyn2->DoOneFrameDepth(
                            RefD2, RefD3, RefD1, pYuv2_work,
                            D2near, D2far, D2trans, D3near, D3far, D3trans, D1near, D1far, D1trans,
                            V2near, V2far, V2trans, fnum, RIGHT, mode2view
        ) ) return false;
      }

    } else {
      if ( m_mode != QUARTER_ODD && m_mode != QUARTER_ODD_L && m_mode != FULL_ODD ) {
        if (!m_pDepthSyn1->DoOneFrameDepth(
                              RefD1, RefD2, RefD3, pYuv1_work, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, fnum, LEFT, mode2view
        ) ) return false;
      } else {
        if (!m_pDepthSyn1->DoOneFrameDepth(
                              RefD1, RefD3, RefD2, pYuv1_work, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, fnum, CENTER, false
        ) ) return false;
      }

      if (m_mode == QUARTER_STACK || m_mode == HALF_QUARTER_STACK) {
        if (!m_pDepthSyn2->DoOneFrameDepth(
                            RefD2, RefD3, RefD1, pYuv2_work, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, fnum, RIGHT, mode2view
        ) ) return false;
      }
    }

    pYuv1_temp->setData444_inIYUV(pYuv1_work);
    // pYuv1_work->LPF_V(pYuv1_temp, false);
    // pYuv1_temp->LPF_H_DSimple(pYuv1_work, false);
    if ( m_mode == QUARTER_STACK || m_mode == HALF_QUARTER_STACK ) {
      pYuv2_temp->setData444_inIYUV(pYuv2_work);
      // pYuv2_work->LPF_V(pYuv2_temp, false);
      // pYuv2_temp->LPF_H_DSimple(pYuv2_work, false);
    }

  } catch (...) {
    return false;
  }
  return true;
}



bool CSubMix::DoSubMix(FILE *fp, FILE *fpD) {
  bool f_1st= ( m_fnum.current == m_fnum.start );
  bool f_2nd= ( m_fnum.current == m_fnum.start + 1 );
  bool f_end= ( m_fnum.current == m_fnum.total - 1 );
  bool RC;

  unsigned int height_bind;
  if ( m_mode == QUARTER )
    height_bind= m_height_bind;
  if ( m_mode == QUARTER_STACK )
    height_bind= m_height_bind;
  if ( m_mode == HALF_QUARTER_STACK )
    height_bind= m_height_bind / 2;
  if ( m_mode == QUARTER_ODD )
    height_bind= m_height_bind;
  if ( m_mode == FULL_ODD )
    height_bind= m_height; //_bind;

  if ( m_mode == FULL_ODD ) {
    return writeYUV( pYuv1_temp, fp, fpD, height_bind );
  }

  if ( m_mode == QUARTER || m_mode == QUARTER_ODD ) { // for 2view or 3view Single Depth /////////
    for ( unsigned int h= 0; h < m_height_bind; h++ )
      for ( unsigned int w= 0; w < m_width_bind; w++ )
        pYuvBind->Y[h][w]= pYuv1_temp->Y[h*2][w*2];

  } else { // QUARTER_STACK, for 3view /////////////////////////////////////////////
    for ( unsigned int h= 0; h < height_bind; h++ )
      for ( unsigned int w= 0; w < m_width_bind; w++ )
        pYuvUnBind1->Y[h][w]= pYuv1_temp->Y[h*2][w*2];
        // pYuvUnBind1a->LPF_V(pYuvUnBind1); // at 444 Quarter to 444 Quarter

    for ( unsigned int h= 0; h < height_bind; h++ )
      for ( unsigned int w= 0; w < m_width_bind; w++ )
        pYuvUnBind2->Y[h][w]= pYuv2_temp->Y[h*2][w*2];
        // pYuvUnBind2a->LPF_V(pYuvUnBind2); // at 444 Quarter to 444 Quarter

    // Jan23,2012 Quarter size to Half-Quarter size.
    // Apr26,2012 Quarter Stack to Half size.
    for ( unsigned int h= 0; h < height_bind/2; h++ )
      for ( unsigned int w= 0; w < m_width_bind; w++ )
        if ( m_mode == QUARTER_STACK )
          pYuvBind->Y[h][w]= pYuvUnBind1->Y[h][w];
        else
          pYuvBind->Y[h][w]= pYuvUnBind1->Y[h*2][w];

    for ( unsigned int h= height_bind/2, h2= 0; h < height_bind; h++, h2++ )
      for ( unsigned int w= 0; w < m_width_bind; w++ )
        if ( m_mode == QUARTER_STACK )
          pYuvBind->Y[h][w]= pYuvUnBind2->Y[h2][w];
        else
          pYuvBind->Y[h][w]= pYuvUnBind2->Y[h2*2][w];
  }

  // if ( m_vparam )
     return writeYUV( pYuvBind, fp, fpD, height_bind );
}



bool CSubMix::writeYUV( CIYuv *yuvSrc, FILE *fp, FILE *fpD, unsigned int height_bind ) {
  cvZero(m_img);
  cvZero(m_imgD);

  // Convert to linear address.
  unsigned int ptv;
  for ( unsigned int h= 0; h < height_bind; h++ ) {
    for ( unsigned int w= 0; w < m_width_bind; w++ ) {
      ptv= w + h * m_width_bind;
      m_img->imageData[ptv * 3 + 0]= yuvSrc->Y[h][w];
      m_img->imageData[ptv * 3 + 1]= (char)128;
      m_img->imageData[ptv * 3 + 2]= (char)128;
      /* ATM13.1 out
      m_imgD->imageData[ptv * 3 + 0]= (char)128;
      m_imgD->imageData[ptv * 3 + 1]= (char)128;
      m_imgD->imageData[ptv * 3 + 2]= (char)128;
      */
    }
  }
  // for
  for ( unsigned int h= 0; h < height_bind * 2; h++ ) {
    for ( unsigned int w= 0; w < m_width_bind * 2; w++ ) {
      ptv= w + h * m_width_bind * 2;
      m_imgD->imageData[ptv * 3 + 0]= (char)128;
      m_imgD->imageData[ptv * 3 + 1]= (char)128;
      m_imgD->imageData[ptv * 3 + 2]= (char)128;
    }
  }

  pYuvOut->setDataFromImgYUV( m_img ); // Jan23,2012 Quarter size to Half-Quarter size.
  if (!pYuvOut->WriteOneFrame(fp)) return false;
  pYuvOutD->setDataFromImgYUV( m_imgD );
  if (!pYuvOutD->WriteOneFrame(fpD)) return false;
  return true;
}



#define DEPTH_TH 3
void CSubMix::smoothYUV( CIYuv *yuvSrc2, CIYuv *yuvSrc1, CIYuv *yuvSrc0, CIYuv *yuvDest, unsigned int height_bind ) {
  for ( unsigned int h= 0; h < height_bind; h++ )
    for ( unsigned int w= 0; w < m_width_bind; w++ )
      if ( abs(yuvSrc2->Y[h][w] - yuvSrc0->Y[h][w]) <= DEPTH_TH )
        yuvDest->Y[h][w]= ( yuvSrc2->Y[h][w] + yuvSrc0->Y[h][w] + 1 ) >> 1;
      else
        yuvDest->Y[h][w]= yuvSrc1->Y[h][w];
}