// View Synthesis Encoder for Residuals
// Dev Enviroment:
//    Windows 8.1 64bit with Visual Studio 2013(VS12) C++ x64
//    OpenCV 2.4.9 64bit
// Original code: View Synthesis Reference Software (VSRS) 3.5
// Last Update: June, 2014.
#pragma once
#include <GVDsetup.h>
#ifdef WIN32
#pragma warning(disable:4996)
#pragma warning(disable:4244)
#pragma warning(disable:4819)
#endif

#include <iostream>
#include <fstream>
#include "yuv.h"
#include <string>

#include "DEBUGOUT.h"

typedef unsigned char uchar;
//using namespace std;

#define NORMAL 0
#define HOLESHIFT 1

#define LEFT 1
#define RIGHT 2
#define CENTER 3

#define INIT 1
#define RESET 2

#define ONESET 24

#define C1  0
#define C2  1
#define C3  2

#define D1  0
#define D2  1

class CResidualsInterpolation
{
public:
  CResidualsInterpolation ();
  virtual ~CResidualsInterpolation ();

  bool InitOneSide (
    unsigned int width, unsigned int height, bool variparam,
    double near1, double far1, double near2, double far2,
    const char *RefCamID1, const char *RefCamID2, const char *VirCamID1, const char *AddCamID1, const char *AddCamID2,
    const unsigned int Side, const unsigned int mode,
    double Mat_In_1[9], double Mat_Ex_1[9], double Mat_Trans_1[3],
    double Mat_In_2[9], double Mat_Ex_2[9], double Mat_Trans_2[3],
    double Mat_In_V1[9], double Mat_Ex_V1[9], double Mat_Trans_V1[3],
    double Mat_In_A1[9], double Mat_Ex_A1[9], double Mat_Trans_A1[3],
    double Mat_In_A2[9], double Mat_Ex_A2[9], double Mat_Trans_A2[3]
  );

  bool Init (
    unsigned int width, unsigned int height, bool variparam,
    double near, double far,
    const unsigned int Side,
    double Mat_In_Ref[9], double Mat_Ex_Ref[9], double Mat_Trans_Ref[3],
    double Mat_In_Vir[9], double Mat_Ex_Vir[9], double Mat_Trans_Vir[3]
  );

  void    xReleaseMemory        ();

  double    getBaselineDistance      ()  { return m_dBaselineDistance; }
  double    getLeftBaselineDistance      ()  { return LeftBaselineDistance; }
  double    getRightBaselineDistance      ()  { return RightBaselineDistance; }
  IplImage*  getHolePixels        ()  { return m_imgHoles; }
  IplImage*  getSynthesizedPixels    ()  { return m_imgSuccessSynthesis; }
  IplImage*  getUnstablePixels      ()  { return m_imgMask[0]; }
  IplImage*  getVirtualImage        ()  { return m_imgVirtualImage; }
  IplImage*  getVirtualDepthMap      ()  { return m_imgVirtualDepth; }
  IplImage*  getWorkDepthMap      ()  { return m_imgWorkDepth; }
  BYTE**    getVirtualDepth        ()  { return m_pVirtualDepth; }
  BYTE**    getNonHoles          ()  { return m_pSuccessSynthesis; }

  void SetWidth (int sWidth ) { Width  = sWidth;  }
  void SetHeight(int sHeight) { Height = sHeight; }
  void SetDepthType(int sDepthType) {DepthType = sDepthType; }
  void SetViewBlending(int sViewBlending) { ViewBlending = sViewBlending; }

  void setImgHoles(IplImage* img) { cvCopy(img, m_imgHoles); }
  void setDepthMap(IplImage* img) { cvCopy(img, m_imgVirtualDepth); }

  void depthInpaint( IplImage* map, IplImage* hole );

  int DoOneFrame ( unsigned int fnum );

  BYTE*** View= NULL;
  BYTE*** CenterView= NULL;
  BYTE** Depth= NULL;
  CIYuv *SynYuvBuffer;
  CIYuv *AreaBuffer;
  double near_C1= 0.0;
  double far_C1= 0.0;
  double trans_C1= 0.0;
  double near_C2= 0.0;
  double far_C2= 0.0;
  double trans_C2= 0.0;
  double near_V1= 0.0;
  double far_V1= 0.0;
  double trans_V1= 0.0;
  double near_A1= 0.0;
  double far_A1= 0.0;
  double trans_A1= 0.0;
  double near_A2= 0.0;
  double far_A2= 0.0;
  double trans_A2= 0.0;
  // unsigned int fnum= 0;
  unsigned int side= 0;
  unsigned int mode= 0;


  IplImage* getImgSynthesizedViewLeft  ();
  IplImage* getImgSynthesizedViewRight  ();

  double getAvgColor(int i) { return avgColor.val[i]; }
  void setAvgColor ( CvScalar color ) { avgColor= color; }

private:
  void convertCameraParam      (CvMat *exMat_dst, CvMat *exMat_src);

  // Depth smooth
  CResidualsInterpolation*  m_pD1;
  // m_pcViewHoleShift
  CResidualsInterpolation*  m_pS1;
  CResidualsInterpolation*  m_pS1a;
  CResidualsInterpolation*  m_pS2;
  CResidualsInterpolation*  m_pS3;
  CResidualsInterpolation*  m_pS2d;
  CResidualsInterpolation*  m_pS3d;
  // m_pcViewHoleLeft
  CResidualsInterpolation*  m_pL1;
  CResidualsInterpolation*  m_pL1a;
  CResidualsInterpolation*  m_pL2;
  CResidualsInterpolation*  m_pL3;
  // m_pcViewHoleRight
  CResidualsInterpolation*  m_pR1;
  CResidualsInterpolation*  m_pR1a;


  bool init_camera_param (
          double Mat_In_Ref[9], double Mat_Ex_Ref[9], double Mat_Trans_Ref[3],
          double Mat_In_Vir[9], double Mat_Ex_Vir[9], double Mat_Trans_Vir[3],
          CvMat *mat_in[2], CvMat *mat_ex_c2w[2], CvMat *mat_proj[2]);

  bool init_3Dwarp( double Z_near, double Z_far );

  bool reset_3Dwarp (
          double near_C, double far_C, double trans_C,
          double near_V, double far_V, double trans_V );

  void image2world_with_z (
          CvMat *mat_Rc2w_invIN_from, CvMat *matEX_c2w_from,
          CvMat *image, CvMat *world );

  void makeHomography (
          CvMat *matH_F2T[256],
          CvMat *matH_T2F[256],
          double adTable[256],
          CvMat *matIN_from, CvMat *matEX_c2w_from, CvMat *matProj_to,
          unsigned int work= RESET );

  bool setView (BYTE ***src);
  bool setDepth(BYTE **pDepthMap);
  bool depthsynthesis_reverse_3Dwarp (BYTE **pDepthMap);
  bool depthShift(BYTE **pDepthMap, unsigned int side, int MODE= NORMAL);

  double m_dWeightLeft;
  double m_dWeightRight;
  double WeightLeft;
  double WeightRight;

  unsigned int m_uiViewBlending;
  unsigned int ViewBlending;

  unsigned int m_uiWidth;
  unsigned int m_uiHeight;
  unsigned int m_uiPicsize;
  unsigned int m_uiPrecision;

  unsigned int m_uiSide;
  unsigned int m_uiResidualsSide;
  unsigned int m_uiResidualsWidth;
  double m_dMaxDisp;
  double m_dispA;
  double m_dispB;

  std::string CameraParameterFile;
  std::string LeftCameraName;
  std::string RightCameraName;
  std::string VirtualCameraName;

  unsigned int Width;
  unsigned int Height;
  unsigned int Picsize;
  unsigned int Precision;
  unsigned int DepthType;
  bool VariParam;

  // for VeriParam
  double Mat_Ex_C[9];
  double Mat_Ex_V[9];
  double Mat_In_C[9];
  double Mat_In_V[9];
  double Mat_Trans_C[3];
  double Mat_Trans_V[3];

  IplImage* m_imgBlended;          //!> Blended image
  IplImage* m_imgInterpolatedView; //!> The final image buffer to be output
  IplImage* getImgInterpolatedView () { return m_imgInterpolatedView; }
  IplImage* m_imgInterpolatedView2;

  double m_dBaselineDistance;
  double LeftBaselineDistance;
  double RightBaselineDistance;

  unsigned int m_ResidualsSide;
  // unsigned int m_ResidualsWidth;

  BYTE** m_pVirtualDepth;
  BYTE** m_pWorkDepth;
  BYTE** m_pSuccessSynthesis;
  BYTE** m_pHole;
  BYTE** m_pHole1;
  // BYTE** m_pWDepth1;
  // BYTE** m_pWDepth2;

  IplImage* m_imgVirtualImage;
  IplImage* m_imgVirtualDepth;
  IplImage* m_imgWorkDepth;
  IplImage* m_imgSuccessSynthesis;
  IplImage* m_imgHoles;
  IplImage* m_imgMask[5];
  IplImage* m_imgTemp[5];

  // Jul1-6,2011 NiCT
  IplImage* m_depthS1;
  IplImage* m_depthS2;
  IplImage* m_depthS3;
  IplImage* m_depthS4;
  IplImage* m_depthR1;
  IplImage* m_depthL1;
  IplImage* m_depthL1a;

  IplImage* m_depthBlended1;
  IplImage* m_depthBlended2;
  IplImage* m_depthHoles1;
  IplImage* m_depthHoles2;
  IplImage* m_depthSuccess;
  void ToFilterDepth(IplImage* pDepthSRC, IplImage* pDepthDST, IplImage* pDepthNEGSRC, IplImage* pDepthNEGDST, unsigned int fsize);

  CvMat* m_matH_R2V[256];
  CvMat* m_matH_V2R[256];

  // for VeriParam
  CvMat* m_Mat_in[2]; //intrinsic parameter of left camera 3x3 matrix
  CvMat* m_Mat_ex_c2w[2]; //extrinsic parameter of left camera 3x4 matrix
  CvMat* m_Mat_proj_w2i[2]; // projection matrix from world to image

  // for Residuals mask
  CvScalar avgColor;
};
