#pragma once
// View Synthesis Encoder for Residuals
// Last Update: Sep, 2013.

// for OpenCV 2.1.0 include <opencv/highgui.h>
#include <opencv2/opencv.hpp>
#include "ResidualSynthesis.h"

#include "DEBUGOUT.h"

class CSubSepMix {

public:
  CSubSepMix();
  virtual ~CSubSepMix();

  bool Init(
        unsigned int mode, unsigned int numView, bool variparam,
        unsigned int width, unsigned int height, unsigned int width_bind, unsigned int height_bind,
        double near1, double far1, double near2, double far2, double near3, double far3,
        const char *strRefCamID1, const char *strRefCamID2, const char *strRefCamID3,
        const char *strVirCamID1, const char *strVirCamID2,
        const char *strAddCamID1, const char *strAddCamID2, const char *strAddCamID3, const char *strAddCamID4,
        double Mat_In_1[9], double Mat_Ex_1[9], double Mat_Trans_1[3],
        double Mat_In_2[9], double Mat_Ex_2[9], double Mat_Trans_2[3],
        double Mat_In_3[9], double Mat_Ex_3[9], double Mat_Trans_3[3],
        double Mat_In_V1[9], double Mat_Ex_V1[9], double Mat_Trans_V1[3],
        double Mat_In_V2[9], double Mat_Ex_V2[9], double Mat_Trans_V2[3],
        double Mat_In_A1[9], double Mat_Ex_A1[9], double Mat_Trans_A1[3],
        double Mat_In_A2[9], double Mat_Ex_A2[9], double Mat_Trans_A2[3],
        double Mat_In_A3[9], double Mat_Ex_A3[9], double Mat_Trans_A3[3],
        double Mat_In_A4[9], double Mat_Ex_A4[9], double Mat_Trans_A4[3]
  );
  bool ReadOneFrame(
        FILE *fp1, FILE *fp2, FILE *fp3, FILE *fpD, unsigned int fnum, unsigned int fnumD,
        // for VariParam
        FILE *f_D1near, FILE *f_D1far, FILE *f_D1trans,
        FILE *f_D2near, FILE *f_D2far, FILE *f_D2trans,
        FILE *f_D3near, FILE *f_D3far, FILE *f_D3trans,
        FILE *f_V1near, FILE *f_V1far, FILE *f_V1trans,
        FILE *f_V2near, FILE *f_V2far, FILE *f_V2trans,
        FILE *f_A1near, FILE *f_A1far, FILE *f_A1trans,
        FILE *f_A2near, FILE *f_A2far, FILE *f_A2trans,
        FILE *f_A3near, FILE *f_A3far, FILE *f_A3trans,
        FILE *f_A4near, FILE *f_A4far, FILE *f_A4trans
  );
#ifndef DEBUGOUT
  bool DoSubMix( FILE *fp );
#else
  bool DoSubMix( FILE *fp1, FILE *fp2 );
#endif
  bool DoResidualsInterpolation();


private:
  CIYuv* pYuv1;
  CIYuv* pYuv2;
  CIYuv* pYuv3;
  CIYuv* pYuvDBind;
  CIYuv* pYuvDBindPRE;
  CIYuv* pYuvD1;
  CIYuv* pYuvD2;
  CIYuv* pYuvD1_temp;
  CIYuv* pYuvD2_temp;
  CIYuv* pYuvD1_half;
  CIYuv* pYuvD2_half;

  CIYuv* pYuvBind;
  CIYuv* pYuvBind2;
  CIYuv* pYuvOut;
  CIYuv* pYuvUnBind1;
  CIYuv* pYuvUnBind2;

  CIYuv* pAreaBind;
  CIYuv* pAreaUnBind1;
  CIYuv* pAreaUnBind2;
  CIYuv* pArea1;
  CIYuv* pArea3;

  CIYuv* pYuv1_work;
  CIYuv* pYuv2_work;
  CIYuv* pYuv3_work;
  CIYuv* pYuvTemp[3];


  CIYuv* pYuvDBind_work;

  IplImage*  m_img;

  unsigned int m_mode;
  unsigned int m_numView;
  unsigned int m_numDepth;
  bool m_globalDepth;
  bool m_fullSize;
  unsigned int m_width;
  unsigned int m_height;
  unsigned int m_width_bind;
  unsigned int m_height_bind;
  bool m_offset;

  CResidualsInterpolation* m_pResidualsSyn[2];

#ifdef DEBUGOUT
  CIYuv* pYuvDebugOut1; // for fullsize out.
  CIYuv* pYuvDebugOut2; // for fullsize out.
#endif
};
