// NICT GVD: View Synthesis Encoder for Residuals
// Dev Enviroment:
//    Windows 8.1 64bit with Visual Studio 2013(VS12) C++ x64
//    OpenCV 2.4.9 64bit
// Original code: View Synthesis Reference Software (VSRS) 3.5
// Designer: Senoh, NiCT, Tokyo. Japan.
// Last Update: June, 2014.
#define GVDENC_MAJOR_VERSION 4
#define GVDENC_MINOR_VERSION 4
#define GVDENC_VERSION CVAUX_STR(GVDENC_MAJOR_VERSION) "." CVAUX_STR(GVDENC_MINOR_VERSION)

#define _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#ifdef WIN32
#include <crtdbg.h>
static void MemReport_Init() {
  _CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF ); //| _CRTDBG_DELAY_FREE_MEM_DF );
  _CrtSetReportMode(_CRT_ERROR, _CRTDBG_MODE_FILE);
  //_CrtSetReportMode(_CRT_WARN, _CRTDBG_MODE_FILE);_CRT_ERROR
  _CrtSetReportFile(_CRT_ERROR, _CRTDBG_FILE_STDOUT);
}
#ifdef _DEBUG
   #ifndef DBG_NEW
      #define DBG_NEW new ( _NORMAL_BLOCK , __FILE__ , __LINE__ )
      #define new DBG_NEW
   #endif
#endif
#endif

#include <opencv2/opencv.hpp>
#include <time.h>

#include "yuv.h"
#include "ParameterEncResidual.h"
#include "SubSepMix.h"
#include "ReturnCode.h"
#include "SubSepOptions.hpp"

namespace op= view_synthesis::residuals_subsep_options;

using namespace std;

int main(int argc , char *argv[])
{
  // MemReport_Init();
  cv::setUseOptimized(true);

  int rc;

  clock_t start, finish, first;
  first = start = clock();

  cout << "GVD Encode, Version " << GVDENC_VERSION << " for Residuals - MPEG 3DV, June 2014 NiCT." << endl << endl;

  if ( NORMAL_END != op::cParameter.Init( argc, argv ) )
    return BAD_COMAMND_LINE;
  if ( UNKNOWN == op::checkOption(op::cParameter.getMode()) )
    return BAD_CONFIG_OPTIONS;
  if ( UNKNOWN == op::checkOffset(op::cParameter.getOffset()) )
    return BAD_CONFIG_OPTIONS;

  rc= op::initParam(); if ( rc != NORMAL_END ) return rc;
  rc= op::initFile(); if ( rc != NORMAL_END ) return rc;
  rc= op::initSubMix(); if ( rc != NORMAL_END ) { op::endFile(); return rc; }

  finish= clock();
  printf( "\nInitialization: %.4f sec\n", (double)(finish - start) / CLOCKS_PER_SEC);
  start= finish;

  for(unsigned int fnum= op::start, fnumD= 0; fnum < op::start + op::total; fnum++) {
    cout << "frame number = " << fnum;
    if (!op::Offset) fnumD= fnum;
    if (op::convOneFrames(fnum, fnumD) != NORMAL_END) {genABEND; break;}
    finish= clock();
    printf("->End (%.4f sec)\n", (double)(finish - start) / CLOCKS_PER_SEC);
    start= finish;
    if (op::Offset) fnumD++;
  }

  finish = clock();
  printf("Total: %.4f sec\n", ((double)(finish-first))/((double)CLOCKS_PER_SEC));

  op::endFile();

  return 0;
}


