// View Synthesis Encoder for Residuals
// Dev Enviroment:
//    Windows 8.1 64bit with Visual Studio 2013(VS12) C++ x64
//    OpenCV 2.4.9 64bit
// Original code: View Synthesis Reference Software (VSRS) 3.5
// Designer: Senoh, NiCT, Tokyo. Japan.
// Last Update: June, 2014.

#ifdef WIN32
#pragma warning(disable : 4996)
#endif

#include "yuv.h"
#include "SubMixOptions.h"
#include "SubSepMix.h"
#include "ResidualSynthesis.h"
#include <omp.h>

CSubSepMix::CSubSepMix() {
  #ifdef PTRACE
    printf("TRACE: CSubSepMix::CSubSepMix()\n");
  #endif

  pYuv1= NULL;
  pYuv2= NULL;
  pYuv3= NULL;

  pYuvDBind= NULL;
  pYuvDBindPRE= NULL;
  pYuvD1= NULL;
  pYuvD2= NULL;
  pYuvD1_temp= NULL;
  pYuvD2_temp= NULL;
  pYuvD1_half= NULL;
  pYuvD2_half= NULL;

  pYuvBind= NULL;
  pYuvOut= NULL;
  pYuvUnBind1= NULL;
  pYuvUnBind2= NULL;

  pYuv1_work= NULL;
  pYuv2_work= NULL;
  pYuv3_work= NULL;

  pArea1= pArea3= NULL;
  pAreaUnBind1= pAreaUnBind2= NULL;
  pAreaBind= NULL;

  pYuvTemp[C1]= NULL;
  pYuvTemp[C2]= NULL;
  pYuvTemp[C3]= NULL;

  pYuvDBind_work= NULL;

  m_img = NULL;

#ifdef DEBUGOUT
  pYuvDebugOut1= NULL;
  pYuvDebugOut2= NULL;
#endif
}



CSubSepMix::~CSubSepMix() {
  #ifdef PTRACE
    printf("TRACE: CSubSepMix::~CSubSepMix()\n");
  #endif
  if (pYuv1 != NULL) { delete pYuv1; pYuv1= NULL; }
  if (pYuv2 != NULL) { delete pYuv2; pYuv2= NULL; }
  if (pYuv3 != NULL) { delete pYuv3; pYuv3= NULL; }

  if (pYuv1_work != NULL) { delete pYuv1_work; pYuv1_work= NULL; }
  if (pYuv2_work != NULL) { delete pYuv2_work; pYuv2_work= NULL; }
  if (pYuv3_work != NULL) { delete pYuv3_work; pYuv3_work= NULL; }

  if (pYuvTemp[C1] != NULL) { delete pYuvTemp[C1]; pYuvTemp[C1]= NULL; }
  if (pYuvTemp[C2] != NULL) { delete pYuvTemp[C2]; pYuvTemp[C2]= NULL; }
  if (pYuvTemp[C3] != NULL) { delete pYuvTemp[C3]; pYuvTemp[C3]= NULL; }

  if (pYuvDBind != NULL) delete pYuvDBind;
  if (pYuvDBindPRE != NULL) delete pYuvDBindPRE;
  if (pYuvDBind_work != NULL) delete pYuvDBind_work;
  if (pYuvD1 != NULL) delete pYuvD1;
  if (pYuvD2 != NULL) delete pYuvD2;
  if (pYuvOut != NULL) { delete pYuvOut; pYuvOut= NULL; }
  if (pYuvD1_temp != NULL) { delete pYuvD1_temp; pYuvD1_temp= NULL; }
  if (pYuvD2_temp != NULL) { delete pYuvD2_temp; pYuvD2_temp= NULL; }
  if (pYuvD1_half != NULL) { delete pYuvD1_half; pYuvD1_half= NULL; }
  if (pYuvD2_half != NULL) { delete pYuvD2_half; pYuvD2_half= NULL; }

  if (pYuvBind != NULL) { delete pYuvBind; pYuvBind= NULL; }
  if (pYuvUnBind1 != NULL) delete pYuvUnBind1;
  if (pYuvUnBind2 != NULL) delete pYuvUnBind2;

  if (pArea1 != NULL) { delete pArea1; pArea1= NULL; }
  if (pArea3 != NULL) { delete pArea3; pArea3= NULL; }
  if (pAreaBind != NULL) { delete pAreaBind; pAreaBind= NULL; }
  if (pAreaUnBind1 != NULL) { delete pAreaUnBind1; pAreaUnBind1= NULL; }
  if (pAreaUnBind2 != NULL) { delete pAreaUnBind2; pAreaUnBind2= NULL; }

  if (m_img != NULL) { cvReleaseImage(&m_img); m_img= NULL; }

  pYuvDBind= NULL;
  pYuvDBindPRE= NULL;
  pYuvD1= NULL;
  pYuvD2= NULL;

  pYuvDBind_work= NULL;

#ifdef DEBUGOUT
  if (pYuvDebugOut1 != NULL) { delete pYuvDebugOut1; pYuvDebugOut1= NULL; }
  if (pYuvDebugOut2 != NULL) { delete pYuvDebugOut2; pYuvDebugOut2= NULL; }
#endif
}



bool CSubSepMix::Init(
      unsigned int mode, unsigned int numView, bool variparam,
      unsigned int width, unsigned int height, unsigned int width_bind, unsigned int height_bind,
      double near1, double far1, double near2, double far2, double near3, double far3,  // 暫定処置
      const char *RefCamID1, const char *RefCamID2, const char *RefCamID3,
      const char *VirCamID1, const char *VirCamID2,
      const char *AddCamID1, const char *AddCamID2, const char *AddCamID3, const char *AddCamID4,
      double Mat_In_1[9], double Mat_Ex_1[9], double Mat_Trans_1[3],
      double Mat_In_2[9], double Mat_Ex_2[9], double Mat_Trans_2[3],
      double Mat_In_3[9], double Mat_Ex_3[9], double Mat_Trans_3[3],
      double Mat_In_V1[9], double Mat_Ex_V1[9], double Mat_Trans_V1[3],
      double Mat_In_V2[9], double Mat_Ex_V2[9], double Mat_Trans_V2[3],
      double Mat_In_A1[9], double Mat_Ex_A1[9], double Mat_Trans_A1[3],
      double Mat_In_A2[9], double Mat_Ex_A2[9], double Mat_Trans_A2[3],
      double Mat_In_A3[9], double Mat_Ex_A3[9], double Mat_Trans_A3[3],
      double Mat_In_A4[9], double Mat_Ex_A4[9], double Mat_Trans_A4[3]
) {
  m_mode= mode;
  m_width= width;
  m_height= height;
  m_width_bind= width_bind;
  m_height_bind= height_bind;

  m_numView= numView;
  if ( mode == QUARTER_STACK || mode == HALF_QUARTER_STACK ) m_numDepth= DEPTH_2; else m_numDepth= DEPTH_1;
  m_globalDepth= ( mode == QUARTER_ODD || mode == QUARTER_ODD_L || m_mode == FULL_ODD_L );
  m_fullSize= ( mode == QUARTER_ODD || mode == QUARTER_ODD_L || m_mode == FULL_ODD_L );

  pYuv1= new CIYuv(height, width, 420);
  pYuv2= new CIYuv(height, width, 420);
  if ( m_numView == VIEW_3 ) pYuv3= new CIYuv(height, width, 420);

  pYuvD1= new CIYuv(height, width, 444);
  if ( m_numDepth == DEPTH_2 ) {
    pYuvD2= new CIYuv(height, width, 444);
  }

  if ( m_mode == QUARTER ) {
    pYuvDBind= new CIYuv(height_bind, width_bind, 420);
    pYuvDBindPRE= new CIYuv(height_bind, width_bind, 420);
    pYuvDBind_work= new CIYuv(height_bind, width_bind, 444);
    pYuvOut= new CIYuv(height_bind, width_bind, 420);
    pYuvDBind_work= new CIYuv(height_bind, width_bind, 444);
    pYuvD1_temp= new CIYuv(height, width, 444);
    pYuvBind= new CIYuv(height_bind, width_bind, 444);
  } else if ( m_mode == QUARTER_L ) {
    pYuvDBind= new CIYuv(height_bind, width_bind, 420);
    pYuvDBindPRE= new CIYuv(height_bind, width_bind, 420);
    pYuvDBind_work= new CIYuv(height_bind, width_bind, 444);
    pYuvOut= new CIYuv(height, width, 420); // Full width
    pYuvDBind_work= new CIYuv(height_bind, width_bind, 444);
    pYuvD1_temp= new CIYuv(height, width, 444);
    pYuvBind= new CIYuv(height_bind, width_bind, 444);
    pYuvD1_half= new CIYuv(height, width, 444);
  } else { // QUARTER_STACK, HALF_QUARTER_STACK, QUARTER_ODD
    if ( m_mode == QUARTER_STACK ) {
      pYuvDBind= new CIYuv(height_bind, width_bind, 420);
      pYuvDBindPRE= new CIYuv(height_bind, width_bind, 420);
      pYuvDBind_work= new CIYuv(height_bind, width_bind, 444);
      pYuvOut= new CIYuv(height_bind, width_bind, 420);
    }
    if ( m_mode == HALF_QUARTER_STACK ) {
      pYuvDBind= new CIYuv(height_bind / 2, width_bind, 420);
      pYuvDBindPRE= new CIYuv(height_bind / 2, width_bind, 420);
      pYuvDBind_work= new CIYuv(height_bind / 2, width_bind, 444);
      pYuvOut= new CIYuv(height_bind, width_bind, 420);
    }
    if ( m_mode == QUARTER_ODD ) {
      pYuvDBind= new CIYuv(height_bind / 2, width_bind, 420);
      pYuvDBindPRE= new CIYuv(height_bind / 2, width_bind, 420);
      pYuvDBind_work= new CIYuv(height_bind / 2, width_bind, 444);
      pYuvOut= new CIYuv(height_bind, width_bind, 420);
    }
    if ( m_mode == QUARTER_ODD_L ) {
      pYuvDBind= new CIYuv(height_bind / 2, width_bind, 420);
      pYuvDBindPRE= new CIYuv(height_bind / 2, width_bind, 420);
      pYuvDBind_work= new CIYuv(height_bind / 2, width_bind, 444);
      pYuvOut= new CIYuv(height_bind, width, 420); // Full width
    }
    if ( m_mode == FULL_ODD_L ) {
      pYuvDBind= new CIYuv(height_bind, width, 420); // DUMMY
      pYuvDBindPRE= new CIYuv(height_bind, width, 420);
      pYuvDBind_work= new CIYuv(height_bind, width, 444); // DUMMY
      pYuvOut= new CIYuv(height_bind, width, 420); // Full width
    }
    pYuvD1_temp= new CIYuv(height_bind, width, 444);
    if ( m_numDepth == DEPTH_2 )
      pYuvD2_temp= new CIYuv(height_bind, width, 444);
    pYuvBind= new CIYuv(height_bind, width_bind, 444);
    pYuvUnBind1= new CIYuv(height_bind / 2, width_bind, 444);
    pYuvUnBind2= new CIYuv(height_bind / 2, width_bind, 444);

    if ( m_mode == QUARTER_STACK ) {
      pYuvD1_half= new CIYuv(m_height_bind, width, 444);
      pYuvD2_half= new CIYuv(m_height_bind, width, 444);
    }
    if ( m_mode == HALF_QUARTER_STACK ) {
      pYuvD1_half= new CIYuv(m_height_bind / 2, width, 444);
      pYuvD2_half= new CIYuv(m_height_bind / 2, width, 444);
    }
    if ( m_mode == QUARTER_ODD || m_mode == QUARTER_ODD_L )
      pYuvD1_half= new CIYuv(m_height_bind, width, 444);
  }

  if (
    pYuv1 == NULL || pYuv2 == NULL || pYuvDBind == NULL || pYuvDBindPRE == NULL
    || pYuvBind == NULL || pYuvOut == NULL
    || pYuvD1 == NULL
    || ( m_numDepth == DEPTH_2 && pYuvD2 == NULL )
    || ( m_numView == VIEW_3 && pYuv3 == NULL )
  ) {
    if (pYuv1 != NULL) delete pYuv1;
    if (pYuv2 != NULL) delete pYuv2;
    if (pYuv3 != NULL) delete pYuv3;
    if (pYuvBind != NULL) delete pYuvBind;
    if (pYuvDBind != NULL) delete pYuvDBind;
    if (pYuvDBindPRE != NULL) delete pYuvDBindPRE;
    if (pYuvD1 != NULL) delete pYuvD1;
    if (pYuvD2 != NULL) delete pYuvD2;
    if (pYuvOut != NULL) delete pYuvOut;
    if (pYuvD1_half != NULL) delete pYuvD1_half;
    if (pYuvD2_half != NULL) delete pYuvD2_half;
    if (pYuvUnBind1 != NULL) delete pYuvUnBind1;
    if (pYuvUnBind2 != NULL) delete pYuvUnBind2;
    return false;
  }

#ifdef DEBUGOUT
  pYuvDebugOut1= new CIYuv(height, width, 420);
  if ( m_numDepth == DEPTH_2 )
    pYuvDebugOut2= new CIYuv(height, width, 420);
#endif

  pYuv1_work= new CIYuv(height, width, 444);
  pYuvTemp[C1]= new CIYuv(height, width, 444);
  pYuvTemp[C1]->setUpsampleFilter(1, 1); // General, Full-pel
  pYuv2_work= new CIYuv(height, width, 444);
  pYuvTemp[C2]= new CIYuv(height, width, 444);
  pYuvTemp[C2]->setUpsampleFilter(1, 1); // General, Full-pel

  if ( m_mode == QUARTER_STACK || m_mode == QUARTER_ODD || m_mode == QUARTER_ODD_L || m_mode == FULL_ODD_L )
    pAreaBind= new CIYuv(height_bind, width_bind, 444);
  if ( m_mode == HALF_QUARTER_STACK )
    pAreaBind= new CIYuv(height_bind / 2, width_bind, 444);
  if ( m_mode == QUARTER || m_mode == QUARTER_L )
    pAreaBind= new CIYuv(height_bind, width_bind, 444);
  pArea1= new CIYuv(height, width, 444);

  if ( m_numView == VIEW_3 ) {
    pYuv3_work= new CIYuv(height, width, 444);
    pYuvTemp[C3]= new CIYuv(height, width, 444);
    pYuvTemp[C3]->setUpsampleFilter(1, 1); // General, Full-pel

    pArea3= new CIYuv(height, width, 444);
    pAreaUnBind1= new CIYuv(height_bind / 2, width_bind, 444);
    pAreaUnBind2= new CIYuv(height_bind / 2, width_bind, 444);
  }

  if (
    pYuv1_work == NULL || pYuvTemp[C1] == NULL || pYuv2_work == NULL || pYuvTemp[C2] == NULL
    || pYuvDBind_work == NULL
    || ( m_numView == VIEW_3 && (pYuv3_work == NULL || pYuvTemp[C3] == NULL ) )
  ) {
    if (pYuv1_work != NULL) delete pYuv1_work;
    if (pYuv2_work != NULL) delete pYuv2_work;
    if (pYuv3_work != NULL) delete pYuv3_work;
    if (pYuvTemp[C1] != NULL) delete pYuvTemp[C1];
    if (pYuvTemp[C2] != NULL) delete pYuvTemp[C2];
    if (pYuvTemp[C3] != NULL) delete pYuvTemp[C3];
    if (pYuvDBind_work != NULL) delete pYuvDBind_work;
    if (pArea1 != NULL) delete pArea1;
    if (pArea3 != NULL) delete pArea3;
    return false;
  }

#ifndef DEBUGOUT
  if ( m_mode == QUARTER_ODD_L || m_mode == FULL_ODD_L || m_mode == FULL_ODD_L )
    m_img= cvCreateImage(cvSize(width, height_bind), 8, 3); // Full width
  else if ( m_mode == QUARTER_L )
    m_img= cvCreateImage(cvSize(width, height), 8, 3); // Full width
  else
    m_img= cvCreateImage(cvSize(width_bind, height_bind), 8, 3);
#else
  m_img= cvCreateImage(cvSize(width, height), 8, 3);
#endif

  m_pResidualsSyn[D1]= new CResidualsInterpolation();
  m_pResidualsSyn[D1]->SetWidth( width );
  m_pResidualsSyn[D1]->SetHeight( height );
  if (
    !m_pResidualsSyn[D1]->InitOneSide(
                      width, height, variparam,
                      near1, far1, near2, far2,
                      RefCamID1, RefCamID2, VirCamID1, AddCamID1, AddCamID2,
                      LEFT, m_mode,
                      Mat_In_1, Mat_Ex_1, Mat_Trans_1,
                      Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                      Mat_In_V1, Mat_Ex_V1, Mat_Trans_V1,
                      Mat_In_A1, Mat_Ex_A1, Mat_Trans_A1,
                      Mat_In_A2, Mat_Ex_A2, Mat_Trans_A2
                      )
  ) return false;

  if ( m_numView == VIEW_3 ) {
    m_pResidualsSyn[D2]= new CResidualsInterpolation();
    m_pResidualsSyn[D2]->SetWidth( width );
    m_pResidualsSyn[D2]->SetHeight( height );
    if (
      !m_pResidualsSyn[D2]->InitOneSide( // CameraOrder: Outer(Right) to Center !!!!!!
                      width, height, variparam,
                      near3, far3, near2, far2,
                      RefCamID3, RefCamID2, VirCamID2, AddCamID4, AddCamID3,
                      RIGHT, m_mode,
                      Mat_In_3, Mat_Ex_3, Mat_Trans_3,
                      Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                      Mat_In_V2, Mat_Ex_V2, Mat_Trans_V2,
                      Mat_In_A4, Mat_Ex_A4, Mat_Trans_A4,
                      Mat_In_A3, Mat_Ex_A3, Mat_Trans_A3
                      )
    ) return false;
  }


  return true;
}



bool CSubSepMix::ReadOneFrame(
                  FILE *fp1, FILE *fp2, FILE *fp3, FILE *fpD,
                  unsigned int fnum, unsigned int fnumD,
                  // for VariParam
                  FILE *f_D1near, FILE *f_D1far, FILE *f_D1trans,
                  FILE *f_D2near, FILE *f_D2far, FILE *f_D2trans,
                  FILE *f_D3near, FILE *f_D3far, FILE *f_D3trans,
                  FILE *f_V1near, FILE *f_V1far, FILE *f_V1trans,
                  FILE *f_V2near, FILE *f_V2far, FILE *f_V2trans,
                  FILE *f_A1near, FILE *f_A1far, FILE *f_A1trans,
                  FILE *f_A2near, FILE *f_A2far, FILE *f_A2trans,
                  FILE *f_A3near, FILE *f_A3far, FILE *f_A3trans,
                  FILE *f_A4near, FILE *f_A4far, FILE *f_A4trans
) {
  BYTE*** RefV1= pYuvTemp[C1]->getData();
  BYTE*** RefV2= pYuvTemp[C2]->getData();
  BYTE*** RefV3; if ( m_numView == VIEW_3 ) RefV3= pYuvTemp[C3]->getData();
  BYTE** RefD1= pYuvD1->Y;
  BYTE** RefD2;  if ( m_numDepth == DEPTH_2 ) RefD2= pYuvD2->Y;
  double D1near, D1far, D1trans, D2near, D2far, D2trans, D3near, D3far, D3trans;
  double V1near, V1far, V1trans, V2near, V2far, V2trans;
  double A1near, A1far, A1trans, A2near, A2far, A2trans, A3near, A3far, A3trans, A4near, A4far, A4trans;

  try {
    pYuv1->ReadOneFrame( fp1, fnum );
    pYuv1_work->setData444_inIYUV(pYuv1);
    pYuvTemp[C1]->upsampling(pYuv1_work);
    pYuv2->ReadOneFrame( fp2, fnum );
    pYuv2_work->setData444_inIYUV(pYuv2);
    pYuvTemp[C2]->upsampling(pYuv2_work);
    if ( m_numView == VIEW_3 ) {
      pYuv3->ReadOneFrame( fp3, fnum );
      pYuv3_work->setData444_inIYUV( pYuv3 );
      pYuvTemp[C3]->upsampling(pYuv3_work);
    }

    pYuvDBindPRE->ReadOneFrame( fpD, fnumD );
    if ( m_mode == FULL_ODD_L ) {
      pYuvD1->setData444_inIYUV( pYuvDBindPRE );
    } else {
      // Depth Split/Resize pYuvDepthBind to pYuvD1/pYuvD2
      pYuvDBind->depthEnhanced( pYuvDBindPRE );
      // pYuvDBind_work->setData444_inIYUV( pYuvDBind );
      if ( m_numDepth == DEPTH_1 ) {
        pYuvDBind_work->setData444_inIYUV( pYuvDBind );
        pYuvD1_temp->LPF_V_DSingle( pYuvDBind_work, 5, false ); // BQF_V_Single 2011 pYuvD1->BQF_Single( pYuvDBind_work );
        pYuvD1->LPF_H_DSmooth( pYuvD1_temp, 5, false );

      } else {
        pYuvDBind_work->setData444_inIYUV( pYuvDBind );
        // 上下分離して、縦半サイズで出力 depth1
        if ( m_mode == QUARTER_ODD || m_mode == QUARTER_ODD_L ) {
          pYuvD1_half->LPF_V_DSplit( pYuvDBind_work, CENTER, 5, false );
        } else {
          pYuvD1_half->LPF_V_DSplit( pYuvDBind_work, LEFT, 5, false );
          pYuvD2_half->LPF_V_DSplit( pYuvDBind_work, RIGHT,5, false );
        }

        // 横ならし
        pYuvD1->LPF_H_DSmooth( pYuvD1_half, 5, false );
        if ( m_numDepth == DEPTH_2 )
          pYuvD2->LPF_H_DSmooth( pYuvD2_half, 5, false );

        if ( m_mode == HALF_QUARTER_STACK ) {
          // 縦引き伸ばし depth2
          pYuvD1_temp->LPF_V_DExpand( pYuvD1_half, 5, false );
          pYuvD2_temp->LPF_V_DExpand( pYuvD2_half, 5, false );
          // 横ならし
          pYuvD1->LPF_H_DSmooth( pYuvD1_temp, 5, false );
          pYuvD2->LPF_H_DSmooth( pYuvD2_temp, 5, false );
        }
      }
    }

    if (f_D1near != NULL) {
      #ifdef WIN32
      if ( fscanf_s(f_D1near, "%lf", &D1near) != 1 ) return false;
      if ( fscanf_s(f_D1far, "%lf", &D1far) != 1 ) return false;
      if ( fscanf_s(f_D1trans, "%lf", &D1trans) != 1 ) return false;
      if ( fscanf_s(f_D2near, "%lf", &D2near) != 1 ) return false;
      if ( fscanf_s(f_D2far, "%lf", &D2far) != 1 ) return false;
      if ( fscanf_s(f_D2trans, "%lf", &D2trans) != 1 ) return false;
      if ( fscanf_s(f_V1near, "%lf", &V1near) != 1 ) return false;
      if ( fscanf_s(f_V1far, "%lf", &V1far) != 1 ) return false;
      if ( fscanf_s(f_V1trans, "%lf", &V1trans) != 1 ) return false;
      if ( fscanf_s(f_A1near, "%lf", &A1near) != 1 ) return false;
      if ( fscanf_s(f_A1far, "%lf", &A1far) != 1 ) return false;
      if ( fscanf_s(f_A1trans, "%lf", &A1trans) != 1 ) return false;
      if ( fscanf_s(f_A2near, "%lf", &A2near) != 1 ) return false;
      if ( fscanf_s(f_A2far, "%lf", &A2far) != 1 ) return false;
      if ( fscanf_s(f_A2trans, "%lf", &A2trans) != 1 ) return false;
      if ( m_numView == VIEW_3 ) {
        if ( fscanf_s(f_D3near, "%lf", &D3near) != 1 ) return false;
        if ( fscanf_s(f_D3far, "%lf", &D3far) != 1 ) return false;
        if ( fscanf_s(f_D3trans, "%lf", &D3trans) != 1 ) return false;
        if ( fscanf_s(f_V2near, "%lf", &V2near) != 1 ) return false;
        if ( fscanf_s(f_V2far, "%lf", &V2far) != 1 ) return false;
        if ( fscanf_s(f_V2trans, "%lf", &V2trans) != 1 ) return false;
        if ( fscanf_s(f_A3near, "%lf", &A3near) != 1 ) return false;
        if ( fscanf_s(f_A3far, "%lf", &A3far) != 1 ) return false;
        if ( fscanf_s(f_A3trans, "%lf", &A3trans) != 1 ) return false;
        if ( fscanf_s(f_A4near, "%lf", &A4near) != 1 ) return false;
        if ( fscanf_s(f_A4far, "%lf", &A4far) != 1 ) return false;
        if ( fscanf_s(f_A4trans, "%lf", &A4trans) != 1 ) return false;
      }

      #else
      if ( fscanf(f_D1near, "%lf", &D1near) != 1 ) return false;
      if ( fscanf(f_D1far, "%lf", &D1far) != 1 ) return false;
      if ( fscanf(f_D1trans, "%lf", &D1trans) != 1 ) return false;
      if ( fscanf(f_D2near, "%lf", &D2near) != 1 ) return false;
      if ( fscanf(f_D2far, "%lf", &D2far) != 1 ) return false;
      if ( fscanf(f_D2trans, "%lf", &D2trans) != 1 ) return false;
      if ( fscanf(f_V1near, "%lf", &V1near) != 1 ) return false;
      if ( fscanf(f_V1far, "%lf", &V1far) != 1 ) return false;
      if ( fscanf(f_V1trans, "%lf", &V1trans) != 1 ) return false;
      if ( fscanf(f_A1near, "%lf", &A1near) != 1 ) return false;
      if ( fscanf(f_A1far, "%lf", &A1far) != 1 ) return false;
      if ( fscanf(f_A1trans, "%lf", &A1trans) != 1 ) return false;
      if ( fscanf(f_A2near, "%lf", &A2near) != 1 ) return false;
      if ( fscanf(f_A2far, "%lf", &A2far) != 1 ) return false;
      if ( fscanf(f_A2trans, "%lf", &A2trans) != 1 ) return false;
      if ( m_numView == VIEW_3 ) {
        if ( fscanf(f_D3near, "%lf", &D3near) != 1 ) return false;
        if ( fscanf(f_D3far, "%lf", &D3far) != 1 ) return false;
        if ( fscanf(f_D3trans, "%lf", &D3trans) != 1 ) return false;
        if ( fscanf(f_V2near, "%lf", &V2near) != 1 ) return false;
        if ( fscanf(f_V2far, "%lf", &V2far) != 1 ) return false;
        if ( fscanf(f_V2trans, "%lf", &V2trans) != 1 ) return false;
        if ( fscanf(f_A3near, "%lf", &A3near) != 1 ) return false;
        if ( fscanf(f_A3far, "%lf", &A3far) != 1 ) return false;
        if ( fscanf(f_A3trans, "%lf", &A3trans) != 1 ) return false;
        if ( fscanf(f_A4near, "%lf", &A4near) != 1 ) return false;
        if ( fscanf(f_A4far, "%lf", &A4far) != 1 ) return false;
        if ( fscanf(f_A4trans, "%lf", &A4trans) != 1 ) return false;
      }
      #endif

      m_pResidualsSyn[D1]->View = RefV1;
      m_pResidualsSyn[D1]->CenterView = RefV2;
      m_pResidualsSyn[D1]->Depth = RefD1;
      m_pResidualsSyn[D1]->SynYuvBuffer= pYuvTemp[C1];
      m_pResidualsSyn[D1]->AreaBuffer= pArea1;
      m_pResidualsSyn[D1]->near_C1= D1near;
      m_pResidualsSyn[D1]->far_C1= D1far;
      m_pResidualsSyn[D1]->trans_C1= D1trans;
      m_pResidualsSyn[D1]->near_C2= D2near;
      m_pResidualsSyn[D1]->far_C2= D2far;
      m_pResidualsSyn[D1]->trans_C2= D2trans;
      m_pResidualsSyn[D1]->near_V1= V1near;
      m_pResidualsSyn[D1]->far_V1= V1far;
      m_pResidualsSyn[D1]->trans_V1= V1trans;
      m_pResidualsSyn[D1]->near_A1= A1near;
      m_pResidualsSyn[D1]->far_A1= A1far;
      m_pResidualsSyn[D1]->trans_A1= A1trans;
      m_pResidualsSyn[D1]->near_A2= A2near;
      m_pResidualsSyn[D1]->far_A2= A2far;
      m_pResidualsSyn[D1]->trans_A2= A2trans;
      m_pResidualsSyn[D1]->side= LEFT;
      m_pResidualsSyn[D1]->mode= m_mode;

      if ( m_numView == VIEW_3 )
        m_pResidualsSyn[D2]->View = RefV3;
        m_pResidualsSyn[D2]->CenterView = RefV2;


        if ( m_numDepth == DEPTH_2 )
          m_pResidualsSyn[D2]->Depth = RefD2;
        else
          m_pResidualsSyn[D2]->Depth = RefD1;  // Not RefD2 !

        m_pResidualsSyn[D2]->SynYuvBuffer= pYuvTemp[C3];
        m_pResidualsSyn[D2]->AreaBuffer= pArea3;
        m_pResidualsSyn[D2]->near_C1= D3near;
        m_pResidualsSyn[D2]->far_C1= D3far;
        m_pResidualsSyn[D2]->trans_C1= D3trans;
        m_pResidualsSyn[D2]->near_C2= D2near;
        m_pResidualsSyn[D2]->far_C2= D2far;
        m_pResidualsSyn[D2]->trans_C2= D2trans;
        m_pResidualsSyn[D2]->near_V1= V2near;
        m_pResidualsSyn[D2]->far_V1= V2far;
        m_pResidualsSyn[D2]->trans_V1= V2trans;
        m_pResidualsSyn[D2]->near_A1= A4near;
        m_pResidualsSyn[D2]->far_A1= A4far;
        m_pResidualsSyn[D2]->trans_A1= A4trans;
        m_pResidualsSyn[D2]->near_A2= A3near;
        m_pResidualsSyn[D2]->far_A2= A3far;
        m_pResidualsSyn[D2]->trans_A2= A3trans;
        m_pResidualsSyn[D2]->side= RIGHT;
        m_pResidualsSyn[D2]->mode= m_mode;

    } else {
      m_pResidualsSyn[D1]->View = RefV1;
      m_pResidualsSyn[D1]->CenterView = RefV2;
      m_pResidualsSyn[D1]->Depth = RefD1;
      m_pResidualsSyn[D1]->SynYuvBuffer= pYuvTemp[C1];
      m_pResidualsSyn[D1]->AreaBuffer= pArea1;
      m_pResidualsSyn[D1]->near_C1= m_pResidualsSyn[D1]->far_C1= m_pResidualsSyn[D1]->trans_C1= 0.0;
      m_pResidualsSyn[D1]->near_C2= m_pResidualsSyn[D1]->far_C2= m_pResidualsSyn[D1]->trans_C2= 0.0;
      m_pResidualsSyn[D1]->near_V1= m_pResidualsSyn[D1]->far_V1= m_pResidualsSyn[D1]->trans_V1= 0.0;
      m_pResidualsSyn[D1]->near_A1= m_pResidualsSyn[D1]->far_A1= m_pResidualsSyn[D1]->trans_A1= 0.0;
      m_pResidualsSyn[D1]->near_A2= m_pResidualsSyn[D1]->far_A2= m_pResidualsSyn[D1]->trans_A2= 0.0;
      m_pResidualsSyn[D1]->side= LEFT;
      m_pResidualsSyn[D1]->mode= m_mode;

      if ( m_numView == VIEW_3 )
        m_pResidualsSyn[D2]->View = RefV3;
        m_pResidualsSyn[D2]->CenterView = RefV2;

        if ( m_numDepth == DEPTH_2 )
          m_pResidualsSyn[D2]->Depth = RefD2;
        else
          m_pResidualsSyn[D2]->Depth = RefD1; // Not RefD2 !

        m_pResidualsSyn[D2]->SynYuvBuffer= pYuvTemp[C3];
        m_pResidualsSyn[D2]->AreaBuffer= pArea3;
        m_pResidualsSyn[D2]->near_C1= m_pResidualsSyn[D2]->far_C1= m_pResidualsSyn[D2]->trans_C1= 0.0;
        m_pResidualsSyn[D2]->near_C2= m_pResidualsSyn[D2]->far_C2= m_pResidualsSyn[D2]->trans_C2= 0.0;
        m_pResidualsSyn[D2]->near_V1= m_pResidualsSyn[D2]->far_V1= m_pResidualsSyn[D2]->trans_V1= 0.0;
        m_pResidualsSyn[D2]->near_A1= m_pResidualsSyn[D2]->far_A1= m_pResidualsSyn[D2]->trans_A1= 0.0;
        m_pResidualsSyn[D2]->near_A2= m_pResidualsSyn[D2]->far_A2= m_pResidualsSyn[D2]->trans_A2= 0.0;
        m_pResidualsSyn[D2]->side= RIGHT;
        m_pResidualsSyn[D2]->mode= m_mode;

    }

#pragma omp parallel for
    for (int side= D1; side <= m_numView - 2; side++ )
      m_pResidualsSyn[side]->DoOneFrame( fnum );

    pYuv1_work->LPF_V( pYuvTemp[C1], false );
    pYuvTemp[C1]->LPF_H( pYuv1_work, false );
    if ( m_numView == VIEW_3 ) {
      pYuv3_work->LPF_V( pYuvTemp[C3], false );
      pYuvTemp[C3]->LPF_H( pYuv3_work, false );
    }
  } catch (...) {
    return false;
  }
  return true;
}



#ifndef DEBUGOUT
bool CSubSepMix::DoSubMix(FILE *fp) {
  cvZero(m_img);
  unsigned int height_bind;
  if ( m_mode == HALF_QUARTER_STACK )
    height_bind= m_height_bind / 2;
  else
    height_bind= m_height_bind;

  if ( m_numView == VIEW_2 ) {
    for ( unsigned int h= 0; h < m_height_bind; h++ ) {
      for ( unsigned int w= 0; w < m_width_bind; w++ ) {
        pYuvBind->Y[h][w]= pYuvTemp[C1]->Y[h*2][w*2];
        pYuvBind->U[h][w]= pYuvTemp[C1]->U[h*2][w*2];
        pYuvBind->V[h][w]= pYuvTemp[C1]->V[h*2][w*2];

        pAreaBind->Y[h][w]= pArea1->Y[h*2][w*2];
      }
    }

  } else {
    for ( unsigned int h= 0; h < m_height_bind / 2; h++ ) {
      for ( unsigned int w= 0; w < m_width_bind; w++ ) {
        pYuvUnBind1->Y[h][w]= pYuvTemp[C1]->Y[h*2][w*2];
        pYuvUnBind1->U[h][w]= pYuvTemp[C1]->U[h*2][w*2];
        pYuvUnBind1->V[h][w]= pYuvTemp[C1]->V[h*2][w*2];

        pAreaUnBind1->Y[h][w]= pArea1->Y[h*2][w*2];
      }
    }
    for ( unsigned int h= 0; h < m_height_bind / 2; h++ ) {
      for ( unsigned int w= 0; w < m_width_bind; w++ ) {
        pYuvUnBind2->Y[h][w]= pYuvTemp[C3]->Y[h*2][w*2];
        pYuvUnBind2->U[h][w]= pYuvTemp[C3]->U[h*2][w*2];
        pYuvUnBind2->V[h][w]= pYuvTemp[C3]->V[h*2][w*2];

        pAreaUnBind2->Y[h][w]= pArea3->Y[h*2][w*2];
      }
    }

    for ( unsigned int h= 0; h < height_bind/2; h++ ) {
      for ( unsigned int w= 0; w < m_width_bind; w++ ) {
        if ( m_mode != HALF_QUARTER_STACK ) {
          pYuvBind->Y[h][w]= pYuvUnBind1->Y[h][w];
          pYuvBind->U[h][w]= pYuvUnBind1->U[h][w];
          pYuvBind->V[h][w]= pYuvUnBind1->V[h][w];

          pAreaBind->Y[h][w]= pAreaUnBind1->Y[h][w];
        } else { // HALF_QUARTER_STACK
          pYuvBind->Y[h][w]= pYuvUnBind1->Y[h*2][w];
          pYuvBind->U[h][w]= pYuvUnBind1->U[h*2][w];
          pYuvBind->V[h][w]= pYuvUnBind1->V[h*2][w];

          pAreaBind->Y[h][w]= pAreaUnBind1->Y[h*2][w];
        }
      }
    }
    for ( unsigned int h= height_bind/2, h2= 0; h < height_bind; h++, h2++ ) {
      for ( unsigned int w= 0; w < m_width_bind; w++ ) {
        if ( m_mode != HALF_QUARTER_STACK ) {
          pYuvBind->Y[h][w]= pYuvUnBind2->Y[h2][w];
          pYuvBind->U[h][w]= pYuvUnBind2->U[h2][w];
          pYuvBind->V[h][w]= pYuvUnBind2->V[h2][w];

          pAreaBind->Y[h][w]= pAreaUnBind2->Y[h2][w];
        } else { // HALF_QUARTER_STACK
          pYuvBind->Y[h][w]= pYuvUnBind2->Y[h2*2][w];
          pYuvBind->U[h][w]= pYuvUnBind2->U[h2*2][w];
          pYuvBind->V[h][w]= pYuvUnBind2->V[h2*2][w];

          pAreaBind->Y[h][w]= pAreaUnBind2->Y[h2*2][w];
        }
      }
    }

  }

  pYuvBind->looseEdgeBindResiduals( pAreaBind, m_mode );

  // Convert to linear address.
  unsigned int ptv;
  if ( m_mode == QUARTER_ODD_L || m_mode == FULL_ODD_L || m_mode == FULL_ODD_L )
    for ( unsigned int h= 0; h < height_bind; h++ ) {
      for ( unsigned int w= 0; w < m_width; w++ ) {
        ptv= w + h * m_width;
        if ( w < m_width_bind ) {
          m_img->imageData[ptv * 3 + 0]= pYuvBind->Y[h][w];
          m_img->imageData[ptv * 3 + 1]= pYuvBind->U[h][w];
          m_img->imageData[ptv * 3 + 2]= pYuvBind->V[h][w];
        } else {
          m_img->imageData[ptv * 3 + 0]= m_pResidualsSyn[D1]->getAvgColor(0);
          m_img->imageData[ptv * 3 + 1]= m_pResidualsSyn[D1]->getAvgColor(1);
          m_img->imageData[ptv * 3 + 2]= m_pResidualsSyn[D1]->getAvgColor(2);
        }
      }
    }

  else if ( m_mode == QUARTER_L )
    for ( unsigned int h= 0; h < m_height; h++ ) {
      for ( unsigned int w= 0; w < m_width; w++ ) {
        ptv= w + h * m_width;
        if ( w < m_width_bind && h < m_height_bind ) {
          m_img->imageData[ptv * 3 + 0]= pYuvBind->Y[h][w];
          m_img->imageData[ptv * 3 + 1]= pYuvBind->U[h][w];
          m_img->imageData[ptv * 3 + 2]= pYuvBind->V[h][w];
        } else {
          m_img->imageData[ptv * 3 + 0]= m_pResidualsSyn[D1]->getAvgColor(0);
          m_img->imageData[ptv * 3 + 1]= m_pResidualsSyn[D1]->getAvgColor(1);
          m_img->imageData[ptv * 3 + 2]= m_pResidualsSyn[D1]->getAvgColor(2);
        }
      }
    }

  else
    for ( unsigned int h= 0; h < height_bind; h++ ) {
      for ( unsigned int w= 0; w < m_width_bind; w++ ) {
        ptv= w + h * m_width_bind;
        m_img->imageData[ptv * 3 + 0]= pYuvBind->Y[h][w];
        m_img->imageData[ptv * 3 + 1]= pYuvBind->U[h][w];
        m_img->imageData[ptv * 3 + 2]= pYuvBind->V[h][w];
      }
    }

  pYuvOut->setDataFromImgYUV( m_img );
  if (!pYuvOut->WriteOneFrame(fp)) return false;
  return true;
}
#else
bool CSubSepMix::DoSubMix(FILE *fp1, FILE *fp2) {
  cvZero(m_img);

  // Convert to linear address.
  unsigned int ptv;
  for ( unsigned int h= 0; h < m_height; h++ ) {
    for ( unsigned int w= 0; w < m_width; w++ ) {
      ptv= w + h * m_width;
      m_img->imageData[ptv * 3 + 0]= pYuv1_work->Y[h][w];
      m_img->imageData[ptv * 3 + 1]= pYuv1_work->U[h][w];
      m_img->imageData[ptv * 3 + 2]= pYuv1_work->V[h][w];
    }
  }

  pYuvDebugOut1->setDataFromImgYUV( m_img );
  if (!pYuvDebugOut1->WriteOneFrame(fp1)) return false;

  if (m_mode == QUARTER_STACK || m_mode == HALF_QUARTER_STACK) {
    cvZero(m_img);

    // Convert to linear address.
    unsigned int ptv;
    for ( unsigned int h= 0; h < m_height; h++ ) {
      for ( unsigned int w= 0; w < m_width; w++ ) {
        ptv= w + h * m_width;
        m_img->imageData[ptv * 3 + 0]= pYuv3_work->Y[h][w];
        m_img->imageData[ptv * 3 + 1]= pYuv3_work->U[h][w];
        m_img->imageData[ptv * 3 + 2]= pYuv3_work->V[h][w];
      }
    }

    pYuvDebugOut2->setDataFromImgYUV( m_img );
    if (!pYuvDebugOut2->WriteOneFrame(fp2)) return false;
  }

  return true;
}
#endif
